@@ -34,7 +34,7 @@ module Base.Types
     -- * Representation of class methods
   , ClassMethod (..), methodName, methodArity, methodType
     -- * Representation of quantification
-  , TypeScheme (..), ExistTypeScheme (..), monoType, polyType, typeScheme
+  , TypeScheme (..), monoType, polyType, typeScheme
   , rawType
     -- * Predefined types
   , arrowType, unitType, predUnitType, boolType, predBoolType, charType
@@ -67,9 +67,6 @@ import Env.Class (ClassEnv, allSuperClasses)
 -- 'Float'. If the type is not restricted, it defaults to the first type
 -- from the constraint list.
 
--- The case 'TypeSkolem' is used for handling skolem types, which
--- result from the use of existentially quantified data constructors.
-
 -- Type variables are represented with deBruijn style indices. Universally
 -- quantified type variables are assigned indices in the order of their
 -- occurrence in the type from left to right. This leads to a canonical
@@ -88,7 +85,6 @@ data Type
   = TypeConstructor QualIdent
   | TypeVariable Int
   | TypeConstrained [Type] Int
-  | TypeSkolem Int
   | TypeApply Type Type
   | TypeArrow Type Type
   | TypeForall [Int] Type
@@ -117,7 +113,6 @@ unapplyType dflt ty = unapply ty []
     unapply (TypeConstrained tys tv) tys'
       | dflt      = unapply (head tys) tys'
       | otherwise = (TypeConstrained tys tv, tys')
-    unapply (TypeSkolem           k) tys  = (TypeSkolem k, tys)
     unapply (TypeForall     tvs ty') tys  = (TypeForall tvs ty', tys)
 
 -- The function 'rootOfType' returns the name of the type constructor at the
@@ -163,21 +158,17 @@ typeConstrs ty = constrs ty [] where
   constrs (TypeVariable      _) tcs = tcs
   constrs (TypeConstrained _ _) tcs = tcs
   constrs (TypeArrow   ty1 ty2) tcs = constrs ty1 (constrs ty2 tcs)
-  constrs (TypeSkolem        _) tcs = tcs
   constrs (TypeForall    _ ty') tcs = constrs ty' tcs
 
--- The methods 'typeVars' and 'typeSkolems' return a list of all type
--- variables and skolems occurring in a type t, respectively. Note that
--- 'TypeConstrained' variables are not included in the set of type
--- variables because they cannot be generalized.
+-- The method 'typeVars' returns a list of all type variables occurring in a
+-- type t. Note that 'TypeConstrained' variables are not included in the set of
+-- type variables because they cannot be generalized.
 
 class IsType t where
   typeVars :: t -> [Int]
-  typeSkolems :: t -> [Int]
 
 instance IsType Type where
   typeVars = typeVars'
-  typeSkolems = typeSkolems'
 
 typeVars' :: Type -> [Int]
 typeVars' ty = vars ty [] where
@@ -186,19 +177,8 @@ typeVars' ty = vars ty [] where
   vars (TypeVariable     tv) tvs = tv : tvs
   vars (TypeConstrained _ _) tvs = tvs
   vars (TypeArrow   ty1 ty2) tvs = vars ty1 (vars ty2 tvs)
-  vars (TypeSkolem        _) tvs = tvs
   vars (TypeForall tvs' ty') tvs = filter (`notElem` tvs') (typeVars' ty') ++ tvs
 
-typeSkolems' :: Type -> [Int]
-typeSkolems' ty = skolems ty [] where
-  skolems (TypeConstructor   _) sks = sks
-  skolems (TypeApply   ty1 ty2) sks = skolems ty1 (skolems ty2 sks)
-  skolems (TypeVariable      _) sks = sks
-  skolems (TypeConstrained _ _) sks = sks
-  skolems (TypeArrow   ty1 ty2) sks = skolems ty1 (skolems ty2 sks)
-  skolems (TypeSkolem        k) sks = k : sks
-  skolems (TypeForall    _ ty') sks = skolems ty' sks
-
 -- The functions 'qualifyType' and 'unqualifyType' add/remove the
 -- qualification with a module identifier for type constructors.
 
@@ -211,7 +191,6 @@ qualifyType m (TypeConstrained tys tv) =
   TypeConstrained (map (qualifyType m) tys) tv
 qualifyType m (TypeArrow      ty1 ty2) =
   TypeArrow (qualifyType m ty1) (qualifyType m ty2)
-qualifyType _ ts@(TypeSkolem        _) = ts
 qualifyType m (TypeForall      tvs ty) = TypeForall tvs (qualifyType m ty)
 
 unqualifyType :: ModuleIdent -> Type -> Type
@@ -223,7 +202,6 @@ unqualifyType m (TypeConstrained tys tv) =
   TypeConstrained (map (unqualifyType m) tys) tv
 unqualifyType m (TypeArrow      ty1 ty2) =
   TypeArrow (unqualifyType m ty1) (unqualifyType m ty2)
-unqualifyType _ ts@(TypeSkolem        _) = ts
 unqualifyType m (TypeForall      tvs ty) = TypeForall tvs (unqualifyType m ty)
 
 qualifyTC :: ModuleIdent -> QualIdent -> QualIdent
@@ -238,8 +216,13 @@ data Pred = Pred QualIdent Type
   deriving (Eq, Show)
 
 -- We provide a custom 'Ord' instance for predicates here where we consider
--- the type component of the predicate before the class component (see predicate
--- sets below for more information).
+-- the type component of the predicate before the class component. This way,
+-- we ensure that a class method's implicit class constraint is always the
+-- minimum w.r.t. this order, because the type expression for that constraint
+-- is a type variable with index 0 and there are no other class constraints
+-- in a predicate set that constraint the same type variable as restrictions
+-- on class variables are not allowed (see predicate sets below for more
+-- information why this order is relevant).
 
 instance Ord Pred where
   Pred qcls1 ty1 `compare` Pred qcls2 ty2 = case ty1 `compare` ty2 of
@@ -249,7 +232,6 @@ instance Ord Pred where
 
 instance IsType Pred where
   typeVars (Pred _ ty) = typeVars ty
-  typeSkolems (Pred _ ty) = typeSkolems ty
 
 qualifyPred :: ModuleIdent -> Pred -> Pred
 qualifyPred m (Pred qcls ty) = Pred (qualQualify m qcls) (qualifyType m ty)
@@ -273,7 +255,6 @@ type PredSet = Set.Set Pred
 
 instance (IsType a, Ord a) => IsType (Set.Set a) where
   typeVars = concat . Set.toList . Set.map typeVars
-  typeSkolems = concat . Set.toList . Set.map typeSkolems
 
 emptyPredSet :: PredSet
 emptyPredSet = Set.empty
@@ -321,7 +302,6 @@ data PredType = PredType PredSet Type
 
 instance IsType PredType where
   typeVars (PredType ps ty) = typeVars ty ++ typeVars ps
-  typeSkolems (PredType ps ty) = typeSkolems ty ++ typeSkolems ps
 
 predType :: Type -> PredType
 predType = PredType emptyPredSet
@@ -342,31 +322,30 @@ unqualifyPredType m (PredType ps ty) =
 -- ---------------------------------------------------------------------------
 
 -- The type 'DataConstr' is used to represent value or record constructors
--- introduced by data or newtype declarations. The 'Int' denotes the number
--- of existentially quantified type variables in the types.
+-- introduced by data or newtype declarations.
 
-data DataConstr = DataConstr   Ident Int PredSet [Type]
-                | RecordConstr Ident Int PredSet [Ident] [Type]
+data DataConstr = DataConstr   Ident [Type]
+                | RecordConstr Ident [Ident] [Type]
   deriving (Eq, Show)
 
 constrIdent :: DataConstr -> Ident
-constrIdent (DataConstr     c _ _ _) = c
-constrIdent (RecordConstr c _ _ _ _) = c
+constrIdent (DataConstr     c _) = c
+constrIdent (RecordConstr c _ _) = c
 
 constrTypes :: DataConstr -> [Type]
-constrTypes (DataConstr     _ _ _ tys) = tys
-constrTypes (RecordConstr _ _ _ _ tys) = tys
+constrTypes (DataConstr     _ tys) = tys
+constrTypes (RecordConstr _ _ tys) = tys
 
 recLabels :: DataConstr -> [Ident]
-recLabels (DataConstr      _ _ _ _) = []
-recLabels (RecordConstr _ _ _ ls _) = ls
+recLabels (DataConstr      _ _) = []
+recLabels (RecordConstr _ ls _) = ls
 
 recLabelTypes :: DataConstr -> [Type]
-recLabelTypes (DataConstr       _ _ _ _) = []
-recLabelTypes (RecordConstr _ _ _ _ tys) = tys
+recLabelTypes (DataConstr       _ _) = []
+recLabelTypes (RecordConstr _ _ tys) = tys
 
 tupleData :: [DataConstr]
-tupleData = [DataConstr (tupleId n) 0 emptyPredSet (take n tvs) | n <- [2 ..]]
+tupleData = [DataConstr (tupleId n) (take n tvs) | n <- [2 ..]]
   where tvs = map TypeVariable [0 ..]
 
 -- ---------------------------------------------------------------------------
@@ -393,27 +372,15 @@ methodType (ClassMethod _ _ pty) = pty
 -- Quantification
 -- ---------------------------------------------------------------------------
 
--- We support two kinds of quantifications of types here, universally
--- quantified type schemes (forall alpha . tau(alpha)) and universally
--- and existentially quantified type schemes
--- (forall alpha exists eta . tau(alpha,eta)). In both, quantified type
--- variables are assigned ascending indices starting from 0. Therefore it
--- is sufficient to record the numbers of quantified type variables in
--- the 'ForAll' and 'ForAllExist' constructors. In case of
--- the latter, the first of the two numbers is the number of universally
--- quantified variables and the second the number of existentially
--- quantified variables.
+-- We support only universally quantified type schemes
+-- (forall alpha . tau(alpha)). Quantified type variables are assigned
+-- ascending indices starting from 0. Therefore it is sufficient to record the
+-- numbers of quantified type variables in the 'ForAll' constructor.
 
 data TypeScheme = ForAll Int PredType deriving (Eq, Show)
-data ExistTypeScheme = ForAllExist Int Int PredType deriving (Eq, Show)
 
 instance IsType TypeScheme where
   typeVars (ForAll _ pty) = [tv | tv <- typeVars pty, tv < 0]
-  typeSkolems (ForAll _ pty) = typeSkolems pty
-
-instance IsType ExistTypeScheme where
-  typeVars (ForAllExist _ _ pty) = [tv | tv <- typeVars pty, tv < 0]
-  typeSkolems (ForAllExist _ _ pty) = typeSkolems pty
 
 -- The functions 'monoType' and 'polyType' translate a type tau into a
 -- monomorphic type scheme and a polymorphic type scheme, respectively.
@@ -502,9 +469,9 @@ fractionalTypes = drop 1 numTypes
 predefTypes :: [(Type, [DataConstr])]
 predefTypes =
   [ (arrowType a b, [])
-  , (unitType     , [ DataConstr unitId 0 emptyPredSet [] ])
-  , (listType a   , [ DataConstr nilId  0 emptyPredSet []
-                    , DataConstr consId 0 emptyPredSet [a, listType a]
+  , (unitType     , [ DataConstr unitId [] ])
+  , (listType a   , [ DataConstr nilId  []
+                    , DataConstr consId [a, listType a]
                     ])
   ]
   where a = TypeVariable 0
