@@ -48,12 +48,12 @@ import Curry.Base.SpanInfo
 import Curry.Base.Pretty
 import Curry.Syntax
 
-import Base.Messages       (Message, internalError, posMessage)
+import Base.Messages       (Message, internalError, spanInfoMessage)
 import Base.TopEnv         (allEntities, origName, localBindings, moduleImports)
 import Base.Types          ( Type (..), unapplyType, arrowBase, PredType (..)
                            , DataConstr (..), constrIdent, recLabels
                            , ClassMethod, methodName
-                           , TypeScheme (..), ExistTypeScheme (..) )
+                           , TypeScheme (..), rawType, rootOfType )
 import Base.Utils          (findMultiples)
 
 import Env.ModuleAlias     (AliasEnv)
@@ -166,8 +166,8 @@ checkThing' f tcExport = do
   where
   justTcOr errFun = maybe (report $ errFun f) (const ok) tcExport
 
-  getTc (DataConstructor  _ _ _ (ForAllExist _ _ (PredType _ ty))) = getTc' ty
-  getTc (NewtypeConstructor _ _ (ForAllExist _ _ (PredType _ ty))) = getTc' ty
+  getTc (DataConstructor  _ _ _ (ForAll _ (PredType _ ty))) = getTc' ty
+  getTc (NewtypeConstructor _ _ (ForAll _ (PredType _ ty))) = getTc' ty
   getTc (Label _ _ (ForAll _ (PredType _ (TypeArrow tc' _)))) =
     let (TypeConstructor tc, _) = unapplyType False tc' in tc
   getTc err = internalError $ currentModuleName ++ ".checkThing'.getTc: " ++ show err
@@ -325,26 +325,32 @@ expandLocalModule = do
   tcEnv <- getTyConsEnv
   tyEnv <- getValueEnv
   return $
-       [ exportType t | (_, t) <- localBindings tcEnv ]
+       [ exportType t
+         | (_, t)              <- localBindings tcEnv ]
+    ++ [ exportLabel l' ty
+         | (l, Label l' _ ty)  <- localBindings tyEnv, hasGlobalScope l ]
     ++ [ Export NoSpanInfo f'
          | (f, Value f' _ _ _) <- localBindings tyEnv, hasGlobalScope f ]
-    ++ [ Export NoSpanInfo l'
-         | (l, Label l' _ _)   <- localBindings tyEnv, hasGlobalScope l ]
 
 -- |Expand a module export
 expandImportedModule :: ModuleIdent -> ECM [Export]
 expandImportedModule m = do
   tcEnv <- getTyConsEnv
   tyEnv <- getValueEnv
-  return $ [exportType t |       (_, t) <- moduleImports m tcEnv]
+  return $ [exportType t        | (_, t)             <- moduleImports m tcEnv]
+        ++ [exportLabel l ty    | (_, Label l _ ty)  <- moduleImports m tyEnv]
         ++ [Export NoSpanInfo f | (_, Value f _ _ _) <- moduleImports m tyEnv]
-        ++ [Export NoSpanInfo l | (_, Label l _ _)   <- moduleImports m tyEnv]
 
 exportType :: TypeInfo -> Export
 exportType t = ExportTypeWith NoSpanInfo tc xs
   where tc = origName t
         xs = elements t
 
+exportLabel :: QualIdent -> TypeScheme -> Export
+exportLabel qid ty = case rawType ty of
+  TypeArrow a _ -> ExportTypeWith NoSpanInfo (rootOfType a) [qidIdent qid]
+  _             -> internalError $ "ExportCheck.exportLabel: " ++ show (qid, ty)
+
 -- -----------------------------------------------------------------------------
 -- Canonicalization and joining of exports
 -- -----------------------------------------------------------------------------
@@ -435,13 +441,13 @@ errAmbiguousType :: QualIdent -> [TypeInfo] -> Message
 errAmbiguousType tc tcs = errAmbiguous "type" tc (map origName tcs)
 
 errAmbiguous :: String -> QualIdent -> [QualIdent] -> Message
-errAmbiguous what qn qns = posMessage qn
+errAmbiguous what qn qns = spanInfoMessage qn
   $   text "Ambiguous" <+> text what <+> text (escQualName qn)
   $+$ text "It could refer to:"
   $+$ nest 2 (vcat (map (text . escQualName) qns))
 
 errModuleNotImported :: ModuleIdent -> Message
-errModuleNotImported m = posMessage m $ hsep $ map text
+errModuleNotImported m = spanInfoMessage m $ hsep $ map text
   ["Module", escModuleName m, "not imported"]
 
 errMultipleName :: [Ident] -> Message
@@ -453,13 +459,13 @@ errMultipleType = errMultiple "type"
 errMultiple :: String -> [Ident] -> Message
 errMultiple _    []     = internalError $
   currentModuleName ++ ".errMultiple: empty list"
-errMultiple what (i:is) = posMessage i $
+errMultiple what (i:is) = spanInfoMessage i $
   text "Multiple exports of" <+> text what <+> text (escName i) <+> text "at:"
   $+$ nest 2 (vcat (map showPos (i:is)))
   where showPos = text . showLine . getPosition
 
 errNonDataTypeOrTypeClass :: QualIdent -> Message
-errNonDataTypeOrTypeClass tc = posMessage tc $ hsep $ map text
+errNonDataTypeOrTypeClass tc = spanInfoMessage tc $ hsep $ map text
   [escQualName tc, "is not a data type or type class"]
 
 errOutsideTypeConstructor :: QualIdent -> QualIdent -> Message
@@ -469,18 +475,18 @@ errOutsideTypeLabel :: QualIdent -> QualIdent -> Message
 errOutsideTypeLabel l tc = errOutsideTypeExport "Label" l tc
 
 errOutsideTypeExport :: String -> QualIdent -> QualIdent -> Message
-errOutsideTypeExport what q tc = posMessage q
+errOutsideTypeExport what q tc = spanInfoMessage q
   $   text what <+> text (escQualName q)
          <+> text "outside type export in export list"
   $+$ text "Use `" <> text (qualName tc) <+> parens (text (qualName q))
   <>  text "' instead"
 
 errUndefinedElement :: QualIdent -> Ident -> Message
-errUndefinedElement tc c = posMessage c $ hsep $ map text
+errUndefinedElement tc c = spanInfoMessage c $ hsep $ map text
   [ escName c, "is not a constructor or label of type", escQualName tc ]
 
 errUndefinedMethod :: QualIdent -> Ident -> Message
-errUndefinedMethod cls f = posMessage f $ hsep $ map text
+errUndefinedMethod cls f = spanInfoMessage f $ hsep $ map text
   [ escName f, "is not a method of class", escQualName cls ]
 
 errUndefinedName :: QualIdent -> Message
@@ -490,5 +496,5 @@ errUndefinedTypeOrClass :: QualIdent -> Message
 errUndefinedTypeOrClass = errUndefined "type or class"
 
 errUndefined :: String -> QualIdent -> Message
-errUndefined what tc = posMessage tc $ hsep $ map text
+errUndefined what tc = spanInfoMessage tc $ hsep $ map text
   ["Undefined", what, escQualName tc, "in export list"]
