@@ -19,21 +19,20 @@
 -}
 module Checks.InstanceCheck (instanceCheck) where
 
-import           Control.Monad.Extra        (concatMapM, whileM)
+import           Control.Monad.Extra        (concatMapM, whileM, unless)
 import qualified Control.Monad.State as S   (State, execState, gets, modify)
 import           Data.List                  (nub, partition, sortBy)
 import qualified Data.Map            as Map
 import qualified Data.Set.Extra      as Set
 
 import Curry.Base.Ident
-import Curry.Base.Position
 import Curry.Base.Pretty
 import Curry.Base.SpanInfo
 import Curry.Syntax hiding (impls)
 import Curry.Syntax.Pretty
 
 import Base.CurryTypes
-import Base.Messages (Message, posMessage, message, internalError)
+import Base.Messages (Message, spanInfoMessage, message, internalError)
 import Base.SCC (scc)
 import Base.TypeExpansion
 import Base.Types
@@ -52,7 +51,7 @@ instanceCheck m tcEnv clsEnv inEnv ds =
     iss -> (inEnv, map (errMultipleInstances tcEnv) iss)
   where
     local = map (flip InstSource m) $ concatMap (genInstIdents m tcEnv) ds
-    imported = map (uncurry InstSource) $ map (fmap fst3) $ Map.toList inEnv
+    imported = map (uncurry InstSource . fmap fst3) $ Map.toList inEnv
     state = INCState m inEnv []
 
 -- In order to provide better error messages, we use the following data type
@@ -95,6 +94,8 @@ ok = return ()
 checkDecls :: TCEnv -> ClassEnv -> [Decl a] -> INCM ()
 checkDecls tcEnv clsEnv ds = do
   mapM_ (bindInstance tcEnv clsEnv) ids
+  mapM (declDeriveDataInfo tcEnv clsEnv) (filter isDataDecl tds) >>=
+    mapM_ (bindDerivedInstances clsEnv) . groupDeriveInfos
   mapM (declDeriveInfo tcEnv clsEnv) (filter hasDerivedInstances tds) >>=
     mapM_ (bindDerivedInstances clsEnv) . groupDeriveInfos
   mapM_ (checkInstance tcEnv clsEnv) ids
@@ -102,12 +103,15 @@ checkDecls tcEnv clsEnv ds = do
   where (tds, ods) = partition isTypeDecl ds
         ids = filter isInstanceDecl ods
         dds = filter isDefaultDecl ods
+        isDataDecl (DataDecl    _ _ _ _ _) = True
+        isDataDecl (NewtypeDecl _ _ _ _ _) = True
+        isDataDecl _                       = False
 
 -- First, the compiler adds all explicit instance declarations to the
 -- instance environment.
 
 bindInstance :: TCEnv -> ClassEnv -> Decl a -> INCM ()
-bindInstance tcEnv clsEnv (InstanceDecl _ cx qcls inst ds) = do
+bindInstance tcEnv clsEnv (InstanceDecl _ _ cx qcls inst ds) = do
   m <- getModuleIdent
   let PredType ps _ = expandPolyType m tcEnv clsEnv $
                         QualTypeExpr NoSpanInfo cx inst
@@ -136,31 +140,52 @@ hasDerivedInstances _                          = False
 -- derived classes with respect to the super class hierarchy so that subclass
 -- instances are added to the instance environment after their super classes.
 
-data DeriveInfo = DeriveInfo Position QualIdent PredType [Type] [QualIdent]
+data DeriveInfo = DeriveInfo SpanInfo QualIdent PredType [Type] [QualIdent]
 
 declDeriveInfo :: TCEnv -> ClassEnv -> Decl a -> INCM DeriveInfo
 declDeriveInfo tcEnv clsEnv (DataDecl p tc tvs cs clss) =
-  mkDeriveInfo tcEnv clsEnv p tc tvs (concat cxs) (concat tyss) clss
-  where (cxs, tyss) = unzip (map constrDeclTypes cs)
-        constrDeclTypes (ConstrDecl     _ _ cx _ tys) = (cx, tys)
-        constrDeclTypes (ConOpDecl  _ _ cx ty1 _ ty2) = (cx, [ty1, ty2])
-        constrDeclTypes (RecordDecl      _ _ cx _ fs) = (cx, tys)
+  mkDeriveInfo tcEnv clsEnv p tc tvs (concat tyss) clss
+  where tyss = map constrDeclTypes cs
+        constrDeclTypes (ConstrDecl     _ _ tys) = tys
+        constrDeclTypes (ConOpDecl  _ ty1 _ ty2) = [ty1, ty2]
+        constrDeclTypes (RecordDecl      _ _ fs) = tys
           where tys = [ty | FieldDecl _ ls ty <- fs, _ <- ls]
 declDeriveInfo tcEnv clsEnv (NewtypeDecl p tc tvs nc clss) =
-  mkDeriveInfo tcEnv clsEnv p tc tvs [] [nconstrType nc] clss
+  mkDeriveInfo tcEnv clsEnv p tc tvs [nconstrType nc] clss
 declDeriveInfo _ _ _ =
   internalError "InstanceCheck.declDeriveInfo: no data or newtype declaration"
 
-mkDeriveInfo :: TCEnv -> ClassEnv -> SpanInfo -> Ident -> [Ident] -> Context
-           -> [TypeExpr] -> [QualIdent] -> INCM DeriveInfo
-mkDeriveInfo tcEnv clsEnv spi tc tvs cx tys clss = do
+declDeriveDataInfo :: TCEnv -> ClassEnv -> Decl a -> INCM DeriveInfo
+declDeriveDataInfo tcEnv clsEnv (DataDecl p tc tvs cs _) =
+  mkDeriveDataInfo tcEnv clsEnv p tc tvs (concat tyss)
+  where tyss = map constrDeclTypes cs
+        constrDeclTypes (ConstrDecl     _ _ tys) = tys
+        constrDeclTypes (ConOpDecl  _ ty1 _ ty2) = [ty1, ty2]
+        constrDeclTypes (RecordDecl      _ _ fs) = tys
+          where tys = [ty | FieldDecl _ ls ty <- fs, _ <- ls]
+declDeriveDataInfo tcEnv clsEnv (NewtypeDecl p tc tvs nc _) =
+  mkDeriveDataInfo tcEnv clsEnv p tc tvs [nconstrType nc]
+declDeriveDataInfo _ _ _ = internalError
+  "InstanceCheck.declDeriveDataInfo: no data or newtype declaration"
+
+mkDeriveInfo :: TCEnv -> ClassEnv -> SpanInfo -> Ident -> [Ident]
+             -> [TypeExpr] -> [QualIdent] -> INCM DeriveInfo
+mkDeriveInfo tcEnv clsEnv spi tc tvs tys clss = do
   m <- getModuleIdent
   let otc = qualifyWith m tc
       oclss = map (flip (getOrigName m) tcEnv) clss
-      PredType ps ty = expandConstrType m tcEnv clsEnv otc tvs cx tys
+      PredType ps ty = expandConstrType m tcEnv clsEnv otc tvs tys
       (tys', ty') = arrowUnapply ty
-  return $ DeriveInfo p otc (PredType ps ty') tys' $ sortClasses clsEnv oclss
-  where p = spanInfo2Pos spi
+  return $ DeriveInfo spi otc (PredType ps ty') tys' $ sortClasses clsEnv oclss
+
+mkDeriveDataInfo :: TCEnv -> ClassEnv -> SpanInfo -> Ident -> [Ident]
+                 -> [TypeExpr] -> INCM DeriveInfo
+mkDeriveDataInfo tcEnv clsEnv spi tc tvs tys = do
+  m <- getModuleIdent
+  let otc = qualifyWith m tc
+      PredType ps ty = expandConstrType m tcEnv clsEnv otc tvs tys
+      (tys', ty') = arrowUnapply ty
+  return $ DeriveInfo spi otc (PredType ps ty') tys' [qDataId]
 
 sortClasses :: ClassEnv -> [QualIdent] -> [QualIdent]
 sortClasses clsEnv clss = map fst $ sortBy compareDepth $ map adjoinDepth clss
@@ -168,28 +193,32 @@ sortClasses clsEnv clss = map fst $ sortBy compareDepth $ map adjoinDepth clss
         adjoinDepth cls = (cls, length $ allSuperClasses cls clsEnv)
 
 groupDeriveInfos :: [DeriveInfo] -> [[DeriveInfo]]
-groupDeriveInfos ds = scc bound free ds
+groupDeriveInfos = scc bound free
   where bound (DeriveInfo _ tc _ _ _) = [tc]
         free (DeriveInfo _ _ _ tys _) = concatMap typeConstrs tys
 
 bindDerivedInstances :: ClassEnv -> [DeriveInfo] -> INCM ()
-bindDerivedInstances clsEnv dis = do
+bindDerivedInstances clsEnv dis = unless (any hasDataFunType dis) $ do
   mapM_ (enterInitialPredSet clsEnv) dis
   whileM $ concatMapM (inferPredSets clsEnv) dis >>= updatePredSets
+  where
+    hasDataFunType (DeriveInfo _ _ _ tys clss) =
+      clss == [qDataId] && any isFunType tys
 
 enterInitialPredSet :: ClassEnv -> DeriveInfo -> INCM ()
-enterInitialPredSet clsEnv (DeriveInfo p tc pty _ clss) =
-  mapM_ (bindDerivedInstance clsEnv p tc pty []) clss
+enterInitialPredSet clsEnv (DeriveInfo spi tc pty _ clss) =
+  mapM_ (bindDerivedInstance clsEnv spi tc pty) clss
 
 -- Note: The methods and arities entered into the instance environment have
 -- to match methods and arities of the later generated instance declarations.
+-- TODO: Add remark about value environment entry
 
-bindDerivedInstance :: ClassEnv -> Position -> QualIdent -> PredType -> [Type]
-                    -> QualIdent -> INCM ()
-bindDerivedInstance clsEnv p tc pty tys cls = do
+bindDerivedInstance :: HasSpanInfo s => ClassEnv -> s -> QualIdent -> PredType -> QualIdent
+                    -> INCM ()
+bindDerivedInstance clsEnv p tc pty cls = do
   m <- getModuleIdent
-  (i, ps) <- inferPredSet clsEnv p tc pty tys cls
-  modifyInstEnv $ bindInstInfo i (m, ps, impls)
+  ((i, ps), _) <- inferPredSet clsEnv p tc pty [] cls
+  modifyInstEnv (bindInstInfo i (m, ps, impls))
   where impls | cls == qEqId      = [(eqOpId, 2)]
               | cls == qOrdId     = [(leqOpId, 2)]
               | cls == qEnumId    = [ (succId, 1), (predId, 1), (toEnumId, 1)
@@ -199,15 +228,16 @@ bindDerivedInstance clsEnv p tc pty tys cls = do
               | cls == qBoundedId = [(maxBoundId, 0), (minBoundId, 0)]
               | cls == qReadId    = [(readsPrecId, 2)]
               | cls == qShowId    = [(showsPrecId, 2)]
+              | cls == qDataId    = [(dataEqId, 2), (aValueId, 0)]
               | otherwise         =
                 internalError "InstanceCheck.bindDerivedInstance.impls"
 
-inferPredSets :: ClassEnv -> DeriveInfo -> INCM [(InstIdent, PredSet)]
-inferPredSets clsEnv (DeriveInfo p tc pty tys clss) =
-  mapM (inferPredSet clsEnv p tc pty tys) clss
+inferPredSets :: ClassEnv -> DeriveInfo -> INCM [((InstIdent, PredSet), Bool)]
+inferPredSets clsEnv (DeriveInfo spi tc pty tys clss) =
+  mapM (inferPredSet clsEnv spi tc pty tys) clss
 
-inferPredSet :: ClassEnv -> Position -> QualIdent -> PredType -> [Type]
-             -> QualIdent -> INCM (InstIdent, PredSet)
+inferPredSet :: HasSpanInfo s => ClassEnv -> s -> QualIdent -> PredType -> [Type]
+             -> QualIdent -> INCM ((InstIdent, PredSet), Bool)
 inferPredSet clsEnv p tc (PredType ps inst) tys cls = do
   m <- getModuleIdent
   let doc = ppPred m $ Pred cls inst
@@ -215,25 +245,34 @@ inferPredSet clsEnv p tc (PredType ps inst) tys cls = do
       ps'   = Set.fromList [Pred cls ty | ty <- tys]
       ps''  = Set.fromList [Pred scls inst | scls <- sclss]
       ps''' = ps `Set.union` ps' `Set.union` ps''
-  ps'''' <- reducePredSet p "derived instance" doc clsEnv ps'''
-  mapM_ (reportUndecidable p "derived instance" doc) $ Set.toList ps''''
-  return ((cls, tc), ps'''')
+  (ps4, novarps) <-
+    reducePredSet (cls == qDataId) p "derived instance" doc clsEnv ps'''
+  let ps5 = filter noPolyPred $ Set.toList ps4
+  if any (isDataPred m) (Set.toList novarps ++ ps5) && cls == qDataId
+    then return    (((cls, tc), ps4), False)
+    else mapM_ (reportUndecidable p "derived instance" doc) ps5
+         >> return (((cls, tc), ps4), True)
+  where
+    noPolyPred (Pred _ (TypeVariable _)) = False
+    noPolyPred (Pred _ _               ) = True
+    isDataPred _ (Pred qid _) = qid == qDataId
 
-updatePredSets :: [(InstIdent, PredSet)] -> INCM Bool
-updatePredSets = (=<<) (return . or) . mapM (uncurry updatePredSet)
+updatePredSets :: [((InstIdent, PredSet), Bool)] -> INCM Bool
+updatePredSets = fmap or . mapM (uncurry updatePredSet)
 
-updatePredSet :: InstIdent -> PredSet -> INCM Bool
-updatePredSet i ps = do
+updatePredSet :: (InstIdent, PredSet) -> Bool -> INCM Bool
+updatePredSet (i, ps) enter = do
   inEnv <- getInstEnv
   case lookupInstInfo i inEnv of
     Just (m, ps', is)
+      | not enter -> modifyInstEnv (removeInstInfo i) >> return False
       | ps == ps' -> return False
       | otherwise -> do
         modifyInstEnv $ bindInstInfo i (m, ps, is)
         return True
     Nothing -> internalError "InstanceCheck.updatePredSet"
 
-reportUndecidable :: Position -> String -> Doc -> Pred -> INCM ()
+reportUndecidable :: HasSpanInfo s => s -> String -> Doc -> Pred -> INCM ()
 reportUndecidable p what doc predicate@(Pred _ ty) = do
   m <- getModuleIdent
   case ty of
@@ -251,7 +290,7 @@ reportUndecidable p what doc predicate@(Pred _ ty) = do
 -- satisfied by cx.
 
 checkInstance :: TCEnv -> ClassEnv -> Decl a -> INCM ()
-checkInstance tcEnv clsEnv (InstanceDecl spi cx cls inst _) = do
+checkInstance tcEnv clsEnv (InstanceDecl _ _ cx cls inst _) = do
   m <- getModuleIdent
   let PredType ps ty = expandPolyType m tcEnv clsEnv $
                          QualTypeExpr NoSpanInfo cx inst
@@ -259,10 +298,9 @@ checkInstance tcEnv clsEnv (InstanceDecl spi cx cls inst _) = do
       ps' = Set.fromList [ Pred scls ty | scls <- superClasses ocls clsEnv ]
       doc = ppPred m $ Pred cls ty
       what = "instance declaration"
-  ps'' <- reducePredSet p what doc clsEnv ps'
-  Set.mapM_ (report . errMissingInstance m p what doc) $
-    ps'' `Set.difference` (maxPredSet clsEnv ps)
-  where p = spanInfo2Pos spi
+  (ps'', _) <- reducePredSet False inst what doc clsEnv ps'
+  Set.mapM_ (report . errMissingInstance m inst what doc) $
+    ps'' `Set.difference` maxPredSet clsEnv ps
 checkInstance _ _ _ = ok
 
 -- All types specified in the optional default declaration of a module
@@ -271,17 +309,18 @@ checkInstance _ _ _ = ok
 -- must be empty.
 
 checkDefault :: TCEnv -> ClassEnv -> Decl a -> INCM ()
-checkDefault tcEnv clsEnv (DefaultDecl p tys) =
-  mapM_ (checkDefaultType (spanInfo2Pos p) tcEnv clsEnv) tys
+checkDefault tcEnv clsEnv (DefaultDecl _ tys) =
+  mapM_ (checkDefaultType tcEnv clsEnv) tys
 checkDefault _ _ _ = ok
 
-checkDefaultType :: Position -> TCEnv -> ClassEnv -> TypeExpr -> INCM ()
-checkDefaultType p tcEnv clsEnv ty = do
+checkDefaultType :: TCEnv -> ClassEnv -> TypeExpr -> INCM ()
+checkDefaultType tcEnv clsEnv ty = do
   m <- getModuleIdent
   let PredType _ ty' = expandPolyType m tcEnv clsEnv $
                          QualTypeExpr NoSpanInfo [] ty
-  ps <- reducePredSet p what empty clsEnv (Set.singleton $ Pred qNumId ty')
-  Set.mapM_ (report . errMissingInstance m p what empty) ps
+  (ps, _) <- reducePredSet False ty what empty clsEnv
+    (Set.singleton $ Pred qNumId ty')
+  Set.mapM_ (report . errMissingInstance m ty what empty) ps
   where what = "default declaration"
 
 -- The function 'reducePredSet' simplifies a predicate set of the form
@@ -290,16 +329,20 @@ checkDefaultType p tcEnv clsEnv ty = do
 -- a type variable. An error is reported if the predicate set cannot
 -- be transformed into this form. In addition, we remove all predicates
 -- that are implied by others within the same set.
+-- When the flag is set, all missing Data preds are ignored
 
-reducePredSet :: Position -> String -> Doc -> ClassEnv -> PredSet
-              -> INCM PredSet
-reducePredSet p what doc clsEnv ps = do
+reducePredSet :: HasSpanInfo s => Bool -> s -> String -> Doc -> ClassEnv -> PredSet
+              -> INCM (PredSet, PredSet)
+reducePredSet b p what doc clsEnv ps = do
   m <- getModuleIdent
   inEnv <- getInstEnv
   let (ps1, ps2) = partitionPredSet $ minPredSet clsEnv $ reducePreds inEnv ps
-  Set.mapM_ (report . errMissingInstance m p what doc) ps2
-  return ps1
+      ps2' = if b then Set.filter (isNotDataPred m) ps2 else ps2
+  Set.mapM_ (reportMissing m) ps2' >> return (ps1, ps2)
   where
+    isNotDataPred _ (Pred qid _) = qid /= qDataId
+    reportMissing m pr@(Pred _ _) =
+      report $ errMissingInstance m p what doc pr
     reducePreds inEnv = Set.concatMap $ reducePred inEnv
     reducePred inEnv predicate = maybe (Set.singleton predicate)
                                        (reducePreds inEnv)
@@ -323,7 +366,7 @@ genInstIdents m tcEnv (DataDecl    _ tc _ _ qclss) =
 genInstIdents m tcEnv (NewtypeDecl _ tc _ _ qclss) =
   map (flip (genInstIdent m tcEnv) $ ConstructorType NoSpanInfo $ qualify tc)
       qclss
-genInstIdents m tcEnv (InstanceDecl _ _ qcls ty _) =
+genInstIdents m tcEnv (InstanceDecl _ _ _ qcls ty _) =
   [genInstIdent m tcEnv qcls ty]
 genInstIdents _ _     _                            = []
 
@@ -344,6 +387,13 @@ unqualInstIdent tcEnv (qcls, tc) = (unqual qcls, unqual tc)
   where
     unqual = head . flip reverseLookupByOrigName tcEnv
 
+isFunType :: Type -> Bool
+isFunType (TypeArrow         _ _) = True
+isFunType (TypeApply       t1 t2) = isFunType t1 || isFunType t2
+isFunType (TypeForall      _  ty) = isFunType ty
+isFunType (TypeConstrained tys _) = any isFunType tys
+isFunType _                       = False
+
 -- ---------------------------------------------------------------------------
 -- Error messages
 -- ---------------------------------------------------------------------------
@@ -356,9 +406,9 @@ errMultipleInstances tcEnv iss = message $
     ppInstSource (InstSource i m) = ppInstIdent (unqualInstIdent tcEnv i) <+>
       parens (text "defined in" <+> ppMIdent m)
 
-errMissingInstance :: ModuleIdent -> Position -> String -> Doc -> Pred
+errMissingInstance :: HasSpanInfo s => ModuleIdent -> s -> String -> Doc -> Pred
                    -> Message
-errMissingInstance m p what doc predicate = posMessage p $ vcat
+errMissingInstance m p what doc predicate = spanInfoMessage (getSpanInfo p) $ vcat
   [ text "Missing instance for" <+> ppPred m predicate
   , text "in" <+> text what <+> doc
   ]
