@@ -30,7 +30,7 @@ import           Data.List                (partition, nub)
 
 import Curry.Base.Ident
 import Curry.Base.Position
-import Curry.Base.SpanInfo ()
+import Curry.Base.SpanInfo
 import Curry.Base.Pretty
 import Curry.Syntax
 import Curry.Syntax.Pretty
@@ -39,7 +39,7 @@ import Base.CurryKinds
 import Base.Expr
 import Base.Kinds
 import Base.KindSubst
-import Base.Messages (Message, posMessage, internalError)
+import Base.Messages (Message, spanInfoMessage, internalError)
 import Base.SCC
 import Base.TopEnv
 import Base.Types
@@ -58,7 +58,7 @@ import Env.TypeConstructor
 -- synonyms into the type constructor environment.
 
 kindCheck :: TCEnv -> ClassEnv -> Module a -> ((TCEnv, ClassEnv), [Message])
-kindCheck tcEnv clsEnv (Module _ _ m _ _ ds) = runKCM check initState
+kindCheck tcEnv clsEnv (Module _ _ _ m _ _ ds) = runKCM check initState
   where
     check = do
       checkNonRecursiveTypes tds &&> checkAcyclicSuperClasses cds
@@ -123,7 +123,7 @@ bt (DataDecl     _ tc _ _ _) = [tc]
 bt (ExternalDataDecl _ tc _) = [tc]
 bt (NewtypeDecl  _ tc _ _ _) = [tc]
 bt (TypeDecl       _ tc _ _) = [tc]
-bt (ClassDecl   _ _ cls _ _) = [cls]
+bt (ClassDecl _ _ _ cls _ _) = [cls]
 bt _                         = []
 
 ft :: ModuleIdent -> Decl a -> [Ident]
@@ -139,25 +139,25 @@ instance HasType a => HasType (Maybe a) where
   fts m = maybe id $ fts m
 
 instance HasType (Decl a) where
-  fts _ (InfixDecl             _ _ _ _) = id
-  fts m (DataDecl        _ _ _ cs clss) = fts m cs . fts m clss
-  fts _ (ExternalDataDecl        _ _ _) = id
-  fts m (NewtypeDecl     _ _ _ nc clss) = fts m nc . fts m clss
-  fts m (TypeDecl             _ _ _ ty) = fts m ty
-  fts m (TypeSig                _ _ ty) = fts m ty
-  fts m (FunctionDecl        _ _ _ eqs) = fts m eqs
-  fts _ (ExternalDecl              _ _) = id
-  fts m (PatternDecl           _ _ rhs) = fts m rhs
-  fts _ (FreeDecl                  _ _) = id
-  fts m (DefaultDecl             _ tys) = fts m tys
-  fts m (ClassDecl         _ cx _ _ ds) = fts m cx . fts m ds
-  fts m (InstanceDecl _ cx cls inst ds) =
+  fts _ (InfixDecl               _ _ _ _) = id
+  fts m (DataDecl          _ _ _ cs clss) = fts m cs . fts m clss
+  fts _ (ExternalDataDecl          _ _ _) = id
+  fts m (NewtypeDecl       _ _ _ nc clss) = fts m nc . fts m clss
+  fts m (TypeDecl               _ _ _ ty) = fts m ty
+  fts m (TypeSig                  _ _ ty) = fts m ty
+  fts m (FunctionDecl          _ _ _ eqs) = fts m eqs
+  fts _ (ExternalDecl                _ _) = id
+  fts m (PatternDecl             _ _ rhs) = fts m rhs
+  fts _ (FreeDecl                    _ _) = id
+  fts m (DefaultDecl               _ tys) = fts m tys
+  fts m (ClassDecl         _ _ cx _ _ ds) = fts m cx . fts m ds
+  fts m (InstanceDecl _ _ cx cls inst ds) =
     fts m cx . fts m cls . fts m inst . fts m ds
 
 instance HasType ConstrDecl where
-  fts m (ConstrDecl     _ _ cx _ tys) = fts m cx . fts m tys
-  fts m (ConOpDecl  _ _ cx ty1 _ ty2) = fts m cx . fts m ty1 . fts m ty2
-  fts m (RecordDecl      _ _ cx _ fs) = fts m cx . fts m fs
+  fts m (ConstrDecl     _ _ tys) = fts m tys
+  fts m (ConOpDecl  _ ty1 _ ty2) = fts m ty1 . fts m ty2
+  fts m (RecordDecl      _ _ fs) = fts m fs
 
 instance HasType FieldDecl where
   fts m (FieldDecl _ _ ty) = fts m ty
@@ -186,8 +186,8 @@ instance HasType (Equation a) where
   fts m (Equation _ _ rhs) = fts m rhs
 
 instance HasType (Rhs a) where
-  fts m (SimpleRhs  _ e  ds) = fts m e . fts m ds
-  fts m (GuardedRhs _ es ds) = fts m es . fts m ds
+  fts m (SimpleRhs  _ _ e  ds) = fts m e . fts m ds
+  fts m (GuardedRhs _ _ es ds) = fts m es . fts m ds
 
 instance HasType (CondExpr a) where
   fts m (CondExpr _ g e) = fts m g . fts m e
@@ -213,15 +213,15 @@ instance HasType (Expression a) where
   fts m (LeftSection         _ e _) = fts m e
   fts m (RightSection        _ _ e) = fts m e
   fts m (Lambda              _ _ e) = fts m e
-  fts m (Let                _ ds e) = fts m ds . fts m e
-  fts m (Do               _ stms e) = fts m stms . fts m e
+  fts m (Let              _ _ ds e) = fts m ds . fts m e
+  fts m (Do             _ _ stms e) = fts m stms . fts m e
   fts m (IfThenElse     _ e1 e2 e3) = fts m e1 . fts m e2 . fts m e3
-  fts m (Case             _ _ e as) = fts m e . fts m as
+  fts m (Case           _ _ _ e as) = fts m e . fts m as
 
 instance HasType (Statement a) where
-  fts m (StmtExpr _   e) = fts m e
-  fts m (StmtDecl _  ds) = fts m ds
-  fts m (StmtBind _ _ e) = fts m e
+  fts m (StmtExpr _    e) = fts m e
+  fts m (StmtDecl _ _ ds) = fts m ds
+  fts m (StmtBind _ _  e) = fts m e
 
 instance HasType (Alt a) where
   fts m (Alt _ _ rhs) = fts m rhs
@@ -277,16 +277,16 @@ fc m = foldr fc' []
 checkAcyclicSuperClasses :: [Decl a] -> KCM ()
 checkAcyclicSuperClasses ds = do
   m <- getModuleIdent
-  mapM_ checkClassDecl $ scc bt (\(ClassDecl _ cx _ _ _) -> fc m cx) ds
+  mapM_ checkClassDecl $ scc bt (\(ClassDecl _ _ cx _ _ _) -> fc m cx) ds
 
 checkClassDecl :: [Decl a] -> KCM ()
 checkClassDecl [] =
   internalError "Checks.KindCheck.checkClassDecl: empty list"
-checkClassDecl [ClassDecl _ cx cls _ _] = do
+checkClassDecl [ClassDecl _ _ cx cls _ _] = do
   m <- getModuleIdent
   when (cls `elem` fc m cx) $ report $ errRecursiveClasses [cls]
-checkClassDecl (ClassDecl _ _ cls _ _ : ds) =
-  report $ errRecursiveClasses $ cls : [cls' | ClassDecl _ _ cls' _ _ <- ds]
+checkClassDecl (ClassDecl _ _ _ cls _ _ : ds) =
+  report $ errRecursiveClasses $ cls : [cls' | ClassDecl _ _ _ cls' _ _ <- ds]
 checkClassDecl _ =
   internalError "Checks.KindCheck.checkClassDecl: no class declaration"
 
@@ -314,37 +314,35 @@ bindKind :: ModuleIdent -> TCEnv -> ClassEnv -> TCEnv -> Decl a -> KCM TCEnv
 bindKind m tcEnv' clsEnv tcEnv (DataDecl _ tc tvs cs _) =
   bindTypeConstructor DataType tc tvs (Just KindStar) (map mkData cs) tcEnv
   where
-    mkData (ConstrDecl _ evs cx     c  tys) = mkData' evs cx c  tys
-    mkData (ConOpDecl  _ evs cx ty1 op ty2) = mkData' evs cx op [ty1, ty2]
-    mkData (RecordDecl _ evs cx     c   fs) =
+    mkData (ConstrDecl _     c  tys) = mkData' c  tys
+    mkData (ConOpDecl  _ ty1 op ty2) = mkData' op [ty1, ty2]
+    mkData (RecordDecl _     c   fs) =
       let (labels, tys) = unzip [(l, ty) | FieldDecl _ ls ty <- fs, l <- ls]
-      in  mkRec evs cx c labels tys
-    mkData' evs cx c tys = DataConstr c (length evs) ps tys'
+      in  mkRec c labels tys
+    mkData' c tys = DataConstr c tys'
       where qtc = qualifyWith m tc
-            tvs' = tvs ++ evs
-            PredType ps ty = expandConstrType m tcEnv' clsEnv qtc tvs' cx tys
+            PredType _ ty = expandConstrType m tcEnv' clsEnv qtc tvs tys
             tys' = arrowArgs ty
-    mkRec evs cx c ls tys =
-      RecordConstr c (length evs) ps ls tys'
+    mkRec c ls tys =
+      RecordConstr c ls tys'
       where qtc = qualifyWith m tc
-            tvs' = tvs ++ evs
-            PredType ps ty = expandConstrType m tcEnv' clsEnv qtc tvs' cx tys
+            PredType _ ty = expandConstrType m tcEnv' clsEnv qtc tvs tys
             tys' = arrowArgs ty
 bindKind _ _     _       tcEnv (ExternalDataDecl _ tc tvs) =
   bindTypeConstructor DataType tc tvs (Just KindStar) [] tcEnv
 bindKind m tcEnv' _      tcEnv (NewtypeDecl _ tc tvs nc _) =
   bindTypeConstructor RenamingType tc tvs (Just KindStar) (mkData nc) tcEnv
   where
-    mkData (NewConstrDecl _ c      ty) = DataConstr c 0 emptyPredSet [ty']
+    mkData (NewConstrDecl _ c      ty) = DataConstr c [ty']
       where ty'  = expandMonoType m tcEnv' tvs ty
-    mkData (NewRecordDecl _ c (l, ty)) = RecordConstr c 0 emptyPredSet [l] [ty']
+    mkData (NewRecordDecl _ c (l, ty)) = RecordConstr c [l] [ty']
       where ty'  = expandMonoType m tcEnv' tvs ty
 bindKind m tcEnv' _      tcEnv (TypeDecl _ tc tvs ty) =
   bindTypeConstructor aliasType tc tvs Nothing ty' tcEnv
   where
     aliasType tc' k = AliasType tc' k $ length tvs
     ty' = expandMonoType m tcEnv' tvs ty
-bindKind m tcEnv' clsEnv tcEnv (ClassDecl _ _ cls tv ds) =
+bindKind m tcEnv' clsEnv tcEnv (ClassDecl _ _ _ cls tv ds) =
   bindTypeClass cls (concatMap mkMethods ds) tcEnv
   where
     mkMethods (TypeSig _ fs qty) = map (mkMethod qty) fs
@@ -392,7 +390,7 @@ bindTypeVar :: Ident -> Kind -> TCEnv -> TCEnv
 bindTypeVar ident k = bindTopEnv ident (TypeVar k)
 
 bindClass :: ModuleIdent -> TCEnv -> ClassEnv -> Decl a -> ClassEnv
-bindClass m tcEnv clsEnv (ClassDecl _  cx cls _ ds) =
+bindClass m tcEnv clsEnv (ClassDecl _ _ cx cls _ ds) =
   bindClassInfo qcls (sclss, ms) clsEnv
   where qcls = qualifyWith m cls
         ms = map (\f -> (f, f `elem` fs)) $ concatMap methods ds
@@ -444,60 +442,54 @@ kcDecl _     (ExternalDataDecl _ _ _) = ok
 kcDecl tcEnv (NewtypeDecl _ tc tvs nc _) = do
   (_, tcEnv') <- bindTypeVars tc tvs tcEnv
   kcNewConstrDecl tcEnv' nc
-kcDecl tcEnv t@(TypeDecl p tc tvs ty) = do
+kcDecl tcEnv t@(TypeDecl _ tc tvs ty) = do
   (k, tcEnv') <- bindTypeVars tc tvs tcEnv
-  kcType tcEnv' p "type declaration" (ppDecl t) k ty
-kcDecl tcEnv (TypeSig p _ qty) = kcTypeSig tcEnv p qty
+  kcType tcEnv' "type declaration" (pPrint t) k ty
+kcDecl tcEnv (TypeSig _ _ qty) = kcTypeSig tcEnv qty
 kcDecl tcEnv (FunctionDecl _ _ _ eqs) = mapM_ (kcEquation tcEnv) eqs
 kcDecl _     (ExternalDecl _ _) = ok
 kcDecl tcEnv (PatternDecl _ _ rhs) = kcRhs tcEnv rhs
 kcDecl _     (FreeDecl _ _) = ok
-kcDecl tcEnv (DefaultDecl p tys) = do
+kcDecl tcEnv (DefaultDecl _ tys) = do
   tcEnv' <- foldM bindFreshKind tcEnv $ nub $ fv tys
-  mapM_ (kcValueType tcEnv' p "default declaration" empty) tys
-kcDecl tcEnv (ClassDecl p cx cls tv ds) = do
+  mapM_ (kcValueType tcEnv' "default declaration" empty) tys
+kcDecl tcEnv (ClassDecl _ _ cx cls tv ds) = do
   m <- getModuleIdent
   let tcEnv' = bindTypeVar tv (clsKind m (qualifyWith m cls) tcEnv) tcEnv
-  kcContext tcEnv' p cx
+  kcContext tcEnv' cx
   mapM_ (kcDecl tcEnv') ds
-kcDecl tcEnv (InstanceDecl p cx qcls inst ds) = do
+kcDecl tcEnv (InstanceDecl p _ cx qcls inst ds) = do
   m <- getModuleIdent
   tcEnv' <- foldM bindFreshKind tcEnv $ fv inst
-  kcContext tcEnv' p cx
-  kcType tcEnv' p what doc (clsKind m qcls tcEnv) inst
+  kcContext tcEnv' cx
+  kcType tcEnv' what doc (clsKind m qcls tcEnv) inst
   mapM_ (kcDecl tcEnv') ds
     where
       what = "instance declaration"
-      doc = ppDecl (InstanceDecl p cx qcls inst [])
+      doc = pPrint (InstanceDecl p WhitespaceLayout cx qcls inst [])
 
 kcConstrDecl :: TCEnv -> ConstrDecl -> KCM ()
-kcConstrDecl tcEnv d@(ConstrDecl p evs cx _ tys) = do
-  tcEnv' <- foldM bindFreshKind tcEnv evs
-  kcContext tcEnv' p cx
-  mapM_ (kcValueType tcEnv' p what doc) tys
+kcConstrDecl tcEnv d@(ConstrDecl _ _ tys) = do
+  mapM_ (kcValueType tcEnv what doc) tys
     where
       what = "data constructor declaration"
-      doc = ppConstr d
-kcConstrDecl tcEnv d@(ConOpDecl p evs cx ty1 _ ty2) = do
-  tcEnv' <- foldM bindFreshKind tcEnv evs
-  kcContext tcEnv' p cx
-  kcValueType tcEnv' p what doc ty1
-  kcValueType tcEnv' p what doc ty2
+      doc = pPrint d
+kcConstrDecl tcEnv d@(ConOpDecl _ ty1 _ ty2) = do
+  kcValueType tcEnv what doc ty1
+  kcValueType tcEnv what doc ty2
     where
       what = "data constructor declaration"
-      doc = ppConstr d
-kcConstrDecl tcEnv (RecordDecl p evs cx _ fs) = do
-  tcEnv' <- foldM bindFreshKind tcEnv evs
-  kcContext tcEnv' p cx
-  mapM_ (kcFieldDecl tcEnv') fs
+      doc = pPrint d
+kcConstrDecl tcEnv (RecordDecl _ _ fs) = do
+  mapM_ (kcFieldDecl tcEnv) fs
 
 kcFieldDecl :: TCEnv -> FieldDecl -> KCM ()
-kcFieldDecl tcEnv d@(FieldDecl p _ ty) =
-  kcValueType tcEnv p "field declaration" (ppFieldDecl d) ty
+kcFieldDecl tcEnv d@(FieldDecl _ _ ty) =
+  kcValueType tcEnv "field declaration" (pPrint d) ty
 
 kcNewConstrDecl :: TCEnv -> NewConstrDecl -> KCM ()
-kcNewConstrDecl tcEnv d@(NewConstrDecl p _ ty) =
-  kcValueType tcEnv p "newtype constructor declaration" (ppNewConstr d) ty
+kcNewConstrDecl tcEnv d@(NewConstrDecl _ _ ty) =
+  kcValueType tcEnv "newtype constructor declaration" (pPrint d) ty
 kcNewConstrDecl tcEnv (NewRecordDecl p _ (l, ty)) =
   kcFieldDecl tcEnv (FieldDecl p [l] ty)
 
@@ -505,110 +497,110 @@ kcEquation :: TCEnv -> Equation a -> KCM ()
 kcEquation tcEnv (Equation _ _ rhs) = kcRhs tcEnv rhs
 
 kcRhs :: TCEnv -> Rhs a -> KCM ()
-kcRhs tcEnv (SimpleRhs p e ds) = do
-  kcExpr tcEnv p e
+kcRhs tcEnv (SimpleRhs _ _ e ds) = do
+  kcExpr tcEnv e
   mapM_ (kcDecl tcEnv) ds
-kcRhs tcEnv (GuardedRhs _ es ds) = do
+kcRhs tcEnv (GuardedRhs _ _ es ds) = do
   mapM_ (kcCondExpr tcEnv) es
   mapM_ (kcDecl tcEnv) ds
 
 kcCondExpr :: TCEnv -> CondExpr a -> KCM ()
-kcCondExpr tcEnv (CondExpr p g e) = kcExpr tcEnv p g >> kcExpr tcEnv p e
+kcCondExpr tcEnv (CondExpr _ g e) = kcExpr tcEnv g >> kcExpr tcEnv e
 
-kcExpr :: HasPosition p => TCEnv -> p -> Expression a -> KCM ()
-kcExpr _     _ (Literal _ _ _) = ok
-kcExpr _     _ (Variable _ _ _) = ok
-kcExpr _     _ (Constructor _ _ _) = ok
-kcExpr tcEnv p (Paren _ e) = kcExpr tcEnv p e
-kcExpr tcEnv p (Typed _ e qty) = do
-  kcExpr tcEnv p e
-  kcTypeSig tcEnv p qty
-kcExpr tcEnv p (Record _ _ _ fs) = mapM_ (kcField tcEnv p) fs
-kcExpr tcEnv p (RecordUpdate _ e fs) = do
-  kcExpr tcEnv p e
-  mapM_ (kcField tcEnv p) fs
-kcExpr tcEnv p (Tuple _ es) = mapM_ (kcExpr tcEnv p) es
-kcExpr tcEnv p (List _ _ es) = mapM_ (kcExpr tcEnv p) es
-kcExpr tcEnv p (ListCompr _ e stms) = do
-  kcExpr tcEnv p e
-  mapM_ (kcStmt tcEnv p) stms
-kcExpr tcEnv p (EnumFrom _ e) = kcExpr tcEnv p e
-kcExpr tcEnv p (EnumFromThen _ e1 e2) = do
-  kcExpr tcEnv p e1
-  kcExpr tcEnv p e2
-kcExpr tcEnv p (EnumFromTo _ e1 e2) = do
-  kcExpr tcEnv p e1
-  kcExpr tcEnv p e2
-kcExpr tcEnv p (EnumFromThenTo _ e1 e2 e3) = do
-  kcExpr tcEnv p e1
-  kcExpr tcEnv p e2
-  kcExpr tcEnv p e3
-kcExpr tcEnv p (UnaryMinus _ e) = kcExpr tcEnv p e
-kcExpr tcEnv p (Apply _ e1 e2) = do
-  kcExpr tcEnv p e1
-  kcExpr tcEnv p e2
-kcExpr tcEnv p (InfixApply _ e1 _ e2) = do
-  kcExpr tcEnv p e1
-  kcExpr tcEnv p e2
-kcExpr tcEnv p (LeftSection _ e _) = kcExpr tcEnv p e
-kcExpr tcEnv p (RightSection _ _ e) = kcExpr tcEnv p e
-kcExpr tcEnv p (Lambda _ _ e) = kcExpr tcEnv p e
-kcExpr tcEnv p (Let _ ds e) = do
+kcExpr :: TCEnv -> Expression a -> KCM ()
+kcExpr _     (Literal _ _ _) = ok
+kcExpr _     (Variable _ _ _) = ok
+kcExpr _     (Constructor _ _ _) = ok
+kcExpr tcEnv (Paren _ e) = kcExpr tcEnv e
+kcExpr tcEnv (Typed _ e qty) = do
+  kcExpr tcEnv e
+  kcTypeSig tcEnv qty
+kcExpr tcEnv (Record _ _ _ fs) = mapM_ (kcField tcEnv) fs
+kcExpr tcEnv (RecordUpdate _ e fs) = do
+  kcExpr tcEnv e
+  mapM_ (kcField tcEnv) fs
+kcExpr tcEnv (Tuple _ es) = mapM_ (kcExpr tcEnv) es
+kcExpr tcEnv (List _ _ es) = mapM_ (kcExpr tcEnv) es
+kcExpr tcEnv (ListCompr _ e stms) = do
+  kcExpr tcEnv e
+  mapM_ (kcStmt tcEnv) stms
+kcExpr tcEnv (EnumFrom _ e) = kcExpr tcEnv e
+kcExpr tcEnv (EnumFromThen _ e1 e2) = do
+  kcExpr tcEnv e1
+  kcExpr tcEnv e2
+kcExpr tcEnv (EnumFromTo _ e1 e2) = do
+  kcExpr tcEnv e1
+  kcExpr tcEnv e2
+kcExpr tcEnv (EnumFromThenTo _ e1 e2 e3) = do
+  kcExpr tcEnv e1
+  kcExpr tcEnv e2
+  kcExpr tcEnv e3
+kcExpr tcEnv (UnaryMinus _ e) = kcExpr tcEnv e
+kcExpr tcEnv (Apply _ e1 e2) = do
+  kcExpr tcEnv e1
+  kcExpr tcEnv e2
+kcExpr tcEnv (InfixApply _ e1 _ e2) = do
+  kcExpr tcEnv e1
+  kcExpr tcEnv e2
+kcExpr tcEnv (LeftSection _ e _) = kcExpr tcEnv e
+kcExpr tcEnv (RightSection _ _ e) = kcExpr tcEnv e
+kcExpr tcEnv (Lambda _ _ e) = kcExpr tcEnv e
+kcExpr tcEnv (Let _ _ ds e) = do
   mapM_ (kcDecl tcEnv) ds
-  kcExpr tcEnv p e
-kcExpr tcEnv p (Do _ stms e) = do
-  mapM_ (kcStmt tcEnv p) stms
-  kcExpr tcEnv p e
-kcExpr tcEnv p (IfThenElse _ e1 e2 e3) = do
-  kcExpr tcEnv p e1
-  kcExpr tcEnv p e2
-  kcExpr tcEnv p e3
-kcExpr tcEnv p (Case _ _ e alts) = do
-  kcExpr tcEnv p e
+  kcExpr tcEnv e
+kcExpr tcEnv (Do _ _ stms e) = do
+  mapM_ (kcStmt tcEnv) stms
+  kcExpr tcEnv e
+kcExpr tcEnv (IfThenElse _ e1 e2 e3) = do
+  kcExpr tcEnv e1
+  kcExpr tcEnv e2
+  kcExpr tcEnv e3
+kcExpr tcEnv (Case _ _ _ e alts) = do
+  kcExpr tcEnv e
   mapM_ (kcAlt tcEnv) alts
 
-kcStmt :: HasPosition p => TCEnv -> p -> Statement a -> KCM ()
-kcStmt tcEnv p (StmtExpr _ e) = kcExpr tcEnv p e
-kcStmt tcEnv _ (StmtDecl _ ds) = mapM_ (kcDecl tcEnv) ds
-kcStmt tcEnv p (StmtBind _ _ e) = kcExpr tcEnv p e
+kcStmt :: TCEnv -> Statement a -> KCM ()
+kcStmt tcEnv (StmtExpr _ e) = kcExpr tcEnv e
+kcStmt tcEnv (StmtDecl _ _ ds) = mapM_ (kcDecl tcEnv) ds
+kcStmt tcEnv (StmtBind _ _ e) = kcExpr tcEnv e
 
 kcAlt :: TCEnv -> Alt a -> KCM ()
 kcAlt tcEnv (Alt _ _ rhs) = kcRhs tcEnv rhs
 
-kcField :: HasPosition p => TCEnv -> p -> Field (Expression a) -> KCM ()
-kcField tcEnv p (Field _ _ e) = kcExpr tcEnv p e
+kcField :: TCEnv -> Field (Expression a) -> KCM ()
+kcField tcEnv (Field _ _ e) = kcExpr tcEnv e
 
-kcContext :: HasPosition p => TCEnv -> p -> Context -> KCM ()
-kcContext tcEnv p = mapM_ (kcConstraint tcEnv p)
+kcContext :: TCEnv -> Context -> KCM ()
+kcContext tcEnv = mapM_ (kcConstraint tcEnv)
 
-kcConstraint :: HasPosition p => TCEnv -> p -> Constraint -> KCM ()
-kcConstraint tcEnv p sc@(Constraint _ qcls ty) = do
+kcConstraint :: TCEnv -> Constraint -> KCM ()
+kcConstraint tcEnv sc@(Constraint _ qcls ty) = do
   m <- getModuleIdent
-  kcType tcEnv p "class constraint" doc (clsKind m qcls tcEnv) ty
+  kcType tcEnv "class constraint" doc (clsKind m qcls tcEnv) ty
   where
-    doc = ppConstraint sc
+    doc = pPrint sc
 
-kcTypeSig :: HasPosition p => TCEnv -> p -> QualTypeExpr -> KCM ()
-kcTypeSig tcEnv p (QualTypeExpr _ cx ty) = do
+kcTypeSig :: TCEnv -> QualTypeExpr -> KCM ()
+kcTypeSig tcEnv (QualTypeExpr _ cx ty) = do
   tcEnv' <- foldM bindFreshKind tcEnv free
-  kcContext tcEnv' p cx
-  kcValueType tcEnv' p "type signature" doc ty
+  kcContext tcEnv' cx
+  kcValueType tcEnv' "type signature" doc ty
   where
     free = filter (null . flip lookupTypeInfo tcEnv) $ nub $ fv ty
-    doc = ppTypeExpr 0 ty
+    doc = pPrintPrec 0 ty
 
-kcValueType :: HasPosition p => TCEnv -> p -> String -> Doc -> TypeExpr -> KCM ()
-kcValueType tcEnv p what doc = kcType tcEnv p what doc KindStar
+kcValueType :: TCEnv -> String -> Doc -> TypeExpr -> KCM ()
+kcValueType tcEnv what doc = kcType tcEnv what doc KindStar
 
-kcType :: HasPosition p => TCEnv -> p -> String -> Doc -> Kind -> TypeExpr -> KCM ()
-kcType tcEnv p what doc k ty = do
-  k' <- kcTypeExpr tcEnv p "type expression" doc' 0 ty
-  unify p what (doc $-$ text "Type:" <+> doc') k k'
+kcType :: TCEnv -> String -> Doc -> Kind -> TypeExpr -> KCM ()
+kcType tcEnv what doc k ty = do
+  k' <- kcTypeExpr tcEnv "type expression" doc' 0 ty
+  unify ty what (doc $-$ text "Type:" <+> doc') k k'
   where
-    doc' = ppTypeExpr 0 ty
+    doc' = pPrintPrec 0 ty
 
-kcTypeExpr :: HasPosition p => TCEnv -> p -> String -> Doc -> Int -> TypeExpr -> KCM Kind
-kcTypeExpr tcEnv p _ _ n (ConstructorType _ tc) = do
+kcTypeExpr :: TCEnv -> String -> Doc -> Int -> TypeExpr -> KCM Kind
+kcTypeExpr tcEnv _ _ n (ConstructorType p tc) = do
   m <- getModuleIdent
   case qualLookupTypeInfo tc tcEnv of
     [AliasType _ _ n' _] -> case n >= n' of
@@ -617,29 +609,29 @@ kcTypeExpr tcEnv p _ _ n (ConstructorType _ tc) = do
         report $ errPartialAlias p tc n' n
         freshKindVar
     _ -> return $ tcKind m tc tcEnv
-kcTypeExpr tcEnv p what doc n (ApplyType _ ty1 ty2) = do
-  (alpha, beta) <- kcTypeExpr tcEnv p what doc (n + 1) ty1 >>=
-    kcArrow p what (doc $-$ text "Type:" <+> ppTypeExpr 0 ty1)
-  kcTypeExpr tcEnv p what doc 0 ty2 >>=
-    unify p what (doc $-$ text "Type:" <+> ppTypeExpr 0 ty2) alpha
+kcTypeExpr tcEnv what doc n (ApplyType p ty1 ty2) = do
+  (alpha, beta) <- kcTypeExpr tcEnv what doc (n + 1) ty1 >>=
+    kcArrow p what (doc $-$ text "Type:" <+> pPrintPrec 0 ty1)
+  kcTypeExpr tcEnv what doc 0 ty2 >>=
+    unify p what (doc $-$ text "Type:" <+> pPrintPrec 0 ty2) alpha
   return beta
-kcTypeExpr tcEnv _ _ _ _ (VariableType _ tv) = return (varKind tv tcEnv)
-kcTypeExpr tcEnv p what doc _ (TupleType _ tys) = do
-  mapM_ (kcValueType tcEnv p what doc) tys
+kcTypeExpr tcEnv _ _ _      (VariableType _ tv) = return (varKind tv tcEnv)
+kcTypeExpr tcEnv what doc _ (TupleType _ tys) = do
+  mapM_ (kcValueType tcEnv what doc) tys
   return KindStar
-kcTypeExpr tcEnv p what doc _ (ListType _ ty) = do
-  kcValueType tcEnv p what doc ty
+kcTypeExpr tcEnv what doc _ (ListType _ ty) = do
+  kcValueType tcEnv what doc ty
   return KindStar
-kcTypeExpr tcEnv p what doc _ (ArrowType _ ty1 ty2) = do
-  kcValueType tcEnv p what doc ty1
-  kcValueType tcEnv p what doc ty2
+kcTypeExpr tcEnv what doc _ (ArrowType _ ty1 ty2) = do
+  kcValueType tcEnv what doc ty1
+  kcValueType tcEnv what doc ty2
   return KindStar
-kcTypeExpr tcEnv p what doc n (ParenType _ ty) = kcTypeExpr tcEnv p what doc n ty
-kcTypeExpr tcEnv p what doc n (ForallType _ vs ty) = do
+kcTypeExpr tcEnv what doc n (ParenType _ ty) = kcTypeExpr tcEnv what doc n ty
+kcTypeExpr tcEnv what doc n (ForallType _ vs ty) = do
   tcEnv' <- foldM bindFreshKind tcEnv vs
-  kcTypeExpr tcEnv' p what doc n ty
+  kcTypeExpr tcEnv' what doc n ty
 
-kcArrow :: HasPosition p => p -> String -> Doc -> Kind -> KCM (Kind, Kind)
+kcArrow :: HasSpanInfo p => p -> String -> Doc -> Kind -> KCM (Kind, Kind)
 kcArrow p what doc k = do
   theta <- getKindSubst
   case subst theta k of
@@ -658,7 +650,7 @@ kcArrow p what doc k = do
 -- ---------------------------------------------------------------------------
 
 -- The unification uses Robinson's algorithm.
-unify :: HasPosition p => p -> String -> Doc -> Kind -> Kind -> KCM ()
+unify :: HasSpanInfo p => p -> String -> Doc -> Kind -> Kind -> KCM ()
 unify p what doc k1 k2 = do
   theta <- getKindSubst
   let k1' = subst theta k1
@@ -718,9 +710,9 @@ isTypeOrNewtypeDecl _                       = False
 errRecursiveTypes :: [Ident] -> Message
 errRecursiveTypes []       = internalError
   "KindCheck.errRecursiveTypes: empty list"
-errRecursiveTypes [tc]     = posMessage tc $ hsep $ map text
+errRecursiveTypes [tc]     = spanInfoMessage tc $ hsep $ map text
   ["Recursive synonym or renaming type", idName tc]
-errRecursiveTypes (tc:tcs) = posMessage tc $
+errRecursiveTypes (tc:tcs) = spanInfoMessage tc $
   text "Mutually recursive synonym and/or renaming types" <+>
     text (idName tc) <> types empty tcs
   where
@@ -733,9 +725,9 @@ errRecursiveTypes (tc:tcs) = posMessage tc $
 errRecursiveClasses :: [Ident] -> Message
 errRecursiveClasses []         = internalError
   "KindCheck.errRecursiveClasses: empty list"
-errRecursiveClasses [cls]      = posMessage cls $ hsep $ map text
+errRecursiveClasses [cls]      = spanInfoMessage cls $ hsep $ map text
   ["Recursive type class", idName cls]
-errRecursiveClasses (cls:clss) = posMessage cls $
+errRecursiveClasses (cls:clss) = spanInfoMessage cls $
   text "Mutually recursive type classes" <+> text (idName cls) <>
     classes empty clss
   where
@@ -745,15 +737,15 @@ errRecursiveClasses (cls:clss) = posMessage cls $
     classPos cls' =
       text (idName cls') <+> parens (text $ showLine $ getPosition cls')
 
-errNonArrowKind :: HasPosition p => p -> String -> Doc -> Kind -> Message
-errNonArrowKind p what doc k = posMessage p $ vcat
+errNonArrowKind :: HasSpanInfo p => p -> String -> Doc -> Kind -> Message
+errNonArrowKind p what doc k = spanInfoMessage p $ vcat
   [ text "Kind error in" <+> text what, doc
   , text "Kind:" <+> ppKind k
   , text "Cannot be applied"
   ]
 
-errPartialAlias :: HasPosition p => p -> QualIdent -> Int -> Int -> Message
-errPartialAlias p tc arity argc = posMessage p $ hsep
+errPartialAlias :: HasSpanInfo p => p -> QualIdent -> Int -> Int -> Message
+errPartialAlias p tc arity argc = spanInfoMessage p $ hsep
   [ text "Type synonym", ppQIdent tc
   , text "requires at least"
   , int arity, text (plural arity "argument") <> comma
@@ -762,8 +754,8 @@ errPartialAlias p tc arity argc = posMessage p $ hsep
   where
     plural n x = if n == 1 then x else x ++ "s"
 
-errKindMismatch ::  HasPosition p => p -> String -> Doc -> Kind -> Kind -> Message
-errKindMismatch p what doc k1 k2 = posMessage p $ vcat
+errKindMismatch :: HasSpanInfo p => p -> String -> Doc -> Kind -> Kind -> Message
+errKindMismatch p what doc k1 k2 = spanInfoMessage p $ vcat
   [ text "Kind error in"  <+> text what, doc
   , text "Inferred kind:" <+> ppKind k2
   , text "Expected kind:" <+> ppKind k1
