@@ -34,33 +34,35 @@ import Prelude hiding ((<>))
 import           Control.Applicative        ((<$>), (<*>))
 #endif
 
-import Control.Monad                      (unless, when)
-import qualified Control.Monad.State as S ( State, runState, gets, modify
-                                          , withState )
-import           Data.Function            (on)
-import           Data.List                (insertBy, intersect, nub, nubBy)
-import qualified Data.Map  as Map         ( Map, empty, findWithDefault
-                                          , fromList, insertWith, keys )
-import           Data.Maybe               (isJust, isNothing)
-import qualified Data.Set as Set          ( Set, empty, insert, member
-                                          , singleton, toList, union)
+import           Control.Monad       (unless, when)
+import qualified Control.Monad.State as S (State, gets, modify, runState,
+                                           withState)
+import           Data.Function       (on)
+import           Data.List           (insertBy, intersect, nub, nubBy)
+import qualified Data.Map            as Map (Map, empty, findWithDefault,
+                                             fromList, insertWith, keys)
+import           Data.Maybe          (isJust, isNothing)
+import qualified Data.Set            as Set (Set, empty, insert, member,
+                                             singleton, toList, union)
 
-import Curry.Base.Ident
-import Curry.Base.Position
-import Curry.Base.Pretty
-import Curry.Base.Span
-import Curry.Base.SpanInfo
-import Curry.Syntax
-import Curry.Syntax.Pretty (ppPattern)
+import           Curry.Base.Ident
+import           Curry.Base.Position
+import           Curry.Base.Pretty
+import           Curry.Base.Span
+import           Curry.Base.SpanInfo
+import           Curry.Syntax
 
-import Base.Expr
-import Base.Messages (Message, posMessage, internalError)
-import Base.NestEnv
-import Base.SCC      (scc)
-import Base.Utils    ((++!), findDouble, findMultiples)
+import           Base.Expr
+import           Base.Messages       (Message, internalError,
+                                      spanInfoMessage)
+import           Base.NestEnv
+import           Base.SCC            (scc)
+import           Base.Utils          (findDouble, findMultiples, (++!))
+
+import           Env.TypeConstructor (TCEnv, clsMethods, getOrigName)
+import           Env.Value           (ValueEnv, ValueInfo (..),
+                                      qualLookupValueUnique)
 
-import Env.TypeConstructor (TCEnv, clsMethods)
-import Env.Value           (ValueEnv, ValueInfo (..))
 
 -- The syntax checking proceeds as follows. First, the compiler extracts
 -- information about all imported values and data constructors from the
@@ -77,7 +79,7 @@ import Env.Value           (ValueEnv, ValueInfo (..))
 
 syntaxCheck :: [KnownExtension] -> TCEnv -> ValueEnv -> Module ()
             -> ((Module (), [KnownExtension]), [Message])
-syntaxCheck exts tcEnv vEnv mdl@(Module _ _ m _ _ ds) =
+syntaxCheck exts tcEnv vEnv mdl@(Module _ _ _ m _ _ ds) =
   case findMultiples cons of
     []  -> case findMultiples (ls ++ fs ++ cons ++ cs) of
              []  -> runSC (checkModule mdl) state
@@ -90,9 +92,9 @@ syntaxCheck exts tcEnv vEnv mdl@(Module _ _ m _ _ ds) =
     cons  = concatMap constrs tds
     ls    = nub $ concatMap recLabels tds
     fs    = nub $ concatMap vars vds
-    cs    = concatMap (concatMap methods) [ds' | ClassDecl _ _ _ _ ds' <- cds]
+    cs    = concatMap (concatMap methods) [ds' | ClassDecl _ _ _ _ _ ds' <- cds]
     rEnv  = globalEnv $ fmap renameInfo vEnv
-    state = initState exts m tcEnv rEnv
+    state = initState exts m tcEnv rEnv vEnv
 
 -- A global state transformer is used for generating fresh integer keys with
 -- which the variables are renamed.
@@ -109,6 +111,7 @@ data SCState = SCState
   , moduleIdent      :: ModuleIdent      -- ^ 'ModuleIdent' of the current module
   , tyConsEnv        :: TCEnv
   , renameEnv        :: RenameEnv        -- ^ Information store
+  , valueEnv         :: ValueEnv         -- ^ To check instance method visibility
   , scopeId          :: Integer          -- ^ Identifier for the current scope
   , nextId           :: Integer          -- ^ Next fresh identifier
   , funcDeps         :: FuncDeps         -- ^ Stores data about functions dependencies
@@ -117,9 +120,10 @@ data SCState = SCState
   }
 
 -- |Initial syntax check state
-initState :: [KnownExtension] -> ModuleIdent -> TCEnv -> RenameEnv -> SCState
-initState exts m tcEnv rEnv =
-  SCState exts m tcEnv rEnv globalScopeId 1 noFuncDeps False []
+initState :: [KnownExtension] -> ModuleIdent -> TCEnv -> RenameEnv -> ValueEnv
+          -> SCState
+initState exts m tcEnv rEnv vEnv =
+  SCState exts m tcEnv rEnv vEnv globalScopeId 1 noFuncDeps False []
 
 -- |Identifier for global (top-level) declarations
 globalScopeId :: Integer
@@ -154,6 +158,10 @@ getTyConsEnv = S.gets tyConsEnv
 getRenameEnv :: SCM RenameEnv
 getRenameEnv = S.gets renameEnv
 
+-- |Retrieve the 'ValueEnv'
+getValueEnv :: SCM ValueEnv
+getValueEnv = S.gets valueEnv
+
 -- |Modify the 'RenameEnv'
 modifyRenameEnv :: (RenameEnv -> RenameEnv) -> SCM ()
 modifyRenameEnv f = S.modify $ \s -> s { renameEnv = f $ renameEnv s }
@@ -333,13 +341,13 @@ bindTypeDecl (NewtypeDecl _ _ _ nc _) = bindNewConstr nc
 bindTypeDecl _                        = ok
 
 bindConstr :: ConstrDecl -> SCM ()
-bindConstr (ConstrDecl _ _ _ c tys) = do
+bindConstr (ConstrDecl _ c tys) = do
   m <- getModuleIdent
   modifyRenameEnv $ bindGlobal False m c (Constr (qualifyWith m c) $ length tys)
-bindConstr (ConOpDecl _ _ _ _ op _) = do
+bindConstr (ConOpDecl _ _ op _) = do
   m <- getModuleIdent
   modifyRenameEnv $ bindGlobal False m op (Constr (qualifyWith m op) 2)
-bindConstr (RecordDecl _ _ _ c fs)  = do
+bindConstr (RecordDecl _ c fs)  = do
   m <- getModuleIdent
   modifyRenameEnv $ bindGlobal False m c (Constr (qualifyWith m c) (length labels))
     where labels = [l | FieldDecl _ ls _ <- fs, l <- ls]
@@ -361,7 +369,7 @@ bindRecordLabels cs =
 bindRecordLabel :: (Ident, [Ident]) -> SCM ()
 bindRecordLabel (l, cs) = do
   m   <- getModuleIdent
-  new <- (null . lookupVar l) <$> getRenameEnv
+  new <- null . lookupVar l <$> getRenameEnv
   unless new $ report $ errDuplicateDefinition l
   modifyRenameEnv $ bindGlobal False m l $
     RecordLabel (qualifyWith m l) (map (qualifyWith m) cs)
@@ -376,7 +384,7 @@ bindFuncDecl tcc m (FunctionDecl _ _ f (eq:_)) env
   = let arty = length $ snd $ getFlatLhs eq
     in  bindGlobal tcc m f (GlobalVar (qualifyWith m f) arty) env
 bindFuncDecl tcc m (TypeSig _ fs (QualTypeExpr _ _ ty)) env
-  = foldr bindTS env $ map (qualifyWith m) fs
+  = foldr (bindTS . qualifyWith m) env fs
   where
     bindTS qf env'
       | null $ qualLookupVar qf env'
@@ -388,8 +396,8 @@ bindFuncDecl _   _ _ env = env
 
 -- |Bind type class information, i.e. class methods
 bindClassDecl :: Decl a -> SCM ()
-bindClassDecl (ClassDecl _ _ _ _ ds) = mapM_ bindClassMethod ds
-bindClassDecl _ = ok
+bindClassDecl (ClassDecl _ _ _ _ _ ds) = mapM_ bindClassMethod ds
+bindClassDecl _                        = ok
 
 bindClassMethod :: Decl a -> SCM ()
 bindClassMethod ts@(TypeSig _ _ _) = do
@@ -406,8 +414,7 @@ bindVarDecl (FunctionDecl    _ _ f eqs) env
   | otherwise = let arty = length $ snd $ getFlatLhs $ head eqs
                 in  bindLocal (unRenameIdent f) (LocalVar f arty) env
 bindVarDecl (PatternDecl         _ t _) env = foldr bindVar env (bv t)
-bindVarDecl (FreeDecl             _ vs) env =
-  foldr bindVar env (map varIdent vs)
+bindVarDecl (FreeDecl             _ vs) env = foldr (bindVar . varIdent) env vs
 bindVarDecl _                           env = env
 
 bindVar :: Ident -> RenameEnv -> RenameEnv
@@ -447,7 +454,7 @@ qualLookupListCons v env
 -- declaration are allowed to be declared).
 
 checkModule :: Module () -> SCM (Module (), [KnownExtension])
-checkModule (Module spi ps m es is ds) = do
+checkModule (Module spi li ps m es is ds) = do
   mapM_ bindTypeDecl tds
   mapM_ bindClassDecl cds
   ds' <- checkTopDecls ds
@@ -456,7 +463,7 @@ checkModule (Module spi ps m es is ds) = do
   let ds'' = updateClassAndInstanceDecls cds' ids' ds'
   checkFuncPatDeps
   exts <- getExtensions
-  return (Module spi ps m es is ds'', exts)
+  return (Module spi li ps m es is ds'', exts)
   where tds = filter isTypeDecl ds
         cds = filter isClassDecl ds
         ids = filter isInstanceDecl ds
@@ -468,7 +475,7 @@ checkFuncPatDeps = do
   fps  <- getFuncPats
   deps <- getGlobalDeps
   let levels   = scc (:[])
-                     (\k -> Set.toList (Map.findWithDefault (Set.empty) k deps))
+                     (\k -> Set.toList (Map.findWithDefault Set.empty k deps))
                      (Map.keys deps)
       levelMap = Map.fromList [ (f, l) | (fs, l) <- zip levels [1 ..], f <- fs ]
       level f  = Map.findWithDefault (0 :: Int) f levelMap
@@ -485,21 +492,48 @@ checkTopDecls ds = do
   checkDeclGroup (bindFuncDecl tcc m) ds
 
 checkClassDecl :: Decl () -> SCM (Decl ())
-checkClassDecl (ClassDecl p cx cls tv ds) = do
+checkClassDecl (ClassDecl p li cx cls tv ds) = do
   checkMethods (qualify cls) (concatMap methods ds) ds
-  ClassDecl p cx cls tv <$> checkTopDecls ds
+  ClassDecl p li cx cls tv <$> checkTopDecls ds
 checkClassDecl _ =
   internalError "SyntaxCheck.checkClassDecl: no class declaration"
 
 checkInstanceDecl :: Decl () -> SCM (Decl ())
-checkInstanceDecl (InstanceDecl p cx qcls ty ds) = do
+checkInstanceDecl (InstanceDecl p li cx qcls ty ds) = do
   m <- getModuleIdent
+  vEnv <- getValueEnv
   tcEnv <- getTyConsEnv
-  checkMethods qcls (clsMethods m qcls tcEnv) ds
-  InstanceDecl p cx qcls ty <$> checkTopDecls ds
+  let clsMthds = clsMethods m qcls tcEnv
+  let orig = getOrigName m qcls tcEnv
+  let mthds =
+        if isLocalIdent m orig
+          then clsMthds
+          else filter (isFromCls orig m vEnv) clsMthds
+  checkMethods qcls mthds ds
+  mapM_ checkAmbiguousMethod ds
+  InstanceDecl p li cx qcls ty <$> checkTopDecls ds
+  where
+    isFromCls orig m vEnv f = case qualLookupValueUnique m (qualify f) vEnv of
+      [Value _ (Just cls) _ _]
+        | cls == orig -> True
+      _               -> False
+
 checkInstanceDecl _ =
   internalError "SyntaxCheck.checkInstanceDecl: no instance declaration"
 
+checkAmbiguousMethod :: Decl a -> SCM ()
+checkAmbiguousMethod (FunctionDecl _ _ f _) = do
+  m <- getModuleIdent
+  rename <- getRenameEnv
+  case lookupVar f rename of
+    rs1@(_:_:_) -> case qualLookupVar (qualifyWith m f) rename of
+      []          -> report $ errAmbiguousIdent rs1 (qualify f)
+      rs2@(_:_:_) -> report $ errAmbiguousIdent rs2 (qualify f)
+      _           -> return ()
+    _           -> return ()
+checkAmbiguousMethod _ =
+  internalError "SyntaxCheck.checkAmbiguousMethod: no function declaration"
+
 checkMethods :: QualIdent -> [Ident] -> [Decl a] -> SCM ()
 checkMethods qcls ms ds =
   mapM_ (report . errUndefinedMethod qcls) $ filter (`notElem` ms) fs
@@ -507,9 +541,9 @@ checkMethods qcls ms ds =
 
 updateClassAndInstanceDecls :: [Decl a] -> [Decl a] -> [Decl a] -> [Decl a]
 updateClassAndInstanceDecls [] [] ds = ds
-updateClassAndInstanceDecls (c:cs) is (ClassDecl _ _ _ _ _:ds) =
+updateClassAndInstanceDecls (c:cs) is (ClassDecl _ _ _ _ _ _:ds) =
   c : updateClassAndInstanceDecls cs is ds
-updateClassAndInstanceDecls cs (i:is) (InstanceDecl _ _ _ _ _:ds) =
+updateClassAndInstanceDecls cs (i:is) (InstanceDecl _ _ _ _ _ _:ds) =
   i : updateClassAndInstanceDecls cs is ds
 updateClassAndInstanceDecls cs is (d:ds) =
   d : updateClassAndInstanceDecls cs is ds
@@ -555,7 +589,7 @@ checkPrecedence :: SpanInfo -> Maybe Precedence -> SCM (Maybe Precedence)
 checkPrecedence _ Nothing  = return Nothing
 checkPrecedence p (Just i) = do
   unless (0 <= i && i <= 9) $ report
-                            $ errPrecedenceOutOfRange (spanInfo2Pos p) i
+                            $ errPrecedenceOutOfRange p i
   return $ Just i
 
 checkVar' :: String -> Var a -> SCM (Var a)
@@ -589,7 +623,7 @@ checkEqLhs pspi toplhs = do
       | not $ isDataConstr f env -> return left
       | k /= globalScopeId       -> return right
       | null infos               -> return left
-      | otherwise                -> do report $ errToplevelPattern p
+      | otherwise                -> do report $ errToplevelPattern pspi
                                        return right
       where f'    = renameIdent f k
             infos = qualLookupVar (qualifyWith m f) env
@@ -600,7 +634,7 @@ checkEqLhs pspi toplhs = do
       | not $ isDataConstr op env -> return left
       | k /= globalScopeId        -> return right
       | null infos                -> return left
-      | otherwise                 -> do report $ errToplevelPattern p
+      | otherwise                 -> do report $ errToplevelPattern pspi
                                         return right
       where op'   = renameIdent op k
             infos = qualLookupVar (qualifyWith m op) env
@@ -620,7 +654,6 @@ checkEqLhs pspi toplhs = do
         r               -> do report $ errNonVariable "curried definition" f
                               return $ r
         where (f, _) = flatLhs lhs
-  where p = spanInfo2Pos pspi
 
 checkOpLhs :: Integer -> RenameEnv -> (Pattern a -> Pattern a)
            -> Pattern a -> Either (Ident, Lhs a) (Pattern a)
@@ -678,7 +711,7 @@ checkDeclRhs _   (PatternDecl    p t rhs) =
 checkDeclRhs _   d                        = return d
 
 checkDeclLabels :: ConstrDecl -> SCM ConstrDecl
-checkDeclLabels rd@(RecordDecl _ _ _ _ fs) = do
+checkDeclLabels rd@(RecordDecl _ _ fs) = do
   onJust (report . errDuplicateLabel "declaration")
          (findDouble $ map qualify labels)
   return rd
@@ -704,7 +737,7 @@ checkLhs p (FunLhs    spi f ts) = FunLhs spi f <$> mapM (checkPattern p) ts
 checkLhs p (OpLhs spi t1 op t2) = do
   let wrongCalls = concatMap (checkParenPattern (Just $ qualify op)) [t1,t2]
   unless (null wrongCalls) $ report $ errInfixWithoutParens
-    (getPosition op) wrongCalls
+    spi wrongCalls
   flip (OpLhs spi) op <$> checkPattern p t1 <*> checkPattern p t2
 checkLhs p (ApLhs   spi lhs ts) =
   ApLhs spi <$> checkLhs p lhs <*> mapM (checkPattern p) ts
@@ -747,7 +780,7 @@ checkPattern _ (LiteralPattern        spi a l) =
 checkPattern _ (NegativePattern       spi a l) =
   return $ NegativePattern spi a l
 checkPattern p (VariablePattern       spi a v)
-  | isAnonId v = (VariablePattern spi a . renameIdent v) <$> newId
+  | isAnonId v = VariablePattern spi a . renameIdent v <$> newId
   | otherwise  = checkConstructorPattern p spi (qualify v) []
 checkPattern p (ConstructorPattern spi _ c ts) =
   checkConstructorPattern p spi c ts
@@ -767,9 +800,9 @@ checkPattern p (LazyPattern             spi t) = do
   t' <- checkPattern p t
   banFPTerm "lazy pattern" p t'
   return (LazyPattern spi t')
-checkPattern _ (FunctionPattern     _ _ _ _) = internalError $
+checkPattern _ (FunctionPattern     _ _ _ _) = internalError
   "SyntaxCheck.checkPattern: function pattern not defined"
-checkPattern _ (InfixFuncPattern  _ _ _ _ _) = internalError $
+checkPattern _ (InfixFuncPattern  _ _ _ _ _) = internalError
   "SyntaxCheck.checkPattern: infix function pattern not defined"
 
 checkConstructorPattern :: SpanInfo -> SpanInfo -> QualIdent -> [Pattern ()]
@@ -803,16 +836,11 @@ checkConstructorPattern p spi c ts = do
     | null ts && not (isQualified c)
     = return $ VariablePattern spi () $ renameIdent (unqualify c) k -- (varIdent r) k
     | otherwise = do
-      let n = arity r
-      checkFuncPatsExtension (spanInfo2Pos p)
+      checkFuncPatsExtension p
       checkFuncPatCall r c
       ts' <- mapM (checkPattern p) ts
       mapM_ (checkFPTerm p) ts'
-      return $ if n' > n
-                 then let (ts1, ts2) = splitAt n ts'
-                      in  genFuncPattAppl
-                          (FunctionPattern spi () (qualVarIdent r) ts1) ts2
-                 else FunctionPattern spi () (qualVarIdent r) ts'
+      return $ FunctionPattern spi () (qualVarIdent r) ts'
 
 checkInfixPattern :: SpanInfo -> SpanInfo -> Pattern () -> QualIdent -> Pattern ()
                   -> SCM (Pattern ())
@@ -835,7 +863,7 @@ checkInfixPattern p spi t1 op t2 = do
     when (n /= 2) $ report $ errWrongArity op n 2
     flip (InfixPattern spi ()) qop <$> checkPattern p t1 <*> checkPattern p t2
   funcPattern r qop = do
-    checkFuncPatsExtension (spanInfo2Pos p)
+    checkFuncPatsExtension p
     checkFuncPatCall r qop
     ts' <- mapM (checkPattern p) [t1,t2]
     let [t1',t2'] = ts'
@@ -871,10 +899,14 @@ checkFuncPatCall r f = case r of
 
 -- Note: process decls first
 checkRhs :: Rhs () -> SCM (Rhs ())
-checkRhs (SimpleRhs spi e ds) = inNestedScope $
-  flip (SimpleRhs spi) <$> checkDeclGroup bindVarDecl ds <*> checkExpr spi e
-checkRhs (GuardedRhs spi es ds) = inNestedScope $
-  flip (GuardedRhs spi) <$> checkDeclGroup bindVarDecl ds <*> mapM checkCondExpr es
+checkRhs (SimpleRhs spi li e ds) = inNestedScope $
+  flip (SimpleRhs spi li) <$>
+    checkDeclGroup bindVarDecl ds <*>
+    checkExpr spi e
+checkRhs (GuardedRhs spi li es ds) = inNestedScope $
+  flip (GuardedRhs spi li) <$>
+    checkDeclGroup bindVarDecl ds <*>
+    mapM checkCondExpr es
 
 checkCondExpr :: CondExpr () -> SCM (CondExpr ())
 checkCondExpr (CondExpr spi g e) =  CondExpr spi <$> checkExpr spi g <*> checkExpr spi e
@@ -909,14 +941,14 @@ checkExpr p (LeftSection        spi e op) =
 checkExpr p (RightSection       spi op e) =
   RightSection spi <$> checkOp op <*> checkExpr p e
 checkExpr p (Lambda             spi ts e) = inNestedScope $ checkLambda p spi ts e
-checkExpr p (Let                spi ds e) = inNestedScope $
-  Let spi <$> checkDeclGroup bindVarDecl ds <*> checkExpr p e
-checkExpr p (Do                spi sts e) = withLocalEnv $
-  Do spi <$> mapM (checkStatement "do sequence" p) sts <*> checkExpr p e
+checkExpr p (Let             spi li ds e) = inNestedScope $
+  Let spi li <$> checkDeclGroup bindVarDecl ds <*> checkExpr p e
+checkExpr p (Do             spi li sts e) = withLocalEnv $
+  Do spi li <$> mapM (checkStatement "do sequence" p) sts <*> checkExpr p e
 checkExpr p (IfThenElse     spi e1 e2 e3) =
   IfThenElse spi <$> checkExpr p e1 <*> checkExpr p e2 <*> checkExpr p e3
-checkExpr p (Case          spi ct e alts) =
-  Case spi ct <$> checkExpr p e <*> mapM checkAlt alts
+checkExpr p (Case       spi li ct e alts) =
+  Case spi li ct <$> checkExpr p e <*> mapM checkAlt alts
 
 checkLambda :: SpanInfo -> SpanInfo -> [Pattern ()] -> Expression ()
             -> SCM (Expression ())
@@ -936,8 +968,8 @@ checkVariable :: SpanInfo -> a -> QualIdent -> SCM (Expression a)
 checkVariable spi a v
     -- anonymous free variable
   | isAnonId (unqualify v) = do
-    checkAnonFreeVarsExtension $ getPosition v
-    (\n -> Variable spi a $ updQualIdent id (flip renameIdent n) v) <$> newId
+    checkAnonFreeVarsExtension $ getSpanInfo v
+    (\n -> Variable spi a $ updQualIdent id (`renameIdent` n) v) <$> newId
     -- return $ Variable v
     -- normal variable
   | otherwise             = do
@@ -947,7 +979,9 @@ checkVariable spi a v
                             return $ Variable spi a v
       [Constr    _ _]   -> return $ Constructor spi a v
       [GlobalVar f _]   -> addGlobalDep f >> return (Variable spi a v)
-      [LocalVar v' _]   -> return $ Variable spi a $ qualify v' @> v
+      [LocalVar v' _]   -> return $ Variable spi a
+                                  $ qualify
+                                  $ spanInfoLike v' (qidIdent v)
       [RecordLabel _ _] -> return $ Variable spi a v
       rs -> do
         m <- getModuleIdent
@@ -956,7 +990,9 @@ checkVariable spi a v
                                 return $ Variable spi a v
           [Constr    _ _]   -> return $ Constructor spi a v
           [GlobalVar f _]   -> addGlobalDep f >> return (Variable spi a v)
-          [LocalVar v' _]   -> return $ Variable spi a $ qualify v' @> v
+          [LocalVar v' _]   -> return $ Variable spi a
+                                      $ qualify
+                                      $ spanInfoLike v' (qidIdent v)
           [RecordLabel _ _] -> return $ Variable spi a v
           rs'               -> do report $ errAmbiguousIdent rs' v
                                   return $ Variable spi a v
@@ -995,11 +1031,11 @@ checkRecordUpdExpr p spi e fs = do
 -- * Because statements are processed list-wise, inNestedEnv can not be
 --   used as this nesting must be visible to following statements.
 checkStatement :: String -> SpanInfo -> Statement () -> SCM (Statement ())
-checkStatement _ p (StmtExpr spi   e) = StmtExpr spi <$> checkExpr p e
-checkStatement s p (StmtBind spi t e) =
+checkStatement _ p (StmtExpr spi     e) = StmtExpr spi <$> checkExpr p e
+checkStatement s p (StmtBind spi   t e) =
   flip (StmtBind spi) <$> checkExpr p e <*> (incNesting >> bindPattern s p t)
-checkStatement _ _ (StmtDecl spi  ds) =
-  StmtDecl spi <$> (incNesting >> checkDeclGroup bindVarDecl ds)
+checkStatement _ _ (StmtDecl spi li ds) =
+  StmtDecl spi li <$> (incNesting >> checkDeclGroup bindVarDecl ds)
 
 bindPattern :: String -> SpanInfo -> Pattern () -> SCM (Pattern ())
 bindPattern s p t = do
@@ -1021,9 +1057,9 @@ banFPTerm s p (ListPattern             _ _ ts) = mapM_ (banFPTerm s p) ts
 banFPTerm s p (AsPattern                _ _ t) = banFPTerm s p t
 banFPTerm s p (LazyPattern                _ t) = banFPTerm s p t
 banFPTerm s p pat@(FunctionPattern    _ _ _ _)
- = report $ errUnsupportedFuncPattern s (spanInfo2Pos p) pat
+ = report $ errUnsupportedFuncPattern s p pat
 banFPTerm s p pat@(InfixFuncPattern _ _ _ _ _)
- = report $ errUnsupportedFuncPattern s (spanInfo2Pos p) pat
+ = report $ errUnsupportedFuncPattern s p pat
 
 checkOp :: InfixOp a -> SCM (InfixOp a)
 checkOp op = do
@@ -1101,9 +1137,10 @@ checkLabels _ (Just c) ls css = do
                            [l | (l, cs) <- zip ls css, c' `notElem` cs]
     _             -> internalError $
                        "Checks.SyntaxCheck.checkLabels: " ++ show c
-checkLabels p Nothing ls css =
-  when (null (foldr1 intersect css))
-    $ report $ errNoCommonCons (spanInfo2Pos p) ls
+checkLabels p Nothing ls css
+  | not (null (foldr1 intersect css)) ||
+    any null css = ok
+  | otherwise    = report $ errNoCommonCons p ls
 
 checkField :: (a -> SCM a) -> Field a -> SCM (Field a)
 checkField check (Field p l x) = Field p l <$> check x
@@ -1134,7 +1171,7 @@ recLabels _                        = []
 -- it is necessary to sort the list of declarations.
 
 sortFuncDecls :: [Decl a] -> [Decl a]
-sortFuncDecls decls = sortFD Set.empty [] decls
+sortFuncDecls = sortFD Set.empty []
  where
  sortFD _   res []              = reverse res
  sortFD env res (decl : decls') = case decl of
@@ -1191,23 +1228,6 @@ qualVarIdent (GlobalVar v _) = v
 qualVarIdent (LocalVar  v _) = qualify v
 qualVarIdent _ = internalError "SyntaxCheck.qualVarIdent: no variable"
 
-arity :: RenameInfo -> Int
-arity (Constr      _ n) = n
-arity (GlobalVar   _ n) = n
-arity (LocalVar    _ n) = n
-arity (RecordLabel _ _) = 1
-
--- Unlike expressions, constructor terms have no possibility to represent
--- over-applications in functional patterns. Therefore it is necessary to
--- transform them to nested function patterns using the prelude function
--- apply. E.g., the function pattern (id id 10) is transformed to
--- (apply (id id) 10).
-
-genFuncPattAppl :: Pattern () -> [Pattern ()] -> Pattern ()
-genFuncPattAppl term []     = term
-genFuncPattAppl term (t:ts)
-   = FunctionPattern NoSpanInfo () qApplyId [genFuncPattAppl term ts, t] -- TODO FIXME major problem
-
 checkFPTerm :: SpanInfo -> Pattern a -> SCM ()
 checkFPTerm _ (LiteralPattern        _ _ _) = ok
 checkFPTerm _ (NegativePattern       _ _ _) = ok
@@ -1219,7 +1239,7 @@ checkFPTerm p (TuplePattern           _ ts) = mapM_ (checkFPTerm p) ts
 checkFPTerm p (ListPattern          _ _ ts) = mapM_ (checkFPTerm p) ts
 checkFPTerm p (AsPattern             _ _ t) = checkFPTerm p t
 checkFPTerm p t@(LazyPattern           _ _) =
-  report $ errUnsupportedFPTerm "Lazy" (spanInfo2Pos p) t
+  report $ errUnsupportedFPTerm "Lazy" p t
 checkFPTerm p (RecordPattern      _ _ _ fs) = mapM_ (checkFPTerm p)
                                             [ t | Field _ _ t <- fs ]
 checkFPTerm _ (FunctionPattern     _ _ _ _) = ok -- do not check again
@@ -1229,19 +1249,19 @@ checkFPTerm _ (InfixFuncPattern  _ _ _ _ _) = ok -- do not check again
 -- Miscellaneous functions
 -- ---------------------------------------------------------------------------
 
-checkFuncPatsExtension :: Position -> SCM ()
-checkFuncPatsExtension p = checkUsedExtension p
+checkFuncPatsExtension :: SpanInfo -> SCM ()
+checkFuncPatsExtension spi = checkUsedExtension spi
   "Functional Patterns" FunctionalPatterns
 
-checkAnonFreeVarsExtension :: Position -> SCM ()
-checkAnonFreeVarsExtension p = checkUsedExtension p
+checkAnonFreeVarsExtension :: SpanInfo -> SCM ()
+checkAnonFreeVarsExtension spi = checkUsedExtension spi
   "Anonymous free variables" AnonFreeVars
 
-checkUsedExtension :: Position -> String -> KnownExtension -> SCM ()
-checkUsedExtension pos msg ext = do
+checkUsedExtension :: SpanInfo -> String -> KnownExtension -> SCM ()
+checkUsedExtension spi msg ext = do
   enabled <- hasExtension ext
   unless enabled $ do
-    report $ errMissingLanguageExtension pos msg ext
+    report $ errMissingLanguageExtension spi msg ext
     enableExtension ext -- to avoid multiple warnings
 
 typeArity :: TypeExpr -> Int
@@ -1259,43 +1279,43 @@ opAnnotation (InfixConstr a _) = a
 -- Error messages
 -- ---------------------------------------------------------------------------
 
-errUnsupportedFPTerm :: String -> Position -> Pattern a -> Message
-errUnsupportedFPTerm s p pat = posMessage p $ text s
+errUnsupportedFPTerm :: String -> SpanInfo -> Pattern a -> Message
+errUnsupportedFPTerm s spi pat = spanInfoMessage spi $ text s
   <+> text "patterns are not supported inside a functional pattern."
-  $+$ ppPattern 0 pat
+  $+$ pPrintPrec 0 pat
 
-errUnsupportedFuncPattern :: String -> Position -> Pattern a -> Message
-errUnsupportedFuncPattern s p pat = posMessage p $
+errUnsupportedFuncPattern :: String -> SpanInfo -> Pattern a -> Message
+errUnsupportedFuncPattern s spi pat = spanInfoMessage spi $
   text "Functional patterns are not supported inside a" <+> text s <> dot
-  $+$ ppPattern 0 pat
+  $+$ pPrintPrec 0 pat
 
 errFuncPatNotGlobal :: QualIdent -> Message
-errFuncPatNotGlobal f = posMessage f $ hsep $ map text
+errFuncPatNotGlobal f = spanInfoMessage f $ hsep $ map text
   ["Function", escQualName f, "in functional pattern is not global"]
 
 errFuncPatCyclic :: QualIdent -> QualIdent -> Message
-errFuncPatCyclic fp f = posMessage fp $ hsep $ map text
+errFuncPatCyclic fp f = spanInfoMessage fp $ hsep $ map text
   [ "Function", escName $ unqualify fp, "used in functional pattern depends on"
   , escName $ unqualify f, " causing a cyclic dependency"]
 
-errPrecedenceOutOfRange :: Position -> Integer -> Message
-errPrecedenceOutOfRange p i = posMessage p $ hsep $ map text
+errPrecedenceOutOfRange :: SpanInfo -> Integer -> Message
+errPrecedenceOutOfRange spi i = spanInfoMessage spi $ hsep $ map text
   ["Precedence out of range:", show i]
 
 errUndefinedVariable :: QualIdent -> Message
-errUndefinedVariable v = posMessage v $ hsep $ map text
+errUndefinedVariable v = spanInfoMessage v $ hsep $ map text
   [escQualName v, "is undefined"]
 
 errUndefinedData :: QualIdent -> Message
-errUndefinedData c = posMessage c $ hsep $ map text
+errUndefinedData c = spanInfoMessage c $ hsep $ map text
   ["Undefined data constructor", escQualName c]
 
 errUndefinedLabel :: QualIdent -> Message
-errUndefinedLabel l = posMessage l $  hsep $ map text
+errUndefinedLabel l = spanInfoMessage l $  hsep $ map text
   ["Undefined record label", escQualName l]
 
 errUndefinedMethod :: QualIdent -> Ident -> Message
-errUndefinedMethod qcls f = posMessage f $ hsep $ map text
+errUndefinedMethod qcls f = spanInfoMessage f $ hsep $ map text
   [escName f, "is not a (visible) method of class", escQualName qcls]
 
 errAmbiguousIdent :: [RenameInfo] -> QualIdent -> Message
@@ -1310,26 +1330,26 @@ errAmbiguousLabel :: [RenameInfo] -> QualIdent -> Message
 errAmbiguousLabel = errAmbiguous "field label"
 
 errAmbiguous :: String -> [RenameInfo] -> QualIdent -> Message
-errAmbiguous what rs qn = posMessage qn
+errAmbiguous what rs qn = spanInfoMessage qn
   $   text "Ambiguous" <+> text what <+> text (escQualName qn)
   $+$ text "It could refer to:"
   $+$ nest 2 (vcat (map ppRenameInfo rs))
 
 errDuplicateDefinition :: Ident -> Message
-errDuplicateDefinition v = posMessage v $ hsep $ map text
+errDuplicateDefinition v = spanInfoMessage v $ hsep $ map text
   ["More than one definition for", escName v]
 
 errDuplicateVariables :: [Ident] -> Message
 errDuplicateVariables [] = internalError
   "SyntaxCheck.errDuplicateVariables: empty list"
-errDuplicateVariables (v:vs) = posMessage v $
+errDuplicateVariables (v:vs) = spanInfoMessage v $
   text (escName v) <+> text "occurs more than one in pattern at:" $+$
   nest 2 (vcat (map (ppPosition . getPosition) (v:vs)))
 
 errMultipleDataConstructor :: [Ident] -> Message
 errMultipleDataConstructor [] = internalError
   "SyntaxCheck.errMultipleDataDeclaration: empty list"
-errMultipleDataConstructor (i:is) = posMessage i $
+errMultipleDataConstructor (i:is) = spanInfoMessage i $
   text "Multiple definitions for data/record constructor" <+> text (escName i)
   <+> text "at:" $+$
   nest 2 (vcat (map (ppPosition . getPosition) (i:is)))
@@ -1337,7 +1357,7 @@ errMultipleDataConstructor (i:is) = posMessage i $
 errMultipleDeclarations :: ModuleIdent -> [Ident] -> Message
 errMultipleDeclarations _ [] = internalError
   "SyntaxCheck.errMultipleDeclarations: empty list"
-errMultipleDeclarations m (i:is) = posMessage i $
+errMultipleDeclarations m (i:is) = spanInfoMessage i $
   text "Multiple declarations of" <+> text (escQualName (qualifyWith m i))
   $+$ text "Declared at:" $+$
   nest 2 (vcat (map (ppPosition . getPosition) (i:is)))
@@ -1345,63 +1365,63 @@ errMultipleDeclarations m (i:is) = posMessage i $
 errDuplicateTypeSig :: [Ident] -> Message
 errDuplicateTypeSig [] = internalError
   "SyntaxCheck.errDuplicateTypeSig: empty list"
-errDuplicateTypeSig (v:vs) = posMessage v $
+errDuplicateTypeSig (v:vs) = spanInfoMessage v $
   text "More than one type signature for" <+> text (escName v)
   <+> text "at:" $+$
   nest 2 (vcat (map (ppPosition . getPosition) (v:vs)))
 
 errDuplicateLabel :: String -> QualIdent -> Message
-errDuplicateLabel what l = posMessage l $ hsep $ map text
+errDuplicateLabel what l = spanInfoMessage l $ hsep $ map text
   ["Field label", escQualName l, "occurs more than once in record", what]
 
 errNonVariable :: String -> Ident -> Message
-errNonVariable what c = posMessage c $ hsep $ map text
+errNonVariable what c = spanInfoMessage c $ hsep $ map text
   ["Data constructor", escName c, "in left hand side of", what]
 
 errNoBody :: Ident -> Message
-errNoBody v = posMessage v $  hsep $ map text ["No body for", escName v]
+errNoBody v = spanInfoMessage v $  hsep $ map text ["No body for", escName v]
 
-errNoCommonCons :: Position -> [QualIdent] -> Message
-errNoCommonCons p ls = posMessage p $
+errNoCommonCons :: SpanInfo -> [QualIdent] -> Message
+errNoCommonCons spi ls = spanInfoMessage spi $
   text "No constructor has all of these fields:"
   $+$ nest 2 (vcat (map (text . escQualName) ls))
 
 errNoLabel :: QualIdent -> QualIdent -> Message
-errNoLabel c l = posMessage l $ hsep $ map text
+errNoLabel c l = spanInfoMessage l $ hsep $ map text
   [escQualName l, "is not a field label of constructor", escQualName c]
 
 errNoTypeSig :: Ident -> Message
-errNoTypeSig f = posMessage f $ hsep $ map text
+errNoTypeSig f = spanInfoMessage f $ hsep $ map text
   ["No type signature for external function", escName f]
 
-errToplevelPattern :: Position -> Message
-errToplevelPattern p = posMessage p $ text
+errToplevelPattern :: SpanInfo -> Message
+errToplevelPattern spi = spanInfoMessage spi $ text
   "Pattern declaration not allowed at top-level"
 
 errDifferentArity :: [Ident] -> Message
 errDifferentArity [] = internalError
   "SyntaxCheck.errDifferentArity: empty list"
-errDifferentArity (i:is) = posMessage i $
+errDifferentArity (i:is) = spanInfoMessage i $
   text "Equations for" <+> text (escName i) <+> text "have different arities"
   <+> text "at:" $+$
   nest 2 (vcat (map (ppPosition . getPosition) (i:is)))
 
 errWrongArity :: QualIdent -> Int -> Int -> Message
-errWrongArity c arity' argc = posMessage c $ hsep (map text
+errWrongArity c arity' argc = spanInfoMessage c $ hsep (map text
   ["Data constructor", escQualName c, "expects", arguments arity'])
   <> comma <+> text "but is applied to" <+> text (show argc)
   where arguments 0 = "no arguments"
         arguments 1 = "1 argument"
         arguments n = show n ++ " arguments"
 
-errMissingLanguageExtension :: Position -> String -> KnownExtension -> Message
-errMissingLanguageExtension p what ext = posMessage p $
+errMissingLanguageExtension :: SpanInfo -> String -> KnownExtension -> Message
+errMissingLanguageExtension spi what ext = spanInfoMessage spi $
   text what <+> text "are not supported in standard Curry." $+$
   nest 2 (text "Use flag or -X" <+> text (show ext)
           <+> text "to enable this extension.")
 
-errInfixWithoutParens :: Position -> [(QualIdent, QualIdent)] -> Message
-errInfixWithoutParens p calls = posMessage p $
+errInfixWithoutParens :: SpanInfo -> [(QualIdent, QualIdent)] -> Message
+errInfixWithoutParens spi calls = spanInfoMessage spi $
   text "Missing parens in infix patterns:" $+$
   vcat (map showCall calls)
   where
