@@ -21,28 +21,34 @@ module Checks.WarnCheck (warnCheck) where
 import Prelude hiding ((<>))
 #endif
 
+import           Control.Applicative
+  ((<|>))
 import           Control.Monad
-  (filterM, foldM_, guard, liftM, liftM2, when, unless)
+  (filterM, foldM_, guard, liftM, liftM2, when, unless, void)
 import           Control.Monad.State.Strict    (State, execState, gets, modify)
 import qualified Data.IntSet         as IntSet
   (IntSet, empty, insert, notMember, singleton, union, unions)
-import qualified Data.Map            as Map    (empty, insert, lookup)
+import qualified Data.Map            as Map    (empty, insert, lookup, (!))
 import           Data.Maybe
   (catMaybes, fromMaybe, listToMaybe)
 import           Data.List
   ((\\), intersect, intersectBy, nub, sort, unionBy)
 import           Data.Char
   (isLower, isUpper, toLower, toUpper, isAlpha)
+import qualified Data.Set.Extra as Set
+import           Data.Tuple.Extra
+  (snd3)
 
 import Curry.Base.Ident
 import Curry.Base.Position
 import Curry.Base.Pretty
 import Curry.Base.SpanInfo
 import Curry.Syntax
-import Curry.Syntax.Pretty (ppDecl, ppPattern, ppExpr, ppIdent)
+import Curry.Syntax.Utils  (typeVariables)
+import Curry.Syntax.Pretty (pPrint)
 
-import Base.CurryTypes (ppTypeScheme)
-import Base.Messages   (Message, posMessage, internalError)
+import Base.CurryTypes (ppTypeScheme, fromPred, toPredSet)
+import Base.Messages   (Message, spanInfoMessage, internalError)
 import Base.NestEnv    ( NestEnv, emptyEnv, localNestEnv, nestEnv, unnestEnv
                        , qualBindNestEnv, qualInLocalNestEnv, qualLookupNestEnv
                        , qualModifyNestEnv)
@@ -66,6 +72,7 @@ import CompilerOpts
 --   - overlapping case alternatives
 --   - non-adjacent function rules
 --   - wrong case mode
+--   - redundant context
 warnCheck :: WarnOpts -> CaseMode -> AliasEnv -> ValueEnv -> TCEnv -> ClassEnv
           -> Module a -> [Message]
 warnCheck wOpts cOpts aEnv valEnv tcEnv clsEnv mdl
@@ -76,7 +83,8 @@ warnCheck wOpts cOpts aEnv valEnv tcEnv clsEnv mdl
       checkMissingTypeSignatures ds
       checkModuleAlias is
       checkCaseMode  ds
-  where Module _ _ mid es is ds = fmap (const ()) mdl
+      checkRedContext ds
+  where Module _ _ _ mid es is ds = fmap (const ()) mdl
 
 type ScopeEnv = NestEnv IdInfo
 
@@ -196,16 +204,16 @@ checkImports = warnFor WarnMultipleImports . foldM_ checkImport Map.empty
   impName (ImportTypeWith _ t _) = t
 
 warnMultiplyImportedModule :: ModuleIdent -> Message
-warnMultiplyImportedModule mid = posMessage mid $ hsep $ map text
+warnMultiplyImportedModule mid = spanInfoMessage mid $ hsep $ map text
   ["Module", moduleName mid, "is imported more than once"]
 
 warnMultiplyImportedSymbol :: ModuleIdent -> Ident -> Message
-warnMultiplyImportedSymbol mid ident = posMessage ident $ hsep $ map text
+warnMultiplyImportedSymbol mid ident = spanInfoMessage ident $ hsep $ map text
   [ "Symbol", escName ident, "from module", moduleName mid
   , "is imported more than once" ]
 
 warnMultiplyHiddenSymbol :: ModuleIdent -> Ident -> Message
-warnMultiplyHiddenSymbol mid ident = posMessage ident $ hsep $ map text
+warnMultiplyHiddenSymbol mid ident = spanInfoMessage ident $ hsep $ map text
   [ "Symbol", escName ident, "from module", moduleName mid
   , "is hidden more than once" ]
 
@@ -244,28 +252,28 @@ checkRuleAdjacency decls = warnFor WarnDisjoinedRules
   check (_    , env) _                     = return (mkIdent "", env)
 
 warnDisjoinedFunctionRules :: Ident -> Position -> Message
-warnDisjoinedFunctionRules ident pos = posMessage ident $ hsep (map text
+warnDisjoinedFunctionRules ident pos = spanInfoMessage ident $ hsep (map text
   [ "Rules for function", escName ident, "are disjoined" ])
   <+> parens (text "first occurrence at" <+> text (showLine pos))
 
 checkDecl :: Decl () -> WCM ()
-checkDecl (DataDecl        _ _ vs cs _) = inNestedScope $ do
+checkDecl (DataDecl          _ _ vs cs _) = inNestedScope $ do
   mapM_ insertTypeVar   vs
   mapM_ checkConstrDecl cs
   reportUnusedTypeVars  vs
-checkDecl (NewtypeDecl     _ _ vs nc _) = inNestedScope $ do
+checkDecl (NewtypeDecl       _ _ vs nc _) = inNestedScope $ do
   mapM_ insertTypeVar   vs
   checkNewConstrDecl nc
   reportUnusedTypeVars vs
-checkDecl (TypeDecl          _ _ vs ty) = inNestedScope $ do
+checkDecl (TypeDecl            _ _ vs ty) = inNestedScope $ do
   mapM_ insertTypeVar  vs
   checkTypeExpr ty
   reportUnusedTypeVars vs
-checkDecl (FunctionDecl      p _ f eqs) = checkFunctionDecl p f eqs
-checkDecl (PatternDecl         _ p rhs) = checkPattern p >> checkRhs rhs
-checkDecl (DefaultDecl           _ tys) = mapM_ checkTypeExpr tys
-checkDecl (ClassDecl        _ _ _ _ ds) = mapM_ checkDecl ds
-checkDecl (InstanceDecl p cx cls ty ds) = do
+checkDecl (FunctionDecl        p _ f eqs) = checkFunctionDecl p f eqs
+checkDecl (PatternDecl           _ p rhs) = checkPattern p >> checkRhs rhs
+checkDecl (DefaultDecl             _ tys) = mapM_ checkTypeExpr tys
+checkDecl (ClassDecl        _ _ _ _ _ ds) = mapM_ checkDecl ds
+checkDecl (InstanceDecl p _ cx cls ty ds) = do
   checkOrphanInstance p cx cls ty
   checkMissingMethodImplementations p cls ds
   mapM_ checkDecl ds
@@ -273,24 +281,15 @@ checkDecl _                             = ok
 
 --TODO: shadowing und context etc.
 checkConstrDecl :: ConstrDecl -> WCM ()
-checkConstrDecl (ConstrDecl     _ vs _ c tys) = inNestedScope $ do
-  mapM_ checkTypeShadowing vs
-  mapM_ insertTypeVar vs
+checkConstrDecl (ConstrDecl     _ c tys) = inNestedScope $ do
   visitId c
   mapM_ checkTypeExpr tys
-  reportUnusedTypeVars vs
-checkConstrDecl (ConOpDecl _ vs _ ty1 op ty2) = inNestedScope $ do
-  mapM_ checkTypeShadowing vs
-  mapM_ insertTypeVar vs
+checkConstrDecl (ConOpDecl _ ty1 op ty2) = inNestedScope $ do
   visitId op
   mapM_ checkTypeExpr [ty1, ty2]
-  reportUnusedTypeVars vs
-checkConstrDecl (RecordDecl      _ vs _ c fs) = inNestedScope $ do
-  mapM_ checkTypeShadowing vs
-  mapM_ insertTypeVar vs
+checkConstrDecl (RecordDecl      _ c fs) = inNestedScope $ do
   visitId c
   mapM_ checkTypeExpr tys
-  reportUnusedTypeVars vs
   where
     tys = [ty | FieldDecl _ _ ty <- fs]
 
@@ -331,12 +330,15 @@ checkFunctionDecl p f eqs = inNestedScope $ do
 checkFunctionPatternMatch :: SpanInfo -> Ident -> [Equation ()] -> WCM ()
 checkFunctionPatternMatch spi f eqs = do
   let pats = map (\(Equation _ lhs _) -> snd (flatLhs lhs)) eqs
-  (nonExhaustive, overlapped, nondet) <- checkPatternMatching pats
+  let guards = map eq2Guards eqs
+  (nonExhaustive, overlapped, nondet) <- checkPatternMatching pats guards
   unless (null nonExhaustive) $ warnFor WarnIncompletePatterns $ report $
-    warnMissingPattern p ("an equation for " ++ escName f) nonExhaustive
+    warnMissingPattern spi ("an equation for " ++ escName f) nonExhaustive
   when (nondet || not (null overlapped)) $ warnFor WarnOverlapping $ report $
-    warnNondetOverlapping p ("Function " ++ escName f)
-  where p = spanInfo2Pos spi
+    warnNondetOverlapping spi ("Function " ++ escName f)
+  where eq2Guards :: Equation () -> [CondExpr ()]
+        eq2Guards (Equation _ _ (GuardedRhs _ _ conds _)) = conds
+        eq2Guards _ = []
 
 -- Check an equation for warnings.
 -- This is done in a seperate scope as the left-hand-side may introduce
@@ -377,11 +379,11 @@ checkPattern _                            = ok
 -- Because local declarations may introduce new variables, we need
 -- another scope nesting.
 checkRhs :: Rhs () -> WCM ()
-checkRhs (SimpleRhs _ e ds) = inNestedScope $ do
+checkRhs (SimpleRhs _ _ e ds) = inNestedScope $ do
   checkLocalDeclGroup ds
   checkExpr e
   reportUnusedVars
-checkRhs (GuardedRhs _ ce ds) = inNestedScope $ do
+checkRhs (GuardedRhs _ _ ce ds) = inNestedScope $ do
   checkLocalDeclGroup ds
   mapM_ checkCondExpr ce
   reportUnusedVars
@@ -416,16 +418,16 @@ checkExpr (Lambda             _ ps e) = inNestedScope $ do
   mapM_ (insertPattern False) ps
   checkExpr e
   reportUnusedVars
-checkExpr (Let                _ ds e) = inNestedScope $ do
+checkExpr (Let              _ _ ds e) = inNestedScope $ do
   checkLocalDeclGroup ds
   checkExpr e
   reportUnusedVars
-checkExpr (Do                _ sts e) = checkStatements sts e
+checkExpr (Do              _ _ sts e) = checkStatements sts e
 checkExpr (IfThenElse     _ e1 e2 e3) = mapM_ checkExpr [e1, e2, e3]
-checkExpr (Case          _ ct e alts) = do
+checkExpr (Case      spi _ ct e alts) = do
   checkExpr e
   mapM_ checkAlt alts
-  checkCaseAlts ct alts
+  checkCaseAlts spi ct alts
 checkExpr _                       = ok
 
 checkStatements :: [Statement ()] -> Expression () -> WCM ()
@@ -435,9 +437,9 @@ checkStatements (s:ss) e = inNestedScope $ do
   reportUnusedVars
 
 checkStatement :: Statement () -> WCM ()
-checkStatement (StmtExpr   _ e) = checkExpr e
-checkStatement (StmtDecl  _ ds) = checkLocalDeclGroup ds
-checkStatement (StmtBind _ p e) = do
+checkStatement (StmtExpr    _ e) = checkExpr e
+checkStatement (StmtDecl _ _ ds) = checkLocalDeclGroup ds
+checkStatement (StmtBind _  p e) = do
   checkPattern p >> insertPattern False p
   checkExpr e
 
@@ -461,11 +463,12 @@ checkOrphanInstance p cx cls ty = warnFor WarnOrphanInstances $ do
   let ocls = getOrigName m cls tcEnv
       otc  = getOrigName m tc  tcEnv
   unless (isLocalIdent m ocls || isLocalIdent m otc) $ report $
-    warnOrphanInstance (spanInfo2Pos p) $ ppDecl $ InstanceDecl p cx cls ty []
+    warnOrphanInstance p $ pPrint $
+    InstanceDecl p WhitespaceLayout cx cls ty []
   where tc = typeConstr ty
 
-warnOrphanInstance :: Position -> Doc -> Message
-warnOrphanInstance p doc = posMessage p $ text "Orphan instance:" <+> doc
+warnOrphanInstance :: SpanInfo -> Doc -> Message
+warnOrphanInstance spi doc = spanInfoMessage spi $ text "Orphan instance:" <+> doc
 
 -- -----------------------------------------------------------------------------
 -- Check for missing method implementations
@@ -478,12 +481,12 @@ checkMissingMethodImplementations p cls ds = warnFor WarnMissingMethods $ do
   clsEnv <- gets classEnv
   let ocls = getOrigName m cls tcEnv
       ms   = classMethods ocls clsEnv
-  mapM_ (report . warnMissingMethodImplementation (spanInfo2Pos p)) $
+  mapM_ (report . warnMissingMethodImplementation p) $
     filter ((null fs ||) . not . flip (hasDefaultImpl ocls) clsEnv) $ ms \\ fs
   where fs = map unRenameIdent $ concatMap impls ds
 
-warnMissingMethodImplementation :: Position -> Ident -> Message
-warnMissingMethodImplementation p f = posMessage p $ hsep $ map text
+warnMissingMethodImplementation :: SpanInfo -> Ident -> Message
+warnMissingMethodImplementation spi f = spanInfoMessage spi $ hsep $ map text
   ["No explicit implementation for method", escName f]
 
 -- -----------------------------------------------------------------------------
@@ -511,7 +514,7 @@ getTyScheme q = do
     _ -> internalError $ "Checks.WarnCheck.getTyScheme: " ++ show q
 
 warnMissingTypeSignature :: ModuleIdent -> Ident -> TypeScheme -> Message
-warnMissingTypeSignature mid i tys = posMessage i $ fsep
+warnMissingTypeSignature mid i tys = spanInfoMessage i $ fsep
   [ text "Top-level binding with no type signature:"
   , nest 2 $ text (showIdent i) <+> text "::" <+> ppTypeScheme mid tys
   ]
@@ -533,14 +536,14 @@ checkModuleAlias is = do
   unless (null aliasClash) $ mapM_ (report . warnAliasNameClash ) aliasClash
 
 warnModuleNameClash :: ModuleIdent -> Message
-warnModuleNameClash mid = posMessage mid $ hsep $ map text
+warnModuleNameClash mid = spanInfoMessage mid $ hsep $ map text
   ["The module alias", escModuleName mid
   , "overlaps with the current module name"]
 
 warnAliasNameClash :: [ModuleIdent] -> Message
 warnAliasNameClash []         = internalError
   "WarnCheck.warnAliasNameClash: empty list"
-warnAliasNameClash mids = posMessage (head mids) $ text
+warnAliasNameClash mids = spanInfoMessage (head mids) $ text
   "Overlapping module aliases" $+$ nest 2 (vcat (map myppAlias mids))
   where myppAlias mid =
           ppLine (getPosition mid) <> text ":" <+> text (escModuleName mid)
@@ -549,23 +552,27 @@ warnAliasNameClash mids = posMessage (head mids) $ text
 -- Check for overlapping/unreachable and non-exhaustive case alternatives
 -- -----------------------------------------------------------------------------
 
-checkCaseAlts :: CaseType -> [Alt ()] -> WCM ()
-checkCaseAlts _  []                   = ok
-checkCaseAlts ct alts@(Alt spi _ _ : _) = do
+checkCaseAlts :: SpanInfo -> CaseType -> [Alt ()] -> WCM ()
+checkCaseAlts _ _ []      = ok
+checkCaseAlts spi ct alts = do
+  let spis = map (\(Alt s _ _) -> s) alts
   let pats = map (\(Alt _ pat _) -> [pat]) alts
-  (nonExhaustive, overlapped, nondet) <- checkPatternMatching pats
+  let guards = map alt2Guards alts
+  (nonExhaustive, overlapped, nondet) <- checkPatternMatching pats guards
   case ct of
     Flex -> do
       unless (null nonExhaustive) $ warnFor WarnIncompletePatterns $ report $
-        warnMissingPattern p "an fcase alternative" nonExhaustive
+        warnMissingPattern spi "an fcase alternative" nonExhaustive
       when (nondet || not (null overlapped)) $ warnFor WarnOverlapping $ report
-        $ warnNondetOverlapping p "An fcase expression"
+        $ warnNondetOverlapping spi "An fcase expression"
     Rigid -> do
       unless (null nonExhaustive) $ warnFor WarnIncompletePatterns $ report $
-        warnMissingPattern p "a case alternative" nonExhaustive
-      unless (null overlapped) $ warnFor WarnOverlapping $ report $
-        warnUnreachablePattern p overlapped
-  where p = spanInfo2Pos spi
+        warnMissingPattern spi "a case alternative" nonExhaustive
+      unless (null overlapped) $ void $ mapM (warnFor WarnOverlapping . report) $
+        map (\(i, pat) -> warnUnreachablePattern (spis !! i) pat) overlapped
+  where alt2Guards :: Alt () -> [CondExpr ()]
+        alt2Guards (Alt _ _ (GuardedRhs _ _ conds _)) = conds
+        alt2Guards _ = []
 
 -- -----------------------------------------------------------------------------
 -- Check for non-exhaustive and overlapping patterns.
@@ -583,19 +590,19 @@ checkCaseAlts ct alts@(Alt spi _ _ : _) = do
 -- matching in function declarations and (f)case expressions.
 -- -----------------------------------------------------------------------------
 
-checkPatternMatching :: [[Pattern ()]]
-                     -> WCM ([ExhaustivePats], [[Pattern ()]], Bool)
-checkPatternMatching pats = do
+checkPatternMatching :: [[Pattern ()]] -> [[CondExpr ()]]
+                     -> WCM ([ExhaustivePats], [OverlappingPats], Bool)
+checkPatternMatching pats guards = do
   -- 1. We simplify the patterns by removing syntactic sugar temporarily
   --    for a simpler implementation.
   simplePats <- mapM (mapM simplifyPat) pats
   -- 2. We compute missing and used pattern matching alternatives
-  (missing, used, nondet) <- processEqs (zip [1..] simplePats)
+  (missing, used, nondet) <- processEqs (zip3 [0..] simplePats guards)
   -- 3. If any, we report the missing patterns, whereby we re-add the syntactic
   --    sugar removed in step (1) for a more precise output.
   nonExhaustive <- mapM tidyExhaustivePats missing
-  let overlap = [ eqn | (i, eqn) <- zip [1..] pats, i `IntSet.notMember` used]
-  return (nonExhaustive , overlap, nondet)
+  let overlap = [(i, eqn) | (i, eqn) <- zip [0..] pats, i `IntSet.notMember` used]
+  return (nonExhaustive, overlap, nondet)
 
 -- |Simplify a 'Pattern' until it only consists of
 --   * Variables
@@ -658,23 +665,38 @@ simplifyListPattern =
 -- number of possible patterns. Missing literals are therefore converted
 -- into the form @ ... x ... with x `notElem` [l1, ..., ln]@.
 type EqnPats = [Pattern ()]
+type EqnGuards = [CondExpr ()]
 type EqnNo   = Int
-type EqnInfo = (EqnNo, EqnPats)
+type EqnInfo = (EqnNo, EqnPats, EqnGuards)
 
 type ExhaustivePats = (EqnPats, [(Ident, [Literal])])
+type OverlappingPats = (EqnNo, EqnPats)
 type EqnSet  = IntSet.IntSet
 
 -- |Compute the missing pattern by inspecting the first patterns and
 -- categorize them as literal, constructor or variable patterns.
 processEqs :: [EqnInfo] -> WCM ([ExhaustivePats], EqnSet, Bool)
 processEqs []              = return ([], IntSet.empty, False)
-processEqs eqs@((n, ps):_)
-  | null ps                    = return ([], IntSet.singleton n, length eqs > 1)
-  | any isLitPat firstPats     = processLits eqs
-  | any isConPat firstPats     = processCons eqs
-  | all isVarPat firstPats     = processVars eqs
-  | otherwise                  = internalError "Checks.WarnCheck.processEqs"
+processEqs eqs@((n, ps, gs):eqs')
+  | null ps                = if guardsExhaustive then return ([], IntSet.singleton n, length eqs > 1)
+                                                 else do -- Current expression is guarded, thus potentially
+                                                         -- non-exhaustive. Therefore process remaining expressions.
+                                                         (missing', used', _) <- processEqs eqs'
+                                                         return (missing', IntSet.insert n used', length eqs > 1)
+  | any isLitPat firstPats = processLits eqs
+  | any isConPat firstPats = processCons eqs
+  | all isVarPat firstPats = processVars eqs
+  | otherwise              = internalError "Checks.WarnCheck.processEqs"
   where firstPats = map firstPat eqs
+        guardsExhaustive = null gs || any guardAlwaysTrue gs
+        guardAlwaysTrue :: CondExpr () -> Bool
+        guardAlwaysTrue (CondExpr _ e _) = case e of
+          Constructor _ _ q -> qidAlwaysTrue q
+          Variable    _ _ q -> qidAlwaysTrue q
+          _ -> False
+        qidAlwaysTrue :: QualIdent -> Bool
+        qidAlwaysTrue q = elem (idName $ qidIdent q) ["True", "success", "otherwise"]
+        
 
 -- |Literal patterns are checked by extracting the matched literals
 --  and constructing a pattern for any missing case.
@@ -697,7 +719,7 @@ processLits qs@(q:_) = do
   defaults   = [ shiftPat q' | q' <- qs, isVarPat (firstPat q') ]
   -- Pattern for all non-matched literals
   defaultPat = ( VariablePattern NoSpanInfo () newVar :
-                   replicate (length (snd q) - 1) wildPat
+                   replicate (length (snd3 q) - 1) wildPat
                , [(newVar, usedLits)]
                )
   newVar     = mkIdent "x"
@@ -744,7 +766,7 @@ processCons qs@(q:_) = do
   -- default alternatives (variable pattern)
   defaults     = [ shiftPat q' | q' <- qs, isVarPat (firstPat q') ]
   -- Pattern for a non-matched constructors
-  defaultPat c = (mkPattern c : replicate (length (snd q) - 1) wildPat, [])
+  defaultPat c = (mkPattern c : replicate (length (snd3 q) - 1) wildPat, [])
   mkPattern  c = ConstructorPattern NoSpanInfo ()
                   (qualifyLike (fst $ head used_cons) (constrIdent c))
                   (replicate (length $ constrTypes c) wildPat)
@@ -764,17 +786,17 @@ processUsedCons cons qs = do
 
   makeCon c a ps = let (args, rest) = splitAt a ps
                    in ConstructorPattern NoSpanInfo () c args : rest
-
-  removeFirstCon c a (n, p:ps)
-    | isVarPat p = (n, replicate a wildPat ++ ps)
-    | isCon c  p = (n, patArgs p           ++ ps)
+  
+  removeFirstCon c a (n, p:ps, gs)
+    | isVarPat p = (n, replicate a wildPat ++ ps, gs)
+    | isCon c  p = (n, patArgs p           ++ ps, gs)
   removeFirstCon _ _ _ = internalError "Checks.WarnCheck.removeFirstCon"
 
 -- |Variable patterns are exhaustive, so they are checked by simply
 -- checking the following patterns.
 processVars :: [EqnInfo] -> WCM ([ExhaustivePats], EqnSet, Bool)
 processVars []               = error "WarnCheck.processVars"
-processVars eqs@((n, _) : _) = do
+processVars eqs@((n, _, _) : _) = do
   let ovlp = length eqs > 1
   (missing, used, nd) <- processEqs (map shiftPat eqs)
   return ( map (\(xs, ys) -> (wildPat : xs, ys)) missing
@@ -793,8 +815,8 @@ getConTy q = do
   tyEnv <- gets valueEnv
   tcEnv <- gets tyConsEnv
   case qualLookupValue q tyEnv of
-    [DataConstructor  _ _ _ (ForAllExist _ _ (PredType _ ty))] -> return ty
-    [NewtypeConstructor _ _ (ForAllExist _ _ (PredType _ ty))] -> return ty
+    [DataConstructor  _ _ _ (ForAll _ (PredType _ ty))] -> return ty
+    [NewtypeConstructor _ _ (ForAll _ (PredType _ ty))] -> return ty
     _ -> case qualLookupTypeInfo q tcEnv of
       [AliasType _ _ _ ty] -> return ty
       _ -> internalError $ "Checks.WarnCheck.getConTy: " ++ show q
@@ -804,14 +826,17 @@ getTyCons :: QualIdent -> WCM [DataConstr]
 getTyCons tc = do
   tc'   <- unAlias tc
   tcEnv <- gets tyConsEnv
-  return $ case lookupTypeInfo (unqualify tc) tcEnv of
-    [DataType     _ _ cs] -> cs
-    [RenamingType _ _ nc] -> [nc]
-    _ -> case qualLookupTypeInfo tc' tcEnv of
-      [DataType     _ _ cs] -> cs
-      [RenamingType _ _ nc] -> [nc]
-      err -> internalError $ "Checks.WarnCheck.getTyCons: " ++
-                               show tc ++ ' ' : show err ++ '\n' : show tcEnv
+  let getTyCons' :: [TypeInfo] -> Either String [DataConstr]
+      getTyCons' ti = case ti of
+        [DataType     _ _ cs] -> Right cs
+        [RenamingType _ _ nc] -> Right $ [nc]
+        _                     -> Left $ "Checks.WarnCheck.getTyCons: " ++ show tc ++ ' ' : show ti ++ '\n' : show tcEnv
+      csResult = getTyCons' (qualLookupTypeInfo tc tcEnv)
+             <|> getTyCons' (qualLookupTypeInfo tc' tcEnv)
+             <|> getTyCons' (lookupTypeInfo (unqualify tc) tcEnv) -- Fall back on unqualified lookup if qualified doesn't work
+  case csResult of
+    Right cs -> return cs
+    Left err -> internalError err
 
 -- |Resugar the exhaustive patterns previously desugared at 'simplifyPat'.
 tidyExhaustivePats :: ExhaustivePats -> WCM ExhaustivePats
@@ -851,13 +876,13 @@ tidyPat p = internalError $ "Checks.WarnCheck.tidyPat: " ++ show p
 
 -- |Get the first pattern of a list.
 firstPat :: EqnInfo -> Pattern ()
-firstPat (_, []   ) = internalError "Checks.WarnCheck.firstPat: empty list"
-firstPat (_, (p:_)) = p
+firstPat (_, [],    _) = internalError "Checks.WarnCheck.firstPat: empty list"
+firstPat (_, (p:_), _) = p
 
 -- |Drop the first pattern of a list.
 shiftPat :: EqnInfo -> EqnInfo
-shiftPat (_, []    ) = internalError "Checks.WarnCheck.shiftPat: empty list"
-shiftPat (n, (_:ps)) = (n, ps)
+shiftPat (_, [],     _ ) = internalError "Checks.WarnCheck.shiftPat: empty list"
+shiftPat (n, (_:ps), gs) = (n, ps, gs)
 
 -- |Wildcard pattern.
 wildPat :: Pattern ()
@@ -914,8 +939,8 @@ patArgs _                             = []
 -- |Warning message for non-exhaustive patterns.
 -- To shorten the output only the first 'maxPattern' are printed,
 -- additional pattern are abbreviated by dots.
-warnMissingPattern :: Position -> String -> [ExhaustivePats] -> Message
-warnMissingPattern p loc pats = posMessage p
+warnMissingPattern :: SpanInfo -> String -> [ExhaustivePats] -> Message
+warnMissingPattern spi loc pats = spanInfoMessage spi
   $   text "Pattern matches are non-exhaustive"
   $+$ text "In" <+> text loc <> char ':'
   $+$ nest 2 (text "Patterns not matched:" $+$ nest 2 (vcat (ppExPats pats)))
@@ -927,31 +952,27 @@ warnMissingPattern p loc pats = posMessage p
   ppExPat (ps, cs)
     | null cs   = ppPats
     | otherwise = ppPats <+> text "with" <+> hsep (map ppCons cs)
-    where ppPats = hsep (map (ppPattern 2) ps)
-  ppCons (i, lits) = ppIdent i <+> text "`notElem`"
-            <+> ppExpr 0 (List NoSpanInfo () (map (Literal NoSpanInfo ()) lits))
+    where ppPats = hsep (map (pPrintPrec 2) ps)
+  ppCons (i, lits) = pPrint i <+> text "`notElem`"
+            <+> pPrintPrec 0 (List NoSpanInfo () (map (Literal NoSpanInfo ()) lits))
 
 -- |Warning message for unreachable patterns.
 -- To shorten the output only the first 'maxPattern' are printed,
 -- additional pattern are abbreviated by dots.
-warnUnreachablePattern :: Position  -> [[Pattern a]] -> Message
-warnUnreachablePattern p pats = posMessage p
+warnUnreachablePattern :: SpanInfo  -> [Pattern a] -> Message
+warnUnreachablePattern spi pats = spanInfoMessage spi
   $   text "Pattern matches are potentially unreachable"
   $+$ text "In a case alternative:"
-  $+$ nest 2 (vcat (ppExPats pats) <+> text "->" <+> text "...")
+  $+$ nest 2 (ppPat pats <+> text "->" <+> text "...")
   where
-  ppExPats ps
-    | length ps > maxPattern = ppPats ++ [text "..."]
-    | otherwise              = ppPats
-    where ppPats = map ppPat (take maxPattern ps)
-  ppPat ps = hsep (map (ppPattern 2) ps)
+  ppPat ps = hsep (map (pPrintPrec 2) ps)
 
 -- |Maximum number of missing patterns to be shown.
 maxPattern :: Int
 maxPattern = 4
 
-warnNondetOverlapping :: Position -> String -> Message
-warnNondetOverlapping p loc = posMessage p $
+warnNondetOverlapping :: SpanInfo -> String -> Message
+warnNondetOverlapping spi loc = spanInfoMessage spi $
   text loc <+> text "is potentially non-deterministic due to overlapping rules"
 
 -- -----------------------------------------------------------------------------
@@ -960,10 +981,6 @@ checkShadowing :: Ident -> WCM ()
 checkShadowing x = warnFor WarnNameShadowing $
   shadowsVar x >>= maybe ok (report . warnShadowing x)
 
-checkTypeShadowing :: Ident -> WCM ()
-checkTypeShadowing x = warnFor WarnNameShadowing $
-  shadowsTypeVar x >>= maybe ok (report . warnTypeShadowing x)
-
 reportUnusedVars :: WCM ()
 reportUnusedVars = reportAllUnusedVars WarnUnusedBindings
 
@@ -986,26 +1003,26 @@ reportUnusedTypeVars vs = warnFor WarnUnusedBindings $ do
 -- sides.
 
 insertDecl :: Decl a -> WCM ()
-insertDecl (DataDecl     _ d _ cs _) = do
+insertDecl (DataDecl      _ d _ cs _) = do
   insertTypeConsId d
   mapM_ insertConstrDecl cs
-insertDecl (ExternalDataDecl  _ d _) = insertTypeConsId d
-insertDecl (NewtypeDecl  _ d _ nc _) = do
+insertDecl (ExternalDataDecl   _ d _) = insertTypeConsId d
+insertDecl (NewtypeDecl   _ d _ nc _) = do
   insertTypeConsId d
   insertNewConstrDecl nc
-insertDecl (TypeDecl       _ t _ ty) = do
+insertDecl (TypeDecl        _ t _ ty) = do
   insertTypeConsId t
   insertTypeExpr ty
-insertDecl (FunctionDecl    _ _ f _) = do
+insertDecl (FunctionDecl     _ _ f _) = do
   cons <- isConsId f
   unless cons $ insertVar f
-insertDecl (ExternalDecl       _ vs) = mapM_ (insertVar . varIdent) vs
-insertDecl (PatternDecl       _ p _) = insertPattern False p
-insertDecl (FreeDecl           _ vs) = mapM_ (insertVar . varIdent) vs
-insertDecl (ClassDecl _ _ cls _  ds) = do
+insertDecl (ExternalDecl        _ vs) = mapM_ (insertVar . varIdent) vs
+insertDecl (PatternDecl        _ p _) = insertPattern False p
+insertDecl (FreeDecl            _ vs) = mapM_ (insertVar . varIdent) vs
+insertDecl (ClassDecl _ _ _ cls _ ds) = do
   insertTypeConsId cls
   mapM_ insertVar $ concatMap methods ds
-insertDecl _                         = ok
+insertDecl _                          = ok
 
 insertTypeExpr :: TypeExpr -> WCM ()
 insertTypeExpr (VariableType       _ _) = ok
@@ -1018,9 +1035,9 @@ insertTypeExpr (ParenType         _ ty) = insertTypeExpr ty
 insertTypeExpr (ForallType      _ _ ty) = insertTypeExpr ty
 
 insertConstrDecl :: ConstrDecl -> WCM ()
-insertConstrDecl (ConstrDecl _ _ _    c _) = insertConsId c
-insertConstrDecl (ConOpDecl  _ _ _ _ op _) = insertConsId op
-insertConstrDecl (RecordDecl _ _ _    c _) = insertConsId c
+insertConstrDecl (ConstrDecl _    c _) = insertConsId c
+insertConstrDecl (ConOpDecl  _ _ op _) = insertConsId op
+insertConstrDecl (RecordDecl _    c _) = insertConsId c
 
 insertNewConstrDecl :: NewConstrDecl -> WCM ()
 insertNewConstrDecl (NewConstrDecl _ c _) = insertConsId c
@@ -1125,9 +1142,6 @@ shadows qid s = do
 shadowsVar :: Ident -> WCM (Maybe Ident)
 shadowsVar v = gets (shadows $ commonId v)
 
-shadowsTypeVar :: Ident -> WCM (Maybe Ident)
-shadowsTypeVar v = gets (shadows $ typeId v)
-
 visitId :: Ident -> WCM ()
 visitId v = modifyScope (qualModifyNestEnv visitVariable (commonId v))
 
@@ -1172,7 +1186,7 @@ isKnown s qid = qualInLocalNestEnv qid (scope s)
 
 isUnref :: WcState -> QualIdent -> Bool
 isUnref s qid = let sc = scope s
-                in  (any (not . variableVisited) (qualLookupNestEnv qid sc))
+                in  any (not . variableVisited) (qualLookupNestEnv qid sc)
                     && qualInLocalNestEnv qid sc
 
 isVar :: QualIdent -> WcState -> Bool
@@ -1238,32 +1252,26 @@ checkCaseModeDecl (PatternDecl _ t rhs) = do
 checkCaseModeDecl (FreeDecl  _ vs) =
   mapM_ (checkCaseModeID isVarName . varIdent) vs
 checkCaseModeDecl (DefaultDecl _ tys) = mapM_ checkTypeExpr tys
-checkCaseModeDecl (ClassDecl _ cx cls tv ds) = do
+checkCaseModeDecl (ClassDecl _ _ cx cls tv ds) = do
   checkCaseModeContext cx
   checkCaseModeID isClassDeclName cls
   checkCaseModeID isVarName tv
   mapM_ checkCaseModeDecl ds
-checkCaseModeDecl (InstanceDecl _ cx _ inst ds) = do
+checkCaseModeDecl (InstanceDecl _ _ cx _ inst ds) = do
   checkCaseModeContext cx
   checkCaseModeTypeExpr inst
   mapM_ checkCaseModeDecl ds
 checkCaseModeDecl _ = ok
 
 checkCaseModeConstr :: ConstrDecl -> WCM ()
-checkCaseModeConstr (ConstrDecl _ evs cx c tys) = do
-  mapM_ (checkCaseModeID isVarName) evs
-  checkCaseModeContext cx
+checkCaseModeConstr (ConstrDecl _ c tys) = do
   checkCaseModeID isConstrName c
   mapM_ checkCaseModeTypeExpr tys
-checkCaseModeConstr (ConOpDecl  _ evs cx ty1 c ty2) = do
-  mapM_ (checkCaseModeID isVarName) evs
-  checkCaseModeContext cx
+checkCaseModeConstr (ConOpDecl  _ ty1 c ty2) = do
   checkCaseModeTypeExpr ty1
   checkCaseModeID isConstrName c
   checkCaseModeTypeExpr ty2
-checkCaseModeConstr (RecordDecl _ evs cx c fs) = do
-  mapM_ (checkCaseModeID isVarName) evs
-  checkCaseModeContext cx
+checkCaseModeConstr (RecordDecl _ c fs) = do
   checkCaseModeID isConstrName c
   mapM_ checkCaseModeFieldDecl fs
 
@@ -1326,10 +1334,10 @@ checkCaseModeLhs (ApLhs _ lhs ts) = do
   mapM_ checkCaseModePattern ts
 
 checkCaseModeRhs :: Rhs a -> WCM ()
-checkCaseModeRhs (SimpleRhs _ e ds) = do
+checkCaseModeRhs (SimpleRhs _ _ e ds) = do
   checkCaseModeExpr e
   mapM_ checkCaseModeDecl ds
-checkCaseModeRhs (GuardedRhs _ es ds) = do
+checkCaseModeRhs (GuardedRhs _ _ es ds) = do
   mapM_ checkCaseModeCondExpr es
   mapM_ checkCaseModeDecl ds
 
@@ -1397,25 +1405,25 @@ checkCaseModeExpr (RightSection _ _ e) = checkCaseModeExpr e
 checkCaseModeExpr (Lambda _ ts e) = do
   mapM_ checkCaseModePattern ts
   checkCaseModeExpr e
-checkCaseModeExpr (Let _ ds e) = do
+checkCaseModeExpr (Let _ _ ds e) = do
   mapM_ checkCaseModeDecl ds
   checkCaseModeExpr e
-checkCaseModeExpr (Do _ stms e) = do
+checkCaseModeExpr (Do _ _ stms e) = do
   mapM_ checkCaseModeStatement stms
   checkCaseModeExpr e
 checkCaseModeExpr (IfThenElse _ e1 e2 e3) = do
   checkCaseModeExpr e1
   checkCaseModeExpr e2
   checkCaseModeExpr e3
-checkCaseModeExpr (Case _ _ e as) = do
+checkCaseModeExpr (Case _ _ _ e as) = do
   mapM_ checkCaseModeAlt as
   checkCaseModeExpr e
 checkCaseModeExpr _ = ok
 
 checkCaseModeStatement :: Statement a -> WCM ()
-checkCaseModeStatement (StmtExpr _ e) = checkCaseModeExpr e
-checkCaseModeStatement (StmtDecl _ ds) = mapM_ checkCaseModeDecl ds
-checkCaseModeStatement (StmtBind _ t e) = do
+checkCaseModeStatement (StmtExpr _    e) = checkCaseModeExpr e
+checkCaseModeStatement (StmtDecl _ _ ds) = mapM_ checkCaseModeDecl ds
+checkCaseModeStatement (StmtBind _  t e) = do
   checkCaseModePattern t
   checkCaseModeExpr e
 
@@ -1457,12 +1465,152 @@ isDataDeclName CaseModeGoedel  (x:_) | isAlpha x = isUpper x
 isDataDeclName CaseModeHaskell (x:_) | isAlpha x = isUpper x
 isDataDeclName _               _     = True
 
+-- ---------------------------------------------------------------------------
+-- Warn for redundant context
+-- ---------------------------------------------------------------------------
+
+--traverse the AST for QualTypeExpr/Context and check for redundancy
+checkRedContext :: [Decl a] -> WCM ()
+checkRedContext = warnFor WarnRedundantContext . mapM_ checkRedContextDecl
+
+getRedPredSet :: ModuleIdent -> ClassEnv -> TCEnv -> PredSet -> PredSet
+getRedPredSet m cenv tcEnv ps =
+  Set.map (pm Map.!) $ Set.difference qps $ minPredSet cenv qps --or fromJust $ Map.lookup
+  where (qps, pm) = Set.foldr qualifyAndAddPred (Set.empty, Map.empty) ps
+        qualifyAndAddPred p@(Pred qid ty) (ps', pm') =
+          let qp = Pred (getOrigName m qid tcEnv) ty
+          in (Set.insert qp ps', Map.insert qp p pm')
+
+getPredFromContext :: Context -> ([Ident], PredSet)
+getPredFromContext cx =
+  let vs = concatMap (\(Constraint _ _ ty) -> typeVariables ty) cx
+  in (vs, toPredSet vs cx)
+
+checkRedContext' :: (Pred -> Message) -> PredSet -> WCM ()
+checkRedContext' f ps = do
+  m     <- gets moduleId
+  cenv  <- gets classEnv
+  tcEnv <- gets tyConsEnv
+  mapM_ (report . f) (getRedPredSet m cenv tcEnv ps)
+
+checkRedContextDecl :: Decl a -> WCM ()
+checkRedContextDecl (TypeSig _ ids (QualTypeExpr _ cx _)) =
+  checkRedContext' (warnRedContext (warnRedFuncString ids) vs) ps
+  where (vs, ps) = getPredFromContext cx
+checkRedContextDecl (FunctionDecl _ _ _ eqs) = mapM_ checkRedContextEq eqs
+checkRedContextDecl (PatternDecl _ _ rhs) = checkRedContextRhs rhs
+checkRedContextDecl (ClassDecl _ _ cx i _ ds) = do
+  checkRedContext'
+    (warnRedContext (text ("class declaration " ++ escName i)) vs)
+    ps
+  mapM_ checkRedContextDecl ds
+  where (vs, ps) = getPredFromContext cx
+checkRedContextDecl (InstanceDecl _ _ cx qid _ ds) = do
+  checkRedContext'
+    (warnRedContext (text ("instance declaration " ++ escQualName qid)) vs)
+    ps
+  mapM_ checkRedContextDecl ds
+  where (vs, ps) = getPredFromContext cx
+checkRedContextDecl _ = return ()
+
+checkRedContextEq :: Equation a -> WCM ()
+checkRedContextEq (Equation _ _ rhs) = checkRedContextRhs rhs
+
+checkRedContextRhs :: Rhs a -> WCM ()
+checkRedContextRhs (SimpleRhs  _ _ e  ds) = do
+  checkRedContextExpr e
+  mapM_ checkRedContextDecl ds
+checkRedContextRhs (GuardedRhs _ _ cs ds) = do
+  mapM_ checkRedContextCond cs
+  mapM_ checkRedContextDecl ds
+
+checkRedContextCond :: CondExpr a -> WCM ()
+checkRedContextCond (CondExpr _ e1 e2) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+
+checkRedContextExpr :: Expression a -> WCM ()
+checkRedContextExpr (Paren _ e) = checkRedContextExpr e
+checkRedContextExpr (Typed _ e (QualTypeExpr _ cx _)) = do
+  checkRedContextExpr e
+  checkRedContext' (warnRedContext (text "type signature") vs) ps
+  where (vs, ps) = getPredFromContext cx
+checkRedContextExpr (Record _ _ _ fs) = mapM_ checkRedContextFieldExpr fs
+checkRedContextExpr (RecordUpdate _ e fs) = do
+  checkRedContextExpr e
+  mapM_ checkRedContextFieldExpr fs
+checkRedContextExpr (Tuple  _ es) = mapM_ checkRedContextExpr es
+checkRedContextExpr (List _ _ es) = mapM_ checkRedContextExpr es
+checkRedContextExpr (ListCompr _ e sts) = do
+  checkRedContextExpr e
+  mapM_ checkRedContextStmt sts
+checkRedContextExpr (EnumFrom _ e) = checkRedContextExpr e
+checkRedContextExpr (EnumFromThen _ e1 e2) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+checkRedContextExpr (EnumFromTo _ e1 e2) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+checkRedContextExpr (EnumFromThenTo _ e1 e2 e3) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+  checkRedContextExpr e3
+checkRedContextExpr (UnaryMinus _ e) = checkRedContextExpr e
+checkRedContextExpr (Apply _ e1 e2) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+checkRedContextExpr (InfixApply _ e1 _ e2) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+checkRedContextExpr (LeftSection  _ e _) = checkRedContextExpr e
+checkRedContextExpr (RightSection _ _ e) = checkRedContextExpr e
+checkRedContextExpr (Lambda _ _ e) = checkRedContextExpr e
+checkRedContextExpr (Let _ _ ds e) = do
+  mapM_ checkRedContextDecl ds
+  checkRedContextExpr e
+checkRedContextExpr (IfThenElse _ e1 e2 e3) = do
+  checkRedContextExpr e1
+  checkRedContextExpr e2
+  checkRedContextExpr e3
+checkRedContextExpr (Case _ _ _ e as) = do
+  checkRedContextExpr e
+  mapM_ checkRedContextAlt as
+checkRedContextExpr _ = return ()
+
+checkRedContextStmt :: Statement a -> WCM ()
+checkRedContextStmt (StmtExpr   _  e) = checkRedContextExpr e
+checkRedContextStmt (StmtDecl _ _ ds) = mapM_ checkRedContextDecl ds
+checkRedContextStmt (StmtBind _ _  e) = checkRedContextExpr e
+
+checkRedContextAlt :: Alt a -> WCM ()
+checkRedContextAlt (Alt _ _ rhs) = checkRedContextRhs rhs
+
+checkRedContextFieldExpr :: Field (Expression a) -> WCM ()
+checkRedContextFieldExpr (Field _ _ e) = checkRedContextExpr e
+
 -- ---------------------------------------------------------------------------
 -- Warnings messages
 -- ---------------------------------------------------------------------------
 
+warnRedFuncString :: [Ident] -> Doc
+warnRedFuncString is = text "type signature for function" <>
+                       text (if length is == 1 then [] else "s") <+>
+                       csep (map (text . escName) is)
+
+-- Doc description -> TypeVars -> Pred -> Warning
+warnRedContext :: Doc -> [Ident] -> Pred -> Message
+warnRedContext d vs p@(Pred qid _) = spanInfoMessage qid $
+  text "Redundant context in" <+> d <> colon <+>
+  quotes (pPrint $ fromPred vs p) -- idents use ` ' as quotes not ' '
+
+-- seperate a list by ', '
+csep :: [Doc] -> Doc
+csep []     = empty
+csep [x]    = x
+csep (x:xs) = x <> comma <+> csep xs
+
 warnCaseMode :: Ident -> CaseMode -> Message
-warnCaseMode i@(Ident _ name _ ) c = posMessage i $
+warnCaseMode i@(Ident _ name _ ) c = spanInfoMessage i $
   text "Wrong case mode in symbol" <+> text (escName i) <+>
   text "due to selected case mode" <+> text (escapeCaseMode c) <> comma <+>
   text "try renaming to" <+> text (caseSuggestion name) <+> text "instead"
@@ -1480,19 +1628,14 @@ escapeCaseMode CaseModeProlog  = "`prolog`"
 escapeCaseMode CaseModeGoedel  = "`goedel`"
 
 warnUnrefTypeVar :: Ident -> Message
-warnUnrefTypeVar v = posMessage v $ hsep $ map text
+warnUnrefTypeVar v = spanInfoMessage v $ hsep $ map text
   [ "Unreferenced type variable", escName v ]
 
 warnUnrefVar :: Ident -> Message
-warnUnrefVar v = posMessage v $ hsep $ map text
+warnUnrefVar v = spanInfoMessage v $ hsep $ map text
   [ "Unused declaration of variable", escName v ]
 
 warnShadowing :: Ident -> Ident -> Message
-warnShadowing x v = posMessage x $
+warnShadowing x v = spanInfoMessage x $
   text "Shadowing symbol" <+> text (escName x)
   <> comma <+> text "bound at:" <+> ppPosition (getPosition v)
-
-warnTypeShadowing :: Ident -> Ident -> Message
-warnTypeShadowing x v = posMessage x $
-  text "Shadowing type variable" <+> text (escName x)
-  <> comma <+> text "bound at:" <+> ppPosition (getPosition v)
