@@ -18,8 +18,9 @@
 -}
 module CompilerOpts
   ( Options (..), CppOpts (..), PrepOpts (..), WarnOpts (..), DebugOpts (..)
-  , CaseMode (..), CymakeMode (..), Verbosity (..), TargetType (..)
-  , WarnFlag (..), KnownExtension (..), DumpLevel (..), dumpLevel
+  , OptimizationOpts(..), CaseMode (..), CymakeMode (..), Verbosity (..)
+  , TargetType (..), WarnFlag (..), KnownExtension (..), DumpLevel (..)
+  , dumpLevel
   , defaultOptions, defaultPrepOpts, defaultWarnOpts, defaultDebugOpts
   , getCompilerOpts, updateOpts, usage
   ) where
@@ -33,7 +34,7 @@ import System.Environment              (getArgs, getProgName)
 import System.FilePath                 ( addTrailingPathSeparator, normalise
                                        , splitSearchPath )
 
-import Curry.Files.Filenames           (currySubdir)
+import Curry.Files.Filenames           (defaultOutDir)
 import Curry.Syntax.Extension
 
 -- -----------------------------------------------------------------------------
@@ -43,24 +44,26 @@ import Curry.Syntax.Extension
 -- |Compiler options
 data Options = Options
   -- general
-  { optMode         :: CymakeMode         -- ^ modus operandi
-  , optVerbosity    :: Verbosity          -- ^ verbosity level
+  { optMode          :: CymakeMode          -- ^ modus operandi
+  , optVerbosity     :: Verbosity           -- ^ verbosity level
   -- compilation
-  , optForce        :: Bool               -- ^ force (re-)compilation of target
-  , optLibraryPaths :: [FilePath]         -- ^ directories to search in
-                                          --   for libraries
-  , optImportPaths  :: [FilePath]         -- ^ directories to search in
-                                          --   for imports
-  , optHtmlDir      :: Maybe FilePath     -- ^ output directory for HTML
-  , optUseSubdir    :: Bool               -- ^ use subdir for output?
-  , optInterface    :: Bool               -- ^ create a FlatCurry interface file?
-  , optPrepOpts     :: PrepOpts           -- ^ preprocessor options
-  , optWarnOpts     :: WarnOpts           -- ^ warning options
-  , optTargetTypes  :: [TargetType]       -- ^ what to generate
-  , optExtensions   :: [KnownExtension]   -- ^ enabled language extensions
-  , optDebugOpts    :: DebugOpts          -- ^ debug options
-  , optCaseMode     :: CaseMode           -- ^ case mode
-  , optCppOpts      :: CppOpts            -- ^ C preprocessor options
+  , optForce         :: Bool                -- ^ force (re-)compilation of target
+  , optLibraryPaths  :: [FilePath]          -- ^ directories to search in
+                                            --   for libraries
+  , optImportPaths   :: [FilePath]          -- ^ directories to search in
+                                            --   for imports
+  , optOutDir        :: FilePath            -- ^ output directory for FlatCurry, ...
+  , optHtmlDir       :: Maybe FilePath      -- ^ output directory for HTML
+  , optUseOutDir     :: Bool                -- ^ use subdir for output?
+  , optInterface     :: Bool                -- ^ create a FlatCurry interface file?
+  , optPrepOpts      :: PrepOpts            -- ^ preprocessor options
+  , optWarnOpts      :: WarnOpts            -- ^ warning options
+  , optTargetTypes   :: [TargetType]        -- ^ what to generate
+  , optExtensions    :: [KnownExtension]    -- ^ enabled language extensions
+  , optDebugOpts     :: DebugOpts           -- ^ debug options
+  , optCaseMode      :: CaseMode            -- ^ case mode
+  , optCppOpts       :: CppOpts             -- ^ C preprocessor options
+  , optOptimizations :: OptimizationOpts -- ^ Optimization options
   } deriving Show
 
 -- |C preprocessor options
@@ -100,24 +103,32 @@ data DebugOpts = DebugOpts
   , dbDumpSimple      :: Bool        -- ^ print more readable environments
   } deriving Show
 
+data OptimizationOpts = OptimizationOpts
+  { optDesugarNewtypes     :: Bool -- ^ Desugar newtypes
+  , optInlineDictionaries  :: Bool -- ^ Inline type class dictionaries
+  , optRemoveUnusedImports :: Bool -- ^ Remove unused imports in IL
+  } deriving Show
+
 -- | Default compiler options
 defaultOptions :: Options
 defaultOptions = Options
-  { optMode         = ModeMake
-  , optVerbosity    = VerbStatus
-  , optForce        = False
-  , optLibraryPaths = []
-  , optImportPaths  = []
-  , optHtmlDir      = Nothing
-  , optUseSubdir    = True
-  , optInterface    = True
-  , optPrepOpts     = defaultPrepOpts
-  , optWarnOpts     = defaultWarnOpts
-  , optTargetTypes  = []
-  , optExtensions   = []
-  , optDebugOpts    = defaultDebugOpts
-  , optCaseMode     = CaseModeFree
-  , optCppOpts      = defaultCppOpts
+  { optMode          = ModeMake
+  , optVerbosity     = VerbStatus
+  , optForce         = False
+  , optLibraryPaths  = []
+  , optImportPaths   = []
+  , optOutDir        = defaultOutDir
+  , optHtmlDir       = Nothing
+  , optUseOutDir     = True
+  , optInterface     = True
+  , optPrepOpts      = defaultPrepOpts
+  , optWarnOpts      = defaultWarnOpts
+  , optTargetTypes   = []
+  , optExtensions    = []
+  , optDebugOpts     = defaultDebugOpts
+  , optCaseMode      = CaseModeFree
+  , optCppOpts       = defaultCppOpts
+  , optOptimizations = defaultOptimizationOpts
   }
 
 -- | Default C preprocessor options
@@ -153,6 +164,13 @@ defaultDebugOpts = DebugOpts
   , dbDumpSimple      = False
   }
 
+defaultOptimizationOpts :: OptimizationOpts
+defaultOptimizationOpts = OptimizationOpts
+  { optDesugarNewtypes     = False
+  , optInlineDictionaries  = True
+  , optRemoveUnusedImports = True
+  }
+
 -- |Modus operandi of the program
 data CymakeMode
   = ModeHelp           -- ^ Show help information and exit
@@ -179,11 +197,11 @@ data TargetType
   | Comments             -- ^ Source code comments
   | Parsed               -- ^ Parsed source code
   | FlatCurry            -- ^ FlatCurry
+  | AnnotatedFlatCurry   -- ^ Annotated FlatCurry
   | TypedFlatCurry       -- ^ Typed FlatCurry
-  | TypeAnnotatedFlatCurry -- ^ Type-annotated FlatCurry
-  | AbstractCurry          -- ^ AbstractCurry
-  | UntypedAbstractCurry   -- ^ Untyped AbstractCurry
-  | Html                   -- ^ HTML documentation
+  | AbstractCurry        -- ^ AbstractCurry
+  | UntypedAbstractCurry -- ^ Untyped AbstractCurry
+  | Html                 -- ^ HTML documentation
   | AST                  -- ^ Abstract-Syntax-Tree after checks
   | ShortAST             -- ^ Abstract-Syntax-Tree with shortened decls
     deriving (Eq, Show)
@@ -200,7 +218,8 @@ data WarnFlag
   | WarnMissingSignatures    -- ^ Warn for missing type signatures
   | WarnMissingMethods       -- ^ Warn for missing method implementations
   | WarnOrphanInstances      -- ^ Warn for orphan instances
-  | WarnIrregularCaseMode
+  | WarnIrregularCaseMode    -- ^ Warn for irregular case mode
+  | WarnRedundantContext     -- ^ Warn for redundant context in type signatures
     deriving (Eq, Bounded, Enum, Show)
 
 -- |Warning flags enabled by default
@@ -209,7 +228,7 @@ stdWarnFlags =
   [ WarnMultipleImports   , WarnDisjoinedRules   --, WarnUnusedGlobalBindings
   , WarnUnusedBindings    , WarnNameShadowing    , WarnOverlapping
   , WarnIncompletePatterns, WarnMissingSignatures, WarnMissingMethods
-  , WarnIrregularCaseMode
+  , WarnIrregularCaseMode , WarnRedundantContext
   ]
 
 -- |Description and flag of warnings flags
@@ -237,6 +256,8 @@ warnFlags =
     , "orphan instances"               )
   , ( WarnIrregularCaseMode   , "irregular-case-mode"
     , "irregular case mode")
+  , ( WarnRedundantContext    , "redundant-context"
+    , "redundant context")
   ]
 
 -- |Dump level
@@ -298,8 +319,6 @@ extensions =
     , "enable anonymous free variables"              )
   , ( CPP                      , "CPP"
     , "run C preprocessor"                           )
-  , ( ExistentialQuantification, "ExistentialQuantification"
-    , "enable existentially quantified types"        )
   , ( FunctionalPatterns       , "FunctionalPatterns"
     , "enable functional patterns"                   )
   , ( NegativeLiterals         , "NegativeLiterals"
@@ -342,6 +361,10 @@ onDebugOpts :: (DebugOpts -> DebugOpts) -> OptErr -> OptErr
 onDebugOpts f (opts, errs)
   = (opts { optDebugOpts = f (optDebugOpts opts) }, errs)
 
+onOptimOpts :: (OptimizationOpts -> OptimizationOpts) -> OptErr -> OptErr
+onOptimOpts f (opts, errs)
+    = (opts { optOptimizations = f (optOptimizations opts) }, errs)
+
 withArg :: ((a -> b) -> OptErr -> OptErr)
         -> (String -> a -> b) -> String -> OptErr -> OptErr
 withArg lift f arg = lift (f arg)
@@ -407,13 +430,16 @@ options =
               map (normalise . addTrailingPathSeparator) (splitSearchPath arg)
               }) "dir[:dir]")
       "search for imports in dir[:dir]"
+  , Option "o"  ["outdir"]
+      (ReqArg (withArg onOpts $ \ arg opts -> opts { optOutDir = arg }) "dir")
+      "write compilation artifacts (FlatCurry, ...) into directory `dir'"
   , Option []   ["htmldir"]
       (ReqArg (withArg onOpts $ \ arg opts -> opts { optHtmlDir =
         Just arg }) "dir")
       "write HTML documentation into directory `dir'"
-  , Option ""   ["no-subdir"]
-      (NoArg (onOpts $ \ opts -> opts { optUseSubdir = False }))
-      ("disable writing to `" ++ currySubdir ++ "' subdirectory")
+  , Option ""   ["no-outdir", "no-subdir"]
+      (NoArg (onOpts $ \ opts -> opts { optUseOutDir = False }))
+      ("disable writing to `" ++ defaultOutDir ++ "' subdirectory")
   , Option ""   ["no-intf"]
       (NoArg (onOpts $ \ opts -> opts { optInterface = False }))
       "do not create an interface file"
@@ -436,7 +462,7 @@ options =
       "generate FlatCurry code"
   , targetOption TypedFlatCurry         "typed-flat"
       "generate typed FlatCurry code"
-  , targetOption TypeAnnotatedFlatCurry "type-annotated-flat"
+  , targetOption AnnotatedFlatCurry     "type-annotated-flat"
       "generate type-annotated FlatCurry code"
   , targetOption AbstractCurry          "acy"
       "generate typed AbstractCurry"
@@ -464,10 +490,11 @@ options =
       (NoArg (onOpts $ \ opts -> opts { optExtensions =
         nub $ kielExtensions ++ optExtensions opts }))
       "enable extended Curry functionalities"
-  , mkOptDescr onOpts      "c" ["case-mode"] "mode" "case mode"          caseModeDescriptions
-  , mkOptDescr onOpts      "X" []            "ext"  "language extension" extDescriptions
-  , mkOptDescr onWarnOpts  "W" []            "opt"  "warning option"     warnDescriptions
-  , mkOptDescr onDebugOpts "d" []            "opt"  "debug option"       debugDescriptions
+  , mkOptDescr onOpts      "c" ["case-mode"] "mode" "case mode"           caseModeDescriptions
+  , mkOptDescr onOpts      "X" []            "ext"  "language extension"  extDescriptions
+  , mkOptDescr onWarnOpts  "W" []            "opt"  "warning option"      warnDescriptions
+  , mkOptDescr onDebugOpts "d" []            "opt"  "debug option"        debugDescriptions
+  , mkOptDescr onOptimOpts "O" []            "opt"  "optimization option" optimizeDescriptions
   , Option ""   ["cpp"]
       (NoArg (onCppOpts $ \ opts -> opts { cppRun = True }))
       "run C preprocessor"
@@ -482,7 +509,7 @@ parseCppDefinition arg optErr
   = onCppOpts (addCppDefinition s v) optErr
   | otherwise
   = addErr (cppDefinitionErr arg) optErr
-  where (s, v) = fmap (drop 1) $ break ('=' ==) arg
+  where (s, v) = drop 1 <$> break ('=' ==) arg
 
 addCppDefinition :: String -> String -> CppOpts -> CppOpts
 addCppDefinition s v opts =
@@ -492,9 +519,9 @@ cppDefinitionErr :: String -> String
 cppDefinitionErr = (++) "Invalid format for option '-D': "
 
 targetOption :: TargetType -> String -> String -> OptDescr (OptErr -> OptErr)
-targetOption ty flag desc
+targetOption ty flag
   = Option "" [flag] (NoArg (onOpts $ \ opts -> opts { optTargetTypes =
-      nub $ ty : optTargetTypes opts })) desc
+      nub $ ty : optTargetTypes opts }))
 
 verbDescriptions :: OptErrTable Options
 verbDescriptions = map toDescr verbosities
@@ -564,11 +591,27 @@ debugDescriptions =
     = (name , "dump code after " ++ desc
         , \ opts -> opts { dbDumpLevels = addFlag flag (dbDumpLevels opts)})
 
+optimizeDescriptions :: OptErrTable OptimizationOpts
+optimizeDescriptions =
+  [ ( "desugar-newtypes"        , "desugars newtypes in FlatCurry"
+    , \ opts -> opts { optDesugarNewtypes     = True    })
+  , ( "inline-dictionaries"     , "inlines type class dictionaries"
+    , \ opts -> opts { optInlineDictionaries  = True    })
+  , ( "remove-unused-imports"   , "removes unused imports"
+    , \ opts -> opts { optRemoveUnusedImports = True    })
+  , ( "no-desugar-newtypes"     , "prevents desugaring of newtypes in FlatCurry"
+    , \ opts -> opts { optDesugarNewtypes     = False   })
+  , ( "no-inline-dictionaries"  , "prevents inlining of type class dictionaries"
+    , \ opts -> opts { optInlineDictionaries  = False   })
+  , ( "no-remove-unused-imports", "prevents removing of unused imports"
+    , \ opts -> opts { optRemoveUnusedImports = False   })
+  ]
+
 addFlag :: Eq a => a -> [a] -> [a]
 addFlag o opts = nub $ o : opts
 
 removeFlag :: Eq a => a -> [a] -> [a]
-removeFlag o opts = filter (/= o) opts
+removeFlag o = filter (/= o)
 
 -- |Update the 'Options' record by the parsed and processed arguments
 updateOpts :: Options -> [String] -> (Options, [String], [String])
