@@ -63,7 +63,7 @@ genAbstractCurry uacy env mdl
 -- ---------------------------------------------------------------------------
 
 trModule :: Module PredType -> GAC CurryProg
-trModule (Module _ _ mid _ is ds) =
+trModule (Module _ _ _ mid _ is ds) =
   CurryProg mid' is' <$> dflt' <*> cds' <*> ids' <*> ts' <*> fs' <*> os'
   where
   mid'  = moduleName mid
@@ -84,7 +84,7 @@ trDefaultDecl (DefaultDecl _ tys) = (\tys' -> [CDefaultDecl tys'])
 trDefaultDecl _                   = return []
 
 trClassDecl :: Decl PredType -> GAC [CClassDecl]
-trClassDecl (ClassDecl _ cx cls tv ds) =
+trClassDecl (ClassDecl _ _ cx cls tv ds) =
   (\cls' v' cx' tv' ds' -> [CClass cls' v' cx' tv' ds'])
     <$> trGlobalIdent cls <*> getTypeVisibility cls <*> trContext cx
     <*> getTVarIndex tv <*> concatMapM (trClassMethodDecl sigs fs) ds
@@ -111,7 +111,7 @@ trClassMethodDecl sigs _ (FunctionDecl _ _ f eqs) =
 trClassMethodDecl _ _ _ = return []
 
 trInstanceDecl :: Decl PredType -> GAC [CInstanceDecl]
-trInstanceDecl (InstanceDecl _ cx qcls ty ds) =
+trInstanceDecl (InstanceDecl _ _ cx qcls ty ds) =
   (\qcls' cx' ty' ds' -> [CInstance qcls' cx' ty' ds']) <$> trQual qcls
   <*> trContext cx <*> trTypeExpr ty <*> mapM (trInstanceMethodDecl qcls ty) ds
 trInstanceDecl _ = return []
@@ -158,23 +158,21 @@ trTypeDecl (ExternalDataDecl _ t vs) =
 trTypeDecl _                       = return []
 
 trConsDecl :: ConstrDecl -> GAC CConsDecl
-trConsDecl (ConstrDecl  _ vs cx c tys) = inNestedTScope $ CCons
-  <$> mapM genTVarIndex vs <*> trContext cx <*> trGlobalIdent c
-  <*> getVisibility c <*> mapM trTypeExpr tys
-trConsDecl (ConOpDecl p vs cx ty1 op ty2) = inNestedTScope $ trConsDecl $
-  ConstrDecl p vs cx op [ty1, ty2]
-trConsDecl (RecordDecl   _ vs cx c fs) = inNestedTScope $ CRecord
-  <$> mapM genTVarIndex vs <*> trContext cx <*> trGlobalIdent c
-  <*> getVisibility c <*> concatMapM trFieldDecl fs
+trConsDecl (ConstrDecl  _ c tys) = inNestedTScope $ CCons
+  <$> trGlobalIdent c <*> getVisibility c <*> mapM trTypeExpr tys
+trConsDecl (ConOpDecl p ty1 op ty2) = inNestedTScope $ trConsDecl $
+  ConstrDecl p op [ty1, ty2]
+trConsDecl (RecordDecl   _ c fs) = inNestedTScope $ CRecord
+  <$> trGlobalIdent c <*> getVisibility c <*> concatMapM trFieldDecl fs
 
 trFieldDecl :: FieldDecl -> GAC [CFieldDecl]
 trFieldDecl (FieldDecl _ ls ty) = T.forM ls $ \l ->
   CField <$> trGlobalIdent l <*> getVisibility l <*> trTypeExpr ty
 
 trNewConsDecl :: NewConstrDecl -> GAC CConsDecl
-trNewConsDecl (NewConstrDecl _ nc      ty) = CCons [] (CContext [])
+trNewConsDecl (NewConstrDecl _ nc      ty) = CCons
   <$> trGlobalIdent nc <*> getVisibility nc <*> ((:[]) <$> trTypeExpr ty)
-trNewConsDecl (NewRecordDecl p nc (l, ty)) = CRecord [] (CContext [])
+trNewConsDecl (NewRecordDecl p nc (l, ty)) = CRecord
   <$> trGlobalIdent nc <*> getVisibility nc <*> trFieldDecl (FieldDecl p [l] ty)
 
 trTypeExpr :: TypeExpr -> GAC CTypeExpr
@@ -232,10 +230,10 @@ trLhs :: Lhs a -> GAC [CPattern]
 trLhs = mapM trPat . snd . flatLhs
 
 trRhs :: Rhs PredType -> GAC CRhs
-trRhs (SimpleRhs _ e ds) = inNestedScope $ do
+trRhs (SimpleRhs _ _ e ds) = inNestedScope $ do
   mapM_ insertDeclLhs ds
   CSimpleRhs <$> trExpr e <*> concatMapM trLocalDecl ds
-trRhs (GuardedRhs _ gs ds) = inNestedScope $ do
+trRhs (GuardedRhs _ _ gs ds) = inNestedScope $ do
   mapM_ insertDeclLhs ds
   CGuardedRhs <$> mapM trCondExpr gs <*> concatMapM trLocalDecl ds
 
@@ -312,14 +310,14 @@ trExpr (RightSection       _ op e) =
   trExpr $ apply (Variable NoSpanInfo undefined qFlip) [infixOp op, e]
 trExpr (Lambda             _ ps e) = inNestedScope $
                                      CLambda <$> mapM trPat ps <*> trExpr e
-trExpr (Let                _ ds e) = inNestedScope $
+trExpr (Let              _ _ ds e) = inNestedScope $
                                      CLetDecl <$> trLocalDecls ds <*> trExpr e
-trExpr (Do                 _ ss e) = inNestedScope $
+trExpr (Do               _ _ ss e) = inNestedScope $
                                      (\ss' e' -> CDoExpr (ss' ++ [CSExpr e']))
                                      <$> mapM trStatement ss <*> trExpr e
 trExpr (IfThenElse     _ e1 e2 e3) =
   trExpr $ apply (Variable NoSpanInfo undefined qIfThenElseId) [e1, e2, e3]
-trExpr (Case            _ ct e bs) = CCase (cvCaseType ct)
+trExpr (Case          _ _ ct e bs) = CCase (cvCaseType ct)
                                      <$> trExpr e <*> mapM trAlt bs
 
 cvCaseType :: CaseType -> CCaseType
@@ -327,9 +325,9 @@ cvCaseType Flex  = CFlex
 cvCaseType Rigid = CRigid
 
 trStatement :: Statement PredType -> GAC CStatement
-trStatement (StmtExpr _   e) = CSExpr     <$> trExpr e
-trStatement (StmtDecl _  ds) = CSLet      <$> trLocalDecls ds
-trStatement (StmtBind _ p e) = flip CSPat <$> trExpr e <*> trPat p
+trStatement (StmtExpr _   e)  = CSExpr     <$> trExpr e
+trStatement (StmtDecl _ _ ds) = CSLet      <$> trLocalDecls ds
+trStatement (StmtBind _ p e)  = flip CSPat <$> trExpr e <*> trPat p
 
 trAlt :: Alt PredType -> GAC (CPattern, CRhs)
 trAlt (Alt _ p rhs) = inNestedScope $ (,) <$> trPat p <*> trRhs rhs
@@ -392,7 +390,7 @@ qNegateId :: QualIdent
 qNegateId = qualifyWith preludeMIdent (mkIdent "negate")
 
 qIfThenElseId :: QualIdent
-qIfThenElseId = qualifyWith preludeMIdent (mkIdent "if_then_else")
+qIfThenElseId = qualifyWith preludeMIdent (mkIdent "ifThenElse")
 
 prelUntyped :: QualIdent
 prelUntyped = qualifyWith preludeMIdent $ mkIdent "untyped"
@@ -418,7 +416,7 @@ data AbstractEnv = AbstractEnv
 
 -- |Initialize the AbstractCurry generator environment
 abstractEnv :: Bool -> CompilerEnv -> Module a -> AbstractEnv
-abstractEnv uacy env (Module _ _ mid es _ ds) = AbstractEnv
+abstractEnv uacy env (Module _ _ _ mid es _ ds) = AbstractEnv
   { moduleId   = mid
   , typeEnv    = valueEnv env
   , tyExports  = foldr (buildTypeExports  mid) Set.empty es'
