@@ -17,7 +17,7 @@
 -}
 module Imports (importInterfaces, importModules, qualifyEnv) where
 
-import           Data.List                  (nubBy)
+import           Data.List                  (nubBy, find)
 import qualified Data.Map            as Map
 import           Data.Maybe                 (catMaybes, fromMaybe, isJust)
 import qualified Data.Set            as Set
@@ -28,8 +28,8 @@ import Curry.Base.Monad
 import Curry.Syntax
 
 import Base.CurryKinds (toKind')
-import Base.CurryTypes ( toQualType, toQualTypes, toQualPredSet, toQualPredType
-                       , toConstrType, toMethodType )
+import Base.CurryTypes ( toQualType, toQualTypes, toQualPredType, toConstrType
+                       , toMethodType )
 
 import Base.Kinds
 import Base.Messages
@@ -49,7 +49,7 @@ import CompilerEnv
 
 importModules :: Monad m => Module a -> InterfaceEnv -> [ImportDecl]
               -> CYT m CompilerEnv
-importModules mdl@(Module _ _ mid _ _ _) iEnv expImps
+importModules mdl@(Module _ _ _ mid _ _ _) iEnv expImps
   = ok $ foldl importModule initEnv expImps
   where
     initEnv = (initCompilerEnv mid)
@@ -164,10 +164,11 @@ importClasses m = flip $ foldr (bindClass m)
 bindClass :: ModuleIdent -> IDecl -> ClassEnv -> ClassEnv
 bindClass m (HidingClassDecl p cx cls k tv) =
   bindClass m (IClassDecl p cx cls k tv [] [])
-bindClass m (IClassDecl _ cx cls _ _ ds _) =
+bindClass m (IClassDecl _ cx cls _ _ ds ids) =
   bindClassInfo (qualQualify m cls) (sclss, ms)
   where sclss = map (\(Constraint _ scls _) -> qualQualify m scls) cx
-        ms = map (\d -> (imethod d, isJust $ imethodArity d)) ds
+        ms = map (\d -> (imethod d, isJust $ imethodArity d)) $ filter isVis ds
+        isVis (IMethodDecl _ idt _ _ ) = idt `notElem` ids
 bindClass _ _ = id
 
 importInstances :: ModuleIdent -> [IDecl] -> InstEnv -> InstEnv
@@ -201,33 +202,28 @@ types :: ModuleIdent -> IDecl -> [TypeInfo]
 types m (IDataDecl _ tc k tvs cs _) =
   [typeCon DataType m tc k tvs (map mkData cs)]
   where
-    mkData (ConstrDecl _ evs cx c tys) =
-      DataConstr c (length evs) (toQualPredSet m tvs' cx)
-        (toQualTypes m tvs' tys)
-      where tvs' = tvs ++ evs
-    mkData (ConOpDecl _ evs cx ty1 c ty2) =
-      DataConstr c (length evs) (toQualPredSet m tvs' cx)
-        (toQualTypes m tvs' [ty1, ty2])
-      where tvs' = tvs ++ evs
-    mkData (RecordDecl _ evs cx c fs) =
-      RecordConstr c (length evs) (toQualPredSet m tvs' cx) labels
-        (toQualTypes m tvs' tys)
-      where tvs' = tvs ++ evs
-            (labels, tys) = unzip [(l, ty) | FieldDecl _ ls ty <- fs, l <- ls]
+    mkData (ConstrDecl _ c tys) =
+      DataConstr c (toQualTypes m tvs tys)
+    mkData (ConOpDecl _  ty1 c ty2) =
+      DataConstr c (toQualTypes m tvs [ty1, ty2])
+    mkData (RecordDecl _ c fs) =
+      RecordConstr c labels (toQualTypes m tvs tys)
+      where (labels, tys) = unzip [(l, ty) | FieldDecl _ ls ty <- fs, l <- ls]
 types m (INewtypeDecl _ tc k tvs nc _) =
   [typeCon RenamingType m tc k tvs (mkData nc)]
   where
     mkData (NewConstrDecl _ c ty) =
-      DataConstr c 0 emptyPredSet [toQualType m tvs ty]
+      DataConstr c [toQualType m tvs ty]
     mkData (NewRecordDecl _ c (l, ty)) =
-      RecordConstr c 0 emptyPredSet [l] [toQualType m tvs ty]
+      RecordConstr c [l] [toQualType m tvs ty]
 types m (ITypeDecl _ tc k tvs ty) =
   [typeCon aliasType m tc k tvs (toQualType m tvs ty)]
   where
     aliasType tc' k' = AliasType tc' k' (length tvs)
-types m (IClassDecl _ _ qcls k tv ds _) =
-  [typeCls m qcls k (map mkMethod ds)]
+types m (IClassDecl _ _ qcls k tv ds ids) =
+  [typeCls m qcls k (map mkMethod $ filter isVis ds)]
   where
+    isVis (IMethodDecl _ f _ _ ) = f `notElem` ids
     mkMethod (IMethodDecl _ f a qty) = ClassMethod f a $
       qualifyPredType m $ normalize 1 $ toMethodType qcls tv qty
 types _ _ = []
@@ -251,7 +247,7 @@ values m (IDataDecl _ tc _ tvs cs hs) =
   map (recLabel m tc' tvs ty') (nubBy sameLabel clabels)
   where tc' = qualQualify m tc
         ty' = constrType tc' tvs
-        labels   = [ (l, lty) | RecordDecl _ _ _ _ fs <- cs
+        labels   = [ (l, lty) | RecordDecl _ _ fs <- cs
                    , FieldDecl _ ls lty <- fs, l <- ls, l `notElem` hs
                    ]
         clabels  = [(l, constr l, ty) | (l, ty) <- labels]
@@ -269,26 +265,32 @@ values m (INewtypeDecl _ tc _ tvs nc hs) =
   where tc' = qualQualify m tc
         ty' = constrType tc' tvs
 values m (IFunctionDecl _ f Nothing a qty) =
-  [Value (qualQualify m f) False a (typeScheme (toQualPredType m [] qty))]
+  [Value (qualQualify m f) Nothing a (typeScheme (toQualPredType m [] qty))]
 values m (IFunctionDecl _ f (Just tv) _ qty) =
-  [Value (qualQualify m f) True 0 (typeScheme (toQualPredType m [tv] qty))]
+  let mcls = case qty of
+        QualTypeExpr _ ctx _ -> fmap (\(Constraint _ qcls _) -> qcls) $
+                                find (\(Constraint _ _ ty) -> isVar ty) ctx
+  in [Value (qualQualify m f) mcls 0 (typeScheme (toQualPredType m [tv] qty))]
+  where
+    isVar (VariableType _ i) = i == tv
+    isVar _                  = False
 values m (IClassDecl _ _ qcls _ tv ds hs) =
-  map (classMethod m qcls' tv) (filter ((`notElem` hs) . imethod) ds)
+  map (classMethod m qcls' tv hs) ds
   where qcls' = qualQualify m qcls
 values _ _                        = []
 
 dataConstr :: ModuleIdent -> QualIdent -> [Ident] -> ConstrDecl -> ValueInfo
-dataConstr m tc tvs (ConstrDecl _ evs cx c tys) =
+dataConstr m tc tvs (ConstrDecl _ c tys) =
   DataConstructor (qualifyLike tc c) a labels $
-    constrType' m tc tvs evs cx tys
+    constrType' m tc tvs tys
   where a      = length tys
         labels = replicate a anonId
-dataConstr m tc tvs (ConOpDecl _ evs cx ty1 op ty2) =
+dataConstr m tc tvs (ConOpDecl _ ty1 op ty2) =
   DataConstructor (qualifyLike tc op) 2 [anonId, anonId] $
-    constrType' m tc tvs evs cx [ty1, ty2]
-dataConstr m tc tvs (RecordDecl _ evs cx c fs) =
+    constrType' m tc tvs [ty1, ty2]
+dataConstr m tc tvs (RecordDecl _ c fs) =
   DataConstructor (qualifyLike tc c) a labels $
-    constrType' m tc tvs evs cx tys
+    constrType' m tc tvs tys
   where fields        = [(l, ty) | FieldDecl _ ls ty <- fs, l <- ls]
         (labels, tys) = unzip fields
         a             = length labels
@@ -296,10 +298,10 @@ dataConstr m tc tvs (RecordDecl _ evs cx c fs) =
 newConstr :: ModuleIdent -> QualIdent -> [Ident] -> NewConstrDecl -> ValueInfo
 newConstr m tc tvs (NewConstrDecl _ c ty1) =
   NewtypeConstructor (qualifyLike tc c) anonId $
-  constrType' m tc tvs [] [] [ty1]
+  constrType' m tc tvs [ty1]
 newConstr m tc tvs (NewRecordDecl _ c (l, ty1)) =
   NewtypeConstructor (qualifyLike tc c) l $
-  constrType' m tc tvs [] [] [ty1]
+  constrType' m tc tvs [ty1]
 
 recLabel :: ModuleIdent -> QualIdent -> [Ident] -> TypeExpr
            -> (Ident, [Ident], TypeExpr) -> ValueInfo
@@ -308,11 +310,9 @@ recLabel m tc tvs ty0 (l, cs, lty) = Label ql qcs tySc
         qcs  = map (qualifyLike tc) cs
         tySc = polyType (toQualType m tvs (ArrowType NoSpanInfo ty0 lty))
 
-constrType' :: ModuleIdent -> QualIdent -> [Ident] -> [Ident] -> Context
-            -> [TypeExpr] -> ExistTypeScheme
-constrType' m tc tvs evs cx tys = ForAllExist (length tvs) (length evs) pty
-  where tvs' = tvs ++ evs
-        pty  = qualifyPredType m $ toConstrType tc tvs' cx tys
+constrType' :: ModuleIdent -> QualIdent -> [Ident] -> [TypeExpr] -> TypeScheme
+constrType' m tc tvs tys = ForAll (length tvs) pty
+  where pty  = qualifyPredType m $ toConstrType tc tvs tys
 
 constrType :: QualIdent -> [Ident] -> TypeExpr
 constrType tc tvs = foldl (ApplyType NoSpanInfo) (ConstructorType NoSpanInfo tc)
@@ -321,10 +321,13 @@ constrType tc tvs = foldl (ApplyType NoSpanInfo) (ConstructorType NoSpanInfo tc)
 -- We always enter class methods with an arity of 0 into the value environment
 -- because there may be different implementations with different arities.
 
-classMethod :: ModuleIdent -> QualIdent -> Ident -> IMethodDecl -> ValueInfo
-classMethod m qcls tv (IMethodDecl _ f _ qty) =
-  Value (qualifyLike qcls f) True 0 $
+classMethod :: ModuleIdent -> QualIdent -> Ident -> [Ident] -> IMethodDecl
+            -> ValueInfo
+classMethod m qcls tv hs (IMethodDecl _ f _ qty) =
+  Value (qualifyLike qcls f) mcls 0 $
     typeScheme $ qualifyPredType m $ toMethodType qcls tv qty
+  where
+    mcls = if f `elem` hs then Nothing else Just qcls
 
 -- ---------------------------------------------------------------------------
 
