@@ -55,7 +55,7 @@ import Env.Interface
 import qualified Curry.AbstractCurry as AC
 import qualified Curry.FlatCurry     as FC
 import qualified Curry.Syntax        as CS
-import qualified IL                  as IL
+import qualified IL
 
 import Checks
 import CompilerEnv
@@ -93,11 +93,10 @@ compileModule opts m fn = do
   writeParsed   opts mdl
   let qmdl = qual mdl
   writeHtml     opts qmdl
-  let umdl = (fst qmdl, fmap (const ()) (snd qmdl))
-  writeAST      opts umdl
-  writeShortAST opts umdl
+  writeAST      opts (fst  mdl, fmap (const ()) (snd  mdl))
+  writeShortAST opts (fst qmdl, fmap (const ()) (snd qmdl))
   mdl' <- expandExports opts mdl
-  qmdl' <- dumpWith opts CS.showModule CS.ppModule DumpQualified $ qual mdl'
+  qmdl' <- dumpWith opts CS.showModule pPrint DumpQualified $ qual mdl'
   writeAbstractCurry opts qmdl'
   -- generate interface file
   let intf = uncurry exportInterface qmdl'
@@ -106,7 +105,10 @@ compileModule opts m fn = do
     ((env, il), mdl'') <- transModule opts qmdl'
     writeFlat opts env (snd mdl'') il
   where
-  withFlat = any (`elem` optTargetTypes opts) [TypedFlatCurry, FlatCurry]
+  withFlat = any (`elem` optTargetTypes opts) [ AnnotatedFlatCurry
+                                              , TypedFlatCurry
+                                              , FlatCurry
+                                              ]
 
 loadAndCheckModule :: Options -> ModuleIdent -> FilePath
                    -> CYIO (CompEnv (CS.Module PredType))
@@ -125,13 +127,14 @@ loadModule opts m fn = do
   -- parse and check module header
   (toks, mdl) <- parseModule opts m fn
   -- load the imported interfaces into an InterfaceEnv
-  let paths = map (addCurrySubdir (optUseSubdir opts))
+  let paths = map (addOutDir (optUseOutDir opts) (optOutDir opts))
                   ("." : optImportPaths opts)
-  iEnv   <- loadInterfaces paths mdl
+  let withPrel = importPrelude opts mdl
+  iEnv   <- loadInterfaces paths withPrel
   checkInterfaces opts iEnv
-  is     <- importSyntaxCheck iEnv mdl
+  is     <- importSyntaxCheck iEnv withPrel
   -- add information of imported modules
-  cEnv   <- importModules mdl iEnv is
+  cEnv   <- importModules withPrel iEnv is
   return (cEnv { filePath = fn, tokens = toks }, mdl)
 
 parseModule :: Options -> ModuleIdent -> FilePath
@@ -150,7 +153,7 @@ parseModule opts m fn = do
       -- they will be issued a second time during parsing.
       spanToks <- liftCYM $ silent $ CS.lexSource fn condC
       ast      <- liftCYM $ CS.parseModule fn condC
-      checked  <- checkModuleHeader opts m fn ast
+      checked  <- checkModuleHeader m fn ast
       return (spanToks, checked)
 
 preprocess :: PrepOpts -> FilePath -> String -> CYIO String
@@ -179,33 +182,33 @@ withTempFile act = do
   removeFile fn
   return res
 
-checkModuleHeader :: Monad m => Options -> ModuleIdent -> FilePath
+checkModuleHeader :: Monad m => ModuleIdent -> FilePath
                   -> CS.Module () -> CYT m (CS.Module ())
-checkModuleHeader opts m fn = checkModuleId m
-                            . importPrelude opts
-                            . CS.patchModuleId fn
+checkModuleHeader m fn = checkModuleId m
+                       . CS.patchModuleId fn
 
 -- |Check whether the 'ModuleIdent' and the 'FilePath' fit together
 checkModuleId :: Monad m => ModuleIdent -> CS.Module () -> CYT m (CS.Module ())
-checkModuleId mid m@(CS.Module _ _ mid' _ _ _)
+checkModuleId mid m@(CS.Module _ _ _ mid' _ _ _)
   | mid == mid' = ok m
   | otherwise   = failMessages [errModuleFileMismatch mid']
 
--- An implicit import of the prelude is added to the declarations of
--- every module, except for the prelude itself, or when the import is disabled
--- by a compiler option. If no explicit import for the prelude is present,
--- the prelude is imported unqualified, otherwise a qualified import is added.
+-- An implicit import of the prelude is temporariliy added to the declarations
+-- of every module, except for the prelude itself, or when the import is
+-- disabled by a compiler option. If no explicit import for the prelude is
+-- present, the prelude is imported unqualified,
+-- otherwise a qualified import is added.
 
 importPrelude :: Options -> CS.Module () -> CS.Module ()
-importPrelude opts m@(CS.Module spi ps mid es is ds)
+importPrelude opts m@(CS.Module spi li ps mid es is ds)
     -- the Prelude itself
-  | mid == preludeMIdent          = m
+  | mid == preludeMIdent         = m
     -- disabled by compiler option
-  | noImpPrelude                  = m
+  | noImpPrelude                 = m
     -- already imported
-  | preludeMIdent `elem` imported = m
+  | preludeMIdent `elem` imports = m
     -- let's add it!
-  | otherwise                     = CS.Module spi ps mid es (preludeImp : is) ds
+  | otherwise                    = CS.Module spi li ps mid es (preludeImp:is) ds
   where
   noImpPrelude = NoImplicitPrelude `elem` optExtensions opts
                  || m `CS.hasLanguageExtension` NoImplicitPrelude
@@ -213,7 +216,7 @@ importPrelude opts m@(CS.Module spi ps mid es is ds)
                   False   -- qualified?
                   Nothing -- no alias
                   Nothing -- no selection of types, functions, etc.
-  imported     = [imp | (CS.ImportDecl _ imp _ _ _) <- is]
+  imports      = [imp | (CS.ImportDecl _ imp _ _ _) <- is]
 
 checkInterfaces :: Monad m => Options -> InterfaceEnv -> CYT m ()
 checkInterfaces opts iEnv = mapM_ checkInterface (Map.elems iEnv)
@@ -223,7 +226,7 @@ checkInterfaces opts iEnv = mapM_ checkInterface (Map.elems iEnv)
     interfaceCheck opts (env, intf)
 
 importSyntaxCheck :: Monad m => InterfaceEnv -> CS.Module a -> CYT m [CS.ImportDecl]
-importSyntaxCheck iEnv (CS.Module _ _ _ _ imps _) = mapM checkImportDecl imps
+importSyntaxCheck iEnv (CS.Module _ _ _ _ _ imps _) = mapM checkImportDecl imps
   where
   checkImportDecl (CS.ImportDecl p m q asM is) = case Map.lookup m iEnv of
     Just intf -> CS.ImportDecl p m q asM `liftM` importCheck intf is
@@ -252,7 +255,7 @@ checkModule opts mdl = do
   where
   dumpCS :: (MonadIO m, Show a) => DumpLevel -> CompEnv (CS.Module a)
          -> m (CompEnv (CS.Module a))
-  dumpCS = dumpWith opts CS.showModule CS.ppModule
+  dumpCS = dumpWith opts CS.showModule pPrint
 
 -- ---------------------------------------------------------------------------
 -- Translating a module
@@ -261,19 +264,23 @@ checkModule opts mdl = do
 transModule :: Options -> CompEnv (CS.Module PredType)
             -> CYIO (CompEnv IL.Module, CompEnv (CS.Module Type))
 transModule opts mdl = do
-  derived    <- dumpCS DumpDerived       $ derive         mdl
-  desugared  <- dumpCS DumpDesugared     $ desugar        derived
-  dicts      <- dumpCS DumpDictionaries  $ insertDicts    desugared
-  newtypes   <- dumpCS DumpNewtypes      $ removeNewtypes dicts
-  simplified <- dumpCS DumpSimplified    $ simplify       newtypes
-  lifted     <- dumpCS DumpLifted        $ lift           simplified
-  il         <- dumpIL DumpTranslated    $ ilTrans        lifted
-  ilCaseComp <- dumpIL DumpCaseCompleted $ completeCase   il
+  derived    <- dumpCS DumpDerived       $ derive               mdl
+  desugared  <- dumpCS DumpDesugared     $ desugar              derived
+  dicts      <- dumpCS DumpDictionaries  $ insertDicts    inlDi desugared
+  newtypes   <- dumpCS DumpNewtypes      $ removeNewtypes remNT dicts
+  simplified <- dumpCS DumpSimplified    $ simplify             newtypes
+  lifted     <- dumpCS DumpLifted        $ lift                 simplified
+  il         <- dumpIL DumpTranslated    $ ilTrans        remIm lifted
+  ilCaseComp <- dumpIL DumpCaseCompleted $ completeCase         il
   return (ilCaseComp, newtypes)
   where
+  optOpts = optOptimizations opts
+  inlDi = optInlineDictionaries  optOpts
+  remIm = optRemoveUnusedImports optOpts
+  remNT = optDesugarNewtypes     optOpts
   dumpCS :: Show a => DumpLevel -> CompEnv (CS.Module a)
          -> CYIO (CompEnv (CS.Module a))
-  dumpCS = dumpWith opts CS.showModule CS.ppModule
+  dumpCS = dumpWith opts CS.showModule pPrint
   dumpIL = dumpWith opts IL.showModule IL.ppModule
 
 -- ---------------------------------------------------------------------------
@@ -293,7 +300,7 @@ writeTokens opts env = when tokTarget $ liftIO $
               (showTokenStream (tokens env))
   where
   tokTarget  = Tokens `elem` optTargetTypes opts
-  useSubDir  = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 writeComments :: Options -> CompilerEnv -> CYIO ()
 writeComments opts env = when tokTarget $ liftIO $
@@ -301,7 +308,7 @@ writeComments opts env = when tokTarget $ liftIO $
               (showCommentTokenStream $ tokens env)
   where
   tokTarget  = Comments `elem` optTargetTypes opts
-  useSubDir  = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 -- |Output the parsed 'Module' on request
 writeParsed :: Show a => Options -> CompEnv (CS.Module a) -> CYIO ()
@@ -309,7 +316,7 @@ writeParsed opts (env, mdl) = when srcTarget $ liftIO $
   writeModule (useSubDir $ sourceRepName (filePath env)) (CS.showModule mdl)
   where
   srcTarget  = Parsed `elem` optTargetTypes opts
-  useSubDir  = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 writeHtml :: Options -> CompEnv (CS.Module a) -> CYIO ()
 writeHtml opts (env, mdl) = when htmlTarget $
@@ -329,8 +336,8 @@ writeInterface opts env intf@(CS.Interface m _ _)
 
   interfaceFile   = interfName (filePath env)
   outputInterface = liftIO $ writeModule
-                    (addCurrySubdirModule (optUseSubdir opts) m interfaceFile)
-                    (show $ CS.ppInterface intf)
+                    (addOutDirModule (optUseOutDir opts) (optOutDir opts) m interfaceFile)
+                    (show $ pPrint intf)
 
 matchInterface :: FilePath -> CS.Interface -> IO Bool
 matchInterface ifn i = do
@@ -342,24 +349,24 @@ matchInterface ifn i = do
 
 writeFlat :: Options -> CompilerEnv -> CS.Module Type -> IL.Module -> CYIO ()
 writeFlat opts env mdl il = do
-  (_, tfc) <- dumpWith opts show (FC.ppProg . genFlatCurry) DumpTypedFlatCurry (env, tfcyProg)
-  when tfcyTarget  $ liftIO $ FC.writeFlatCurry (useSubDir tfcyName) tafcyProg
-  when tafcyTarget $ liftIO $ FC.writeFlatCurry (useSubDir tafcyName) tfc
+  _ <- dumpWith opts show (pPrint . genFlatCurry) DumpTypedFlatCurry (env, afcy)
+  when afcyTarget $ liftIO $ FC.writeFlatCurry (useSubDir afcyName) afcy
+  when tfcyTarget  $ liftIO $ FC.writeFlatCurry (useSubDir tfcyName)  tfcy
   when fcyTarget $ do
-    (_, fc) <- dumpWith opts show FC.ppProg DumpFlatCurry (env, fcyProg)
-    liftIO $ FC.writeFlatCurry (useSubDir fcyName) fc
-  writeFlatIntf opts env fcyProg
+    _ <- dumpWith opts show pPrint DumpFlatCurry (env, fcy)
+    liftIO $ FC.writeFlatCurry (useSubDir fcyName) fcy
+  writeFlatIntf opts env fcy
   where
-  tfcyName    = typedFlatName (filePath env)
-  tfcyProg    = genTypedFlatCurry env mdl il
-  tfcyTarget  = TypedFlatCurry `elem` optTargetTypes opts
-  tafcyName   = typeAnnFlatName (filePath env)
-  tafcyProg   = genTypeAnnotatedFlatCurry env mdl il
-  tafcyTarget = TypeAnnotatedFlatCurry `elem` optTargetTypes opts
-  fcyName     = flatName (filePath env)
-  fcyProg     = genFlatCurry tfcyProg
-  fcyTarget   = FlatCurry `elem` optTargetTypes opts
-  useSubDir   = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  afcy       = genAnnotatedFlatCurry env mdl il
+  afcyName   = annotatedFlatName (filePath env)
+  afcyTarget = AnnotatedFlatCurry `elem` optTargetTypes opts
+  tfcy       = genTypedFlatCurry afcy
+  tfcyName   = typedFlatName (filePath env)
+  tfcyTarget = TypedFlatCurry `elem` optTargetTypes opts
+  fcy        = genFlatCurry afcy
+  fcyName    = flatName (filePath env)
+  fcyTarget  = FlatCurry `elem` optTargetTypes opts
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 writeFlatIntf :: Options -> CompilerEnv -> FC.Prog -> CYIO ()
 writeFlatIntf opts env prog
@@ -369,12 +376,12 @@ writeFlatIntf opts env prog
       mfint <- liftIO $ FC.readFlatInterface targetFile
       let oldInterface = fromMaybe emptyIntf mfint
       when (mfint == mfint) $ return () -- necessary to close file -- TODO
-      unless (oldInterface `eqInterface` fint) $ outputInterface
+      unless (oldInterface `eqInterface` fint) outputInterface
   where
   targetFile      = flatIntName (filePath env)
   emptyIntf       = FC.Prog "" [] [] [] []
   fint            = genFlatInterface prog
-  useSubDir       = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir       = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
   outputInterface = liftIO $ FC.writeFlatCurry (useSubDir targetFile) fint
 
 writeAbstractCurry :: Options -> CompEnv (CS.Module PredType) -> CYIO ()
@@ -388,7 +395,7 @@ writeAbstractCurry opts (env, mdl) = do
   where
   acyTarget  = AbstractCurry        `elem` optTargetTypes opts
   uacyTarget = UntypedAbstractCurry `elem` optTargetTypes opts
-  useSubDir  = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 
 writeAST :: Options -> CompEnv (CS.Module ()) -> CYIO ()
@@ -396,7 +403,7 @@ writeAST opts (env, mdl) = when astTarget $ liftIO $
   writeModule (useSubDir $ astName (filePath env)) (CS.showModule mdl)
   where
   astTarget  = AST `elem` optTargetTypes opts
-  useSubDir  = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 
 writeShortAST :: Options -> CompEnv (CS.Module ()) -> CYIO ()
@@ -405,7 +412,7 @@ writeShortAST opts (env, mdl) = when astTarget $ liftIO $
               (CS.showModule $ shortenModuleAST mdl)
   where
   astTarget  = ShortAST `elem` optTargetTypes opts
-  useSubDir  = addCurrySubdirModule (optUseSubdir opts) (moduleIdent env)
+  useSubDir  = addOutDirModule (optUseOutDir opts) (optOutDir opts) (moduleIdent env)
 
 
 type Dump = (DumpLevel, CompilerEnv, String)
