@@ -30,27 +30,33 @@ import           Control.Applicative        ((<$>), (<*>))
 
 import           Control.Monad.Extra         (concatMapM)
 import qualified Control.Monad.Reader as R
+import qualified Control.Monad.State  as S
 import           Data.List                   (nub, partition)
-import qualified Data.Map             as Map (Map, empty, insert, lookup)
+import           Data.Maybe                  (fromJust)
+import qualified Data.Map             as Map
 import qualified Data.Set             as Set (Set, empty, insert, delete, toList)
 
 import Curry.Base.Ident
 import Curry.Syntax hiding (caseAlt)
 
-import Base.CurryTypes (toType)
 import Base.Expr
 import Base.Messages (internalError)
-import Base.Types
+import Base.Types hiding (polyType)
+import Base.Kinds
 import Base.Typing
 import Base.Utils (foldr2)
 
+import Env.TypeConstructor
 import Env.Value (ValueEnv, ValueInfo (..), qualLookupValue)
 
 import qualified IL as IL
 
-ilTrans :: ValueEnv -> Module Type -> IL.Module
-ilTrans vEnv (Module _ _ m _ _ ds) = IL.Module m (imports m ds') ds'
-  where ds' = R.runReader (concatMapM trDecl ds) (TransEnv m vEnv)
+ilTrans :: Bool -> ValueEnv -> TCEnv -> Module Type -> IL.Module
+ilTrans remIm vEnv tcEnv (Module _ _ _ m _ im ds) = IL.Module m im' ds'
+  where ds' = R.runReader (concatMapM trDecl ds) (TransEnv m vEnv tcEnv)
+        im' = preludeMIdent : if remIm then imports m ds' else map moduleImport im
+        moduleImport (ImportDecl _ mdl _ _ _) = mdl
+
 
 -- -----------------------------------------------------------------------------
 -- Computation of necessary imports
@@ -65,9 +71,11 @@ imports m = Set.toList . Set.delete m . foldr mdlsDecl Set.empty
 mdlsDecl :: IL.Decl -> Set.Set ModuleIdent -> Set.Set ModuleIdent
 mdlsDecl (IL.DataDecl       _ _ cs) ms = foldr mdlsConstrsDecl ms cs
   where mdlsConstrsDecl (IL.ConstrDecl _ tys) ms' = foldr mdlsType ms' tys
+mdlsDecl (IL.NewtypeDecl    _ _ nc) ms = mdlsNewConstrDecl nc
+  where mdlsNewConstrDecl (IL.NewConstrDecl _ ty) = mdlsType ty ms
 mdlsDecl (IL.ExternalDataDecl  _ _) ms = ms
 mdlsDecl (IL.FunctionDecl _ _ ty e) ms = mdlsType ty (mdlsExpr e ms)
-mdlsDecl (IL.ExternalDecl     _ ty) ms = mdlsType ty ms
+mdlsDecl (IL.ExternalDecl   _ _ ty) ms = mdlsType ty ms
 
 mdlsType :: IL.Type -> Set.Set ModuleIdent -> Set.Set ModuleIdent
 mdlsType (IL.TypeConstructor tc tys) ms = modules tc (foldr mdlsType ms tys)
@@ -103,6 +111,7 @@ modules x ms = maybe ms (`Set.insert` ms) (qidModule x)
 data TransEnv = TransEnv
   { moduleIdent :: ModuleIdent
   , valueEnv    :: ValueEnv
+  , tyconEnv    :: TCEnv
   }
 
 type TransM a = R.Reader TransEnv a
@@ -110,27 +119,40 @@ type TransM a = R.Reader TransEnv a
 getValueEnv :: TransM ValueEnv
 getValueEnv = R.asks valueEnv
 
+getTCEnv :: TransM TCEnv
+getTCEnv = R.asks tyconEnv
+
 trQualify :: Ident -> TransM QualIdent
 trQualify i = flip qualifyWith i <$> R.asks moduleIdent
 
--- Return the type of a variable
-varType :: QualIdent -> TransM Type
-varType f = do
-  tyEnv <- getValueEnv
-  case qualLookupValue f tyEnv of
-    [Value _ _ _ (ForAll _ (PredType _ ty))] -> return ty
-    [Label _ _ (ForAll _ (PredType _ ty))] -> return ty
-    _ -> internalError $ "CurryToIL.varType: " ++ show f
+getArity :: QualIdent -> TransM Int
+getArity qid = do
+    vEnv <- getValueEnv
+    return $ case qualLookupValue qid vEnv of
+      [DataConstructor  _ a _ _] -> a
+      [NewtypeConstructor _ _ _] -> 1
+      [Value            _ _ a _] -> a
+      [Label              _ _ _] -> 1
+      _                          ->
+        internalError $ "CurryToIL.getArity: " ++ show qid
 
 -- Return the type of a constructor
 constrType :: QualIdent -> TransM Type
 constrType c = do
   vEnv <- getValueEnv
   case qualLookupValue c vEnv of
-    [DataConstructor  _ _ _ (ForAllExist _ _ (PredType _ ty))] -> return ty
-    [NewtypeConstructor _ _ (ForAllExist _ _ (PredType _ ty))] -> return ty
+    [DataConstructor  _ _ _ (ForAll _ (PredType _ ty))] -> return ty
+    [NewtypeConstructor _ _ (ForAll _ (PredType _ ty))] -> return ty
     _ -> internalError $ "CurryToIL.constrType: " ++ show c
 
+-- Return the kinds of a type constructor's type variables
+tcTVarKinds :: QualIdent -> TransM [Kind]
+tcTVarKinds qid = do
+  tcEnv <- getTCEnv
+  let mid = fromJust $ qidModule qid
+      kind = tcKind mid qid tcEnv
+  return $ kindArgs kind
+
 -- -----------------------------------------------------------------------------
 -- Translation
 -- -----------------------------------------------------------------------------
@@ -144,59 +166,177 @@ constrType c = do
 
 trDecl :: Decl Type -> TransM [IL.Decl]
 trDecl (DataDecl     _ tc tvs cs _) = (:[]) <$> trData tc tvs cs
+trDecl (NewtypeDecl  _ tc tvs nc _) = (:[]) <$> trNewtype tc tvs nc
 trDecl (ExternalDataDecl  _ tc tvs) = (:[]) <$> trExternalData tc tvs
-trDecl (FunctionDecl    _ ty f eqs) = (:[]) <$> trFunction f ty eqs
+trDecl (FunctionDecl    _ _ f  eqs) = (:[]) <$> trFunction f eqs
 trDecl (ExternalDecl          _ vs) = mapM trExternal vs
 trDecl _                            = return []
 
 trData :: Ident -> [Ident] -> [ConstrDecl] -> TransM IL.Decl
 trData tc tvs cs = do
   tc' <- trQualify tc
-  IL.DataDecl tc' (length tvs) <$> mapM trConstrDecl cs
+  ks <- tcTVarKinds tc'
+  IL.DataDecl tc' (transKind <$> ks) <$> mapM trConstrDecl cs
+
+trNewtype :: Ident -> [Ident] -> NewConstrDecl -> TransM IL.Decl
+trNewtype tc tvs nc = do
+  tc' <- trQualify tc
+  ks <- tcTVarKinds tc'
+  IL.NewtypeDecl tc' (transKind <$> ks) <$> trNewConstrDecl nc
 
 trConstrDecl :: ConstrDecl -> TransM IL.ConstrDecl
 trConstrDecl d = do
   c' <- trQualify (constr d)
   ty' <- arrowArgs <$> constrType c'
-  return $ IL.ConstrDecl c' (map transType ty')
+  tcEnv <- getTCEnv
+  return $ IL.ConstrDecl c' (map (transType tcEnv) ty')
   where
-  constr (ConstrDecl    _ _ _ c _) = c
-  constr (ConOpDecl  _ _ _ _ op _) = op
-  constr (RecordDecl    _ _ _ c _) = c
+  constr (ConstrDecl    _ c _) = c
+  constr (ConOpDecl  _ _ op _) = op
+  constr (RecordDecl    _ c _) = c
+
+trNewConstrDecl :: NewConstrDecl -> TransM IL.NewConstrDecl
+trNewConstrDecl d = do
+  c' <- trQualify (constr d)
+  ty' <- arrowArgs <$> constrType c'
+  tcEnv <- getTCEnv
+  case ty' of
+    [ty] -> return $ IL.NewConstrDecl c' (transType tcEnv ty)
+    _    -> internalError "CurryToIL.trNewConstrDecl: invalid constructor type"
+  where
+  constr (NewConstrDecl    _ c _) = c
+  constr (NewRecordDecl    _ c _) = c
 
 trExternalData :: Ident -> [Ident] -> TransM IL.Decl
-trExternalData tc tvs = flip IL.ExternalDataDecl (length tvs) <$> trQualify tc
+trExternalData tc tvs = do
+  tc' <- trQualify tc
+  ks <- tcTVarKinds tc'
+  return $ IL.ExternalDataDecl tc' (transKind <$> ks)
 
 trExternal :: Var Type -> TransM IL.Decl
-trExternal (Var ty f) = flip IL.ExternalDecl (transType ty) <$> trQualify f
+trExternal (Var ty f) = do
+  tcEnv <- getTCEnv
+  f' <- trQualify f
+  a <- getArity f'
+  return $ IL.ExternalDecl f' a (transType tcEnv $ polyType ty)
 
 -- The type representation in the intermediate language does not support
 -- types with higher order kinds. Therefore, the type transformations has
 -- to transform all types to first order terms. To that end, we assume the
--- existence of a type synonym 'type @ f a = f a'. In addition, the type
+-- existence of a type synonym 'type Apply f a = f a'. In addition, the type
 -- representation of the intermediate language does not support constrained
 -- type variables and skolem types. The former are fixed and the later are
 -- replaced by fresh type constructors.
 
-transType :: Type -> IL.Type
-transType ty = transType' ty []
-
-transType' :: Type -> [IL.Type] -> IL.Type
-transType' (TypeConstructor    tc) = IL.TypeConstructor tc
-transType' (TypeApply     ty1 ty2) = transType' ty1 . (transType ty2 :)
-transType' (TypeVariable       tv) = foldl applyType' (IL.TypeVariable tv)
-transType' (TypeConstrained tys _) = transType' (head tys)
-transType' (TypeSkolem          k) =
-  foldl applyType' (IL.TypeConstructor (qualify (mkIdent ("_" ++ show k))) [])
-transType' (TypeArrow     ty1 ty2) =
-  foldl applyType' (IL.TypeArrow (transType ty1) (transType ty2))
-transType' (TypeForall     tvs ty) =
-  foldl applyType' (IL.TypeForall tvs (transType ty))
+transType :: TCEnv -> Type -> IL.Type
+transType tcEnv ty' = transType' ty' []
+  where
+    ks = transTVars tcEnv ty'
+    transType' (TypeConstructor    tc) = IL.TypeConstructor tc
+    transType' (TypeApply     ty1 ty2) = transType' ty1 . (transType' ty2 [] :)
+    transType' (TypeVariable       tv) = foldl applyType' (IL.TypeVariable tv)
+    transType' (TypeConstrained tys _) = transType' (head tys)
+    transType' (TypeArrow     ty1 ty2) =
+      foldl applyType' (IL.TypeArrow (transType' ty1 []) (transType' ty2 []))
+    transType' (TypeForall     tvs ty) =
+      foldl applyType' (IL.TypeForall tvs' (transType' ty []))
+      where tvs' = filter ((`elem` tvs) . fst) ks
 
 applyType' :: IL.Type -> IL.Type -> IL.Type
 applyType' ty1 ty2 =
   IL.TypeConstructor (qualifyWith preludeMIdent (mkIdent "Apply")) [ty1, ty2]
 
+-- We need to existentially quantify all variables in some types
+polyType :: Type -> Type
+polyType (TypeForall _ ty) = polyType ty
+polyType ty                =
+  let vs = nub $ typeVars ty
+  in if null vs then ty else TypeForall vs ty
+
+-- We need to infer kinds for the quantified variables.
+-- We already checked the correctness of all Kinds earlier,
+-- thus we know that we will be able to unify all the inferred equations.
+-- We can also keep a flat environment,
+-- as all variables have already been renamed.
+
+data KIS = KIS
+  { _nextId :: Int
+  , kinds  :: Map.Map Int IL.Kind
+  }
+
+freshId :: S.State KIS Int
+freshId = do
+  KIS i ks <- S.get
+  S.put (KIS (i+1) ks)
+  return i
+
+transTVars :: TCEnv -> Type -> [(Int, IL.Kind)]
+transTVars tcEnv ty' =
+  Map.toList $ kinds $ S.execState (build ty' IL.KindStar) (KIS 0 Map.empty)
+  where
+    build :: Type -> IL.Kind -> S.State KIS ()
+    build (TypeArrow     ty1 ty2) _ =
+      build ty1 IL.KindStar >> build ty2 IL.KindStar
+    build (TypeConstrained tys _) k =
+      build (head tys) k
+    build (TypeForall       _ ty) k =
+      build ty k
+    build (TypeVariable       tv) k = do
+      KIS i ks <- S.get
+      -- get current kind
+      let k' = Map.findWithDefault k tv ks
+      -- unify it
+      let s = unifyKind k k'
+      -- apply substitution
+      let ks' = fmap (applyKindSubst s) $ Map.insert tv k' ks
+      S.put (KIS i ks')
+    build (TypeConstructor     _) _ = return ()
+    build ta@(TypeApply       _ _) k =
+      let (ty, tys) = unapplyType True ta
+      in case ty of
+        TypeConstructor tc -> do
+          let k' = tcKind (fromJust $ qidModule tc) tc tcEnv
+          mapM_ (uncurry build) (zip tys $ unarrowKind $ transKind k')
+        _ -> do -- var of forall
+          -- construct new kind vars
+          ks <- mapM (const (freshId >>= return . IL.KindVariable)) tys
+          -- infer kind for v
+          build ty (foldr IL.KindArrow k ks)
+          -- infer kinds for args
+          mapM_ (uncurry build) (zip tys ks)
+
+type KindSubst = Map.Map Int IL.Kind
+
+transKind :: Kind -> IL.Kind
+transKind KindStar          = IL.KindStar
+transKind (KindVariable  _) = IL.KindStar
+transKind (KindArrow k1 k2) = IL.KindArrow (transKind k1) (transKind k2)
+
+unarrowKind :: IL.Kind -> [IL.Kind]
+unarrowKind (IL.KindArrow k1 k2) = k1 : unarrowKind k2
+unarrowKind k                    = [k]
+
+applyKindSubst :: KindSubst -> IL.Kind -> IL.Kind
+applyKindSubst _ IL.KindStar =
+  IL.KindStar
+applyKindSubst s (IL.KindArrow k1 k2) =
+  IL.KindArrow (applyKindSubst s k1) (applyKindSubst s k2)
+applyKindSubst s v@(IL.KindVariable i) =
+  Map.findWithDefault v i s
+
+composeKindSubst :: KindSubst -> KindSubst -> KindSubst
+composeKindSubst s1 s2 = Map.map (applyKindSubst s1) s2 `Map.union` s1
+
+unifyKind :: IL.Kind -> IL.Kind -> KindSubst
+unifyKind IL.KindStar          IL.KindStar            = Map.empty
+unifyKind (IL.KindVariable i)  k                      = Map.singleton i k
+unifyKind k                    (IL.KindVariable i)    = Map.singleton i k
+unifyKind (IL.KindArrow k1 k2) (IL.KindArrow k1' k2') =
+  let s1 = unifyKind k1 k1'
+      s2 = unifyKind (applyKindSubst s1 k2) (applyKindSubst s1 k2')
+  in s1 `composeKindSubst` s2
+unifyKind k1 k2 = error $ "Transformation.CurryToIL.unifyKind: " ++ show k1 ++ ", " ++ show k2
+
 -- Each function in the program is translated into a function of the
 -- intermediate language. The arguments of the function are renamed such
 -- that all variables occurring in the same position (in different
@@ -217,17 +357,19 @@ applyType' ty1 ty2 =
 -- selector function have to be renamed according to the name mapping
 -- computed for its first argument.
 
-trFunction :: Ident -> Type -> [Equation Type] -> TransM IL.Decl
-trFunction f ty eqs = do
+trFunction :: Ident -> [Equation Type] -> TransM IL.Decl
+trFunction f eqs = do
   f' <- trQualify f
-  let ty' = transType ty
-      vs' = zip (map (transType . typeOf) ts) vs
+  tcEnv <- getTCEnv
+  let tys = map typeOf ts
+      ty' = transType tcEnv $ polyType $ foldr TypeArrow (typeOf rhs) tys
+      vs' = zip (map (transType tcEnv) tys) vs
   alts <- mapM (trEquation vs ws) eqs
   return $ IL.FunctionDecl f' vs' ty' (flexMatch vs' alts)
   where
   -- vs are the variables needed for the function: _1, _2, etc.
   -- ws is an infinite list for introducing additional variables later
-  Equation _ lhs _ = head eqs
+  Equation _ lhs rhs = head eqs
   (_, ts) = flatLhs lhs
   (vs, ws) = splitAt (length ts) (argNames (mkIdent ""))
 
@@ -241,7 +383,8 @@ trEquation vs vs' (Equation _ (FunLhs _ _ ts) rhs) = do
   -- translate right-hand-side
   rhs' <- trRhs vs' patternRenaming rhs
   -- convert patterns
-  return (zipWith trPattern vs ts, rhs')
+  tcEnv <- getTCEnv
+  return (zipWith (trPattern tcEnv) vs ts, rhs')
 trEquation _  _    _
   = internalError "Translation of non-FunLhs euqation not defined"
 
@@ -259,8 +402,8 @@ bindRenameEnv _ _                           _
   = internalError "CurryToIL.bindRenameEnv"
 
 trRhs :: [Ident] -> RenameEnv -> Rhs Type -> TransM IL.Expression
-trRhs vs env (SimpleRhs _ e _) = trExpr vs env e
-trRhs _  _   (GuardedRhs _ _  _) = internalError "CurryToIL.trRhs: GuardedRhs"
+trRhs vs env (SimpleRhs _ _ e _) = trExpr vs env e
+trRhs _  _   (GuardedRhs _ _ _ _) = internalError "CurryToIL.trRhs: GuardedRhs"
 
 -- Note that the case matching algorithm assumes that the matched
 -- expression is accessible through a variable. The translation of case
@@ -271,22 +414,30 @@ trRhs _  _   (GuardedRhs _ _  _) = internalError "CurryToIL.trRhs: GuardedRhs"
 -- instance, if one of the alternatives contains an as-pattern.
 
 trExpr :: [Ident] -> RenameEnv -> Expression Type -> TransM IL.Expression
-trExpr _  _   (Literal     _ ty l) = return $ IL.Literal (transType ty) (trLiteral l)
+trExpr _  _   (Literal     _ ty l) = do
+  tcEnv <- getTCEnv
+  return $ IL.Literal (transType tcEnv ty) (trLiteral l)
 trExpr _  env (Variable    _ ty v)
-  | isQualified v = fun
-  | otherwise     = case Map.lookup (unqualify v) env of
-      Nothing -> fun
-      Just v' -> return $ IL.Variable (transType ty) v' -- apply renaming
-  where fun = (IL.Function (transType ty) v . arrowArity) <$> varType v
-trExpr _  _   (Constructor _ ty c)
-  = (IL.Constructor (transType ty) c . arrowArity) <$> constrType c
+  | isQualified v = getTCEnv >>= fun
+  | otherwise     = do
+    tcEnv <- getTCEnv
+    case Map.lookup (unqualify v) env of
+      Nothing -> error $ "unexpected variable" ++ show v --TODO: Replace case by fromJust?
+      Just v' -> return $ IL.Variable (transType tcEnv ty) v' -- apply renaming
+  where
+    fun tcEnv = IL.Function (transType tcEnv ty) v <$> getArity v
+trExpr _  _   (Constructor _ ty c) = do
+  tcEnv <- getTCEnv
+  IL.Constructor (transType tcEnv ty) c <$> getArity c
 trExpr vs env (Apply     _ e1 e2)
   = IL.Apply <$> trExpr vs env e1 <*> trExpr vs env e2
-trExpr vs env (Let        _ ds e) = do
+trExpr vs env (Let      _ _ ds e) = do
   e' <- trExpr vs env' e
   case ds of
     [FreeDecl _ vs']
-       -> return $ foldr (\ (Var ty v) -> IL.Exist v (transType ty)) e' vs'
+       -> do tcEnv <- getTCEnv
+             return $
+               foldr (\ (Var ty v) -> IL.Exist v (transType tcEnv ty)) e' vs'
     [d] | all (`notElem` bv d) (qfv emptyMIdent d)
       -> flip IL.Let    e' <$>      trBinding d
     _ -> flip IL.Letrec e' <$> mapM trBinding ds
@@ -296,12 +447,13 @@ trExpr vs env (Let        _ ds e) = do
   trBinding (PatternDecl _ (VariablePattern _ _ v) rhs)
     = IL.Binding v <$> trRhs vs env' rhs
   trBinding p = error $ "unexpected binding: " ++ show p
-trExpr (v:vs) env (Case _ ct e alts) = do
+trExpr (v:vs) env (Case _ _ ct e alts) = do
   -- the ident v is used for the case expression subject, as this could
   -- be referenced in the case alternatives by a variable pattern
   e' <- trExpr vs env e
+  tcEnv <- getTCEnv
   let matcher = if ct == Flex then flexMatch else rigidMatch
-      ty'     = transType $ typeOf e
+      ty'     = transType tcEnv $ typeOf e
   expr <- matcher [(ty', v)] <$> mapM (trAlt (v:vs) env) alts
   return $ case expr of
     IL.Case mode (IL.Variable _ v') alts'
@@ -311,15 +463,17 @@ trExpr (v:vs) env (Case _ ct e alts) = do
         -- subject is referenced -> introduce binding for v as subject
       | v `elem` fv expr                -> IL.Let (IL.Binding v e') expr
       | otherwise                       -> expr
-trExpr  vs env (Typed _ e (QualTypeExpr _ _ ty)) =
-  flip IL.Typed ty' <$> trExpr vs env e
-  where ty' = transType (toType [] ty)
+trExpr vs env (Typed _ e _) = do
+  tcEnv <- getTCEnv
+  e' <- trExpr vs env e
+  return $ IL.Typed e' (transType tcEnv $ typeOf e)
 trExpr _ _ _ = internalError "CurryToIL.trExpr"
 
 trAlt :: [Ident] -> RenameEnv -> Alt Type -> TransM Match
 trAlt ~(v:vs) env (Alt _ t rhs) = do
+  tcEnv <- getTCEnv
   rhs' <- trRhs vs (bindRenameEnv v t env) rhs
-  return ([trPattern v t], rhs')
+  return ([trPattern tcEnv v t], rhs')
 
 trLiteral :: Literal -> IL.Literal
 trLiteral (Char  c) = IL.Char c
@@ -339,18 +493,20 @@ pattern (NestedTerm t _) = t
 arguments :: NestedTerm -> [NestedTerm]
 arguments (NestedTerm _ ts) = ts
 
-trPattern :: Ident -> Pattern Type -> NestedTerm
-trPattern _ (LiteralPattern        _ ty l)
-  = NestedTerm (IL.LiteralPattern (transType ty) $ trLiteral l) []
-trPattern v (VariablePattern       _ ty _)
-  = NestedTerm (IL.VariablePattern (transType ty) v) []
-trPattern v (ConstructorPattern _ ty c ts)
-  = NestedTerm (IL.ConstructorPattern (transType ty) c vs')
-               (zipWith trPattern vs ts)
+trPattern :: TCEnv -> Ident -> Pattern Type -> NestedTerm
+trPattern tcEnv _ (LiteralPattern        _ ty l)
+  = NestedTerm (IL.LiteralPattern (transType tcEnv ty) $ trLiteral l) []
+trPattern tcEnv v (VariablePattern       _ ty _)
+  = NestedTerm (IL.VariablePattern (transType tcEnv ty) v) []
+trPattern tcEnv v (ConstructorPattern _ ty c ts)
+  = NestedTerm (IL.ConstructorPattern (transType tcEnv ty) c vs')
+               (zipWith (trPattern tcEnv) vs ts)
   where vs  = argNames v
-        vs' = zip (map (transType . typeOf) ts) vs
-trPattern v (AsPattern              _ _ t) = trPattern v t
-trPattern _ _                            = internalError "CurryToIL.trPattern"
+        vs' = zip (map (transType tcEnv . typeOf) ts) vs
+trPattern tcEnv v (AsPattern              _ _ t)
+  = trPattern tcEnv v t
+trPattern _ _ _
+  = internalError "CurryToIL.trPattern"
 
 argNames :: Ident -> [Ident]
 argNames v = [mkIdent (prefix ++ show i) | i <- [1 :: Integer ..] ]
