@@ -62,7 +62,6 @@ data DTState = DTState
   , classEnv    :: ClassEnv
   , instEnv     :: InstEnv
   , opPrecEnv   :: OpPrecEnv
-  , augmentEnv  :: AugmentEnv -- for augmenting nullary class methods
   , dictEnv     :: DictEnv    -- for dictionary insertion
   , specEnv     :: SpecEnv    -- for dictionary specialization
   , nextId      :: Integer
@@ -70,15 +69,15 @@ data DTState = DTState
 
 type DTM = S.State DTState
 
-insertDicts :: InterfaceEnv -> TCEnv -> ValueEnv -> ClassEnv -> InstEnv
-            -> OpPrecEnv -> Module PredType
+insertDicts :: Bool -> InterfaceEnv -> TCEnv -> ValueEnv -> ClassEnv
+            -> InstEnv -> OpPrecEnv -> Module PredType
             -> (Module Type, InterfaceEnv, TCEnv, ValueEnv, OpPrecEnv)
-insertDicts intfEnv tcEnv vEnv clsEnv inEnv pEnv mdl@(Module _ _ m _ _ _) =
+insertDicts inlDi intfEnv tcEnv vEnv clsEnv inEnv pEnv mdl@(Module _ _ _ m _ _ _) =
   (mdl', intfEnv', tcEnv', vEnv', pEnv')
   where initState =
-          DTState m tcEnv vEnv clsEnv inEnv pEnv emptyAugEnv emptyDictEnv emptySpEnv 1
+          DTState m tcEnv vEnv clsEnv inEnv pEnv emptyDictEnv emptySpEnv 1
         (mdl', tcEnv', vEnv', pEnv') =
-          runDTM (augment mdl >>= dictTrans >>= specialize >>= cleanup) initState
+          runDTM (dictTrans mdl >>= (if inlDi then specialize else return) >>= cleanup) initState
         intfEnv' = dictTransInterfaces vEnv' clsEnv intfEnv
 
 runDTM :: DTM a -> DTState -> (a, TCEnv, ValueEnv, OpPrecEnv)
@@ -122,12 +121,6 @@ getPrecEnv = S.gets opPrecEnv
 modifyPrecEnv :: (OpPrecEnv -> OpPrecEnv) -> DTM ()
 modifyPrecEnv f = S.modify $ \s -> s { opPrecEnv = f $ opPrecEnv s }
 
-getAugEnv :: DTM AugmentEnv
-getAugEnv = S.gets augmentEnv
-
-setAugEnv :: AugmentEnv -> DTM ()
-setAugEnv augEnv = S.modify $ \s -> s { augmentEnv = augEnv }
-
 getDictEnv :: DTM DictEnv
 getDictEnv = S.gets dictEnv
 
@@ -153,167 +146,6 @@ getNextId = do
   S.modify $ \s -> s { nextId = succ nid }
   return nid
 
--- -----------------------------------------------------------------------------
--- Augmenting nullary class methods
--- -----------------------------------------------------------------------------
-
--- To prevent unwanted sharing of non-determinism for nullary class methods
--- we augment them with an additional unit argument.
-
-type AugmentEnv = [QualIdent]
-
-emptyAugEnv :: AugmentEnv
-emptyAugEnv = []
-
-initAugEnv :: ValueEnv -> AugmentEnv
-initAugEnv = foldr (bindValue . snd) emptyAugEnv . allBindings
-  where bindValue (Value f True _ (ForAll _ (PredType _ ty)))
-          | arrowArity ty == 0 = (f :)
-        bindValue _ = id
-
-isAugmented :: AugmentEnv -> QualIdent -> Bool
-isAugmented = flip elem
-
-augmentValues :: ValueEnv -> ValueEnv
-augmentValues = fmap augmentValueInfo
-
-augmentValueInfo :: ValueInfo -> ValueInfo
-augmentValueInfo (Value f True a (ForAll n (PredType ps ty)))
-  | arrowArity ty == 0 = Value f True a $ ForAll n $ PredType ps ty'
-  where ty' = augmentType ty
-augmentValueInfo vi = vi
-
-augmentTypes :: TCEnv -> TCEnv
-augmentTypes = fmap augmentTypeInfo
-
-augmentTypeInfo :: TypeInfo -> TypeInfo
-augmentTypeInfo (TypeClass cls k ms) =
-  TypeClass cls k $ map augmentClassMethod ms
-augmentTypeInfo ti = ti
-
-augmentClassMethod :: ClassMethod -> ClassMethod
-augmentClassMethod mthd@(ClassMethod f a (PredType ps ty))
-  | arrowArity ty == 0 =
-    ClassMethod f (Just $ fromMaybe 0 a + 1) $ PredType ps $ augmentType ty
-  | otherwise = mthd
-
-augmentInstances :: AugmentEnv -> InstEnv -> InstEnv
-augmentInstances = Map.mapWithKey . augmentInstInfo
-
-augmentInstInfo :: AugmentEnv -> InstIdent -> InstInfo -> InstInfo
-augmentInstInfo augEnv (cls, _) (m, ps, is) =
-  (m, ps, map (augmentInstImpl augEnv cls) is)
-
-augmentInstImpl :: AugmentEnv -> QualIdent -> (Ident, Int) -> (Ident, Int)
-augmentInstImpl augEnv cls (f, a)
-  | isAugmented augEnv (qualifyLike cls f) = (f, a + 1)
-  | otherwise = (f, a)
-
-class Augment a where
-  augment :: a PredType -> DTM (a PredType)
-
-instance Augment Module where
-  augment (Module spi ps m es is ds) = do
-    augEnv <- initAugEnv <$> getValueEnv
-    setAugEnv augEnv
-    modifyValueEnv $ augmentValues
-    modifyTyConsEnv $ augmentTypes
-    modifyInstEnv $ augmentInstances augEnv
-    Module spi ps m es is <$> mapM (augmentDecl Nothing) ds
-
--- The first parameter of the functions 'augmentDecl', 'augmentEquation' and
--- 'augmentLhs' determines whether we have to unrename the function identifiers
--- before checking if the function has to augmented or not. Furthermore, it
--- specifies the module unqualified identifiers have to be qualified with.
--- The unrenaming is necessary for both class and instance declarations as all
--- identifiers within these have been renamed during the syntax check, while the
--- qualifying is needed for function declarations within instance declarations
--- as the implemented class methods can originate from another module. If not
--- qualified properly, the lookup in the augmentation environment would fail.
-
--- Since type signatures remain only in class declarations due to desugaring,
--- we can always perform the unrenaming and it is safe to assume that all other
--- functions mentioned in a type signature have to be augmented as well if the
--- first one is affected.
-
-augmentDecl :: Maybe ModuleIdent -> Decl PredType -> DTM (Decl PredType)
-augmentDecl _  d@(TypeSig          p fs qty) = do
-  m <- getModuleIdent
-  augEnv <- getAugEnv
-  return $ if isAugmented augEnv (qualifyWith m $ unRenameIdent $ head fs)
-              then TypeSig p fs $ augmentQualTypeExpr qty
-              else d
-augmentDecl mm (FunctionDecl    p pty f eqs) = do
-    eqs' <- mapM (augmentEquation mm) eqs
-    m <- maybe getModuleIdent return mm
-    augEnv <- getAugEnv
-    if isAugmented augEnv (qualifyWith m $ unRenameIdentIf (isJust mm) f)
-      then return $ FunctionDecl p (augmentPredType pty) f eqs'
-      else return $ FunctionDecl p pty f eqs'
-augmentDecl _  (PatternDecl         p t rhs) = PatternDecl p t <$> augment rhs
-augmentDecl _  (ClassDecl    p cx cls tv ds) = do
-  m <- getModuleIdent
-  ClassDecl p cx cls tv <$> mapM (augmentDecl $ Just m) ds
-augmentDecl _  (InstanceDecl p cx cls ty ds) =
-  InstanceDecl p cx cls ty <$> mapM (augmentDecl $ qidModule cls) ds
-augmentDecl _ d                             = return d
-
-augmentEquation :: Maybe ModuleIdent -> Equation PredType
-                -> DTM (Equation PredType)
-augmentEquation mm (Equation p lhs rhs) =
-  Equation p <$> augmentLhs mm lhs <*> augment rhs
-
-augmentLhs :: Maybe ModuleIdent -> Lhs PredType -> DTM (Lhs PredType)
-augmentLhs mm lhs@(FunLhs spi f ts) = do
-    m <- maybe getModuleIdent return mm
-    augEnv <- getAugEnv
-    if isAugmented augEnv (qualifyWith m $ unRenameIdentIf (isJust mm) f)
-      then return $ FunLhs spi f
-                  $ ConstructorPattern NoSpanInfo predUnitType qUnitId [] : ts
-      else return lhs
-augmentLhs _ lhs               =
-  internalError $ "Dictionary.augmentLhs" ++ show lhs
-
-instance Augment Rhs where
-  augment (SimpleRhs p e []) = simpleRhs p <$> augment e
-  augment rhs                =
-    internalError $ "Dictionary.augment: " ++ show rhs
-
-instance Augment Expression where
-  augment l@(Literal     _ _ _) = return l
-  augment v@(Variable _ pty v') = do
-    augEnv <- getAugEnv
-    return $ if isAugmented augEnv v'
-               then apply (Variable NoSpanInfo (augmentPredType pty) v')
-                      [Constructor NoSpanInfo predUnitType qUnitId]
-               else v
-  augment c@(Constructor _ _ _) = return c
-  augment (Typed       spi e qty) = flip (Typed spi) qty <$> augment e
-  augment (Apply       spi e1 e2) = Apply spi <$> augment e1 <*> augment e2
-  augment (Lambda       spi ts e) = Lambda spi ts <$> augment e
-  augment (Let          spi ds e) =
-    Let spi <$> mapM (augmentDecl Nothing) ds <*> augment e
-  augment (Case      spi ct e as) =
-    Case spi ct <$> augment e <*> mapM augment as
-  augment e                     =
-    internalError $ "Dictionary.augment: " ++ show e
-
-instance Augment Alt where
-  augment (Alt p t rhs) = Alt p t <$> augment rhs
-
-augmentPredType :: PredType -> PredType
-augmentPredType (PredType ps ty) = PredType ps $ augmentType ty
-
-augmentType :: Type -> Type
-augmentType = TypeArrow unitType
-
-augmentQualTypeExpr :: QualTypeExpr -> QualTypeExpr
-augmentQualTypeExpr (QualTypeExpr spi cx ty) =
-  QualTypeExpr spi cx $ augmentTypeExpr ty
-
-augmentTypeExpr :: TypeExpr -> TypeExpr
-augmentTypeExpr = ArrowType NoSpanInfo $ ConstructorType NoSpanInfo qUnitId
-
 -- -----------------------------------------------------------------------------
 -- Lifting class and instance declarations
 -- -----------------------------------------------------------------------------
@@ -324,10 +156,10 @@ augmentTypeExpr = ArrowType NoSpanInfo $ ConstructorType NoSpanInfo qUnitId
 
 liftDecls :: Decl PredType -> DTM [Decl PredType]
 liftDecls (DefaultDecl _ _) = return []
-liftDecls (ClassDecl _ _ cls tv ds) = do
+liftDecls (ClassDecl _ _ _ cls tv ds) = do
   m <- getModuleIdent
   liftClassDecls (qualifyWith m cls) tv ds
-liftDecls (InstanceDecl _ cx cls ty ds) = do
+liftDecls (InstanceDecl _ _ cx cls ty ds) = do
   clsEnv <- getClassEnv
   let PredType ps ty' = toPredType [] $ QualTypeExpr NoSpanInfo cx ty
       ps' = minPredSet clsEnv ps
@@ -373,14 +205,10 @@ createClassDictDecl cls tv ds = do
 
 createClassDictConstrDecl :: QualIdent -> Ident -> [Decl a] -> DTM ConstrDecl
 createClassDictConstrDecl cls tv ds = do
-  clsEnv <- getClassEnv
-  let sclss = superClasses cls clsEnv
-      cx    = [Constraint NoSpanInfo scls (VariableType NoSpanInfo tv)
-              | scls <- sclss]
-      tvs  = tv : filter (unRenameIdent tv /=) identSupply
+  let tvs  = tv : filter (unRenameIdent tv /=) identSupply
       mtys = map (fromType tvs . generalizeMethodType . transformMethodPredType)
                  [toMethodType cls tv qty | TypeSig _ fs qty <- ds, _ <- fs]
-  return $ ConstrDecl NoSpanInfo [] cx (dictConstrId cls) mtys
+  return $ ConstrDecl NoSpanInfo (dictConstrId cls) mtys
 
 classDictConstrPredType :: ValueEnv -> ClassEnv -> QualIdent -> PredType
 classDictConstrPredType vEnv clsEnv cls = PredType ps $ foldr TypeArrow ty mtys
@@ -394,7 +222,7 @@ classDictConstrPredType vEnv clsEnv cls = PredType ps $ foldr TypeArrow ty mtys
 createInstDictDecl :: PredSet -> QualIdent -> Type -> DTM (Decl PredType)
 createInstDictDecl ps cls ty = do
   pty <- PredType ps . arrowBase <$> getInstDictConstrType cls ty
-  funDecl NoSpanInfo pty (instFunId cls ty) [] <$> createInstDictExpr cls ty
+  funDecl NoSpanInfo pty (instFunId cls ty) [ConstructorPattern NoSpanInfo predUnitType qUnitId []] <$> createInstDictExpr cls ty
 
 createInstDictExpr :: QualIdent -> Type -> DTM (Expression PredType)
 createInstDictExpr cls ty = do
@@ -418,13 +246,7 @@ createClassMethodDecl cls =
 defaultClassMethodDecl :: QualIdent -> Ident -> DTM (Decl PredType)
 defaultClassMethodDecl cls f = do
   pty@(PredType _ ty) <- getClassMethodType cls f
-  augEnv <- getAugEnv
-  let augmented = isAugmented augEnv (qualifyLike cls f)
-      pats = if augmented
-               then [ConstructorPattern NoSpanInfo predUnitType qUnitId []]
-               else []
-      ty' = if augmented then arrowBase ty else ty
-  return $ funDecl NoSpanInfo pty f pats $ preludeError (instType ty') $
+  return $ funDecl NoSpanInfo pty f [] $ preludeError (instType ty) $
     "No instance or default method for class operation " ++ escName f
 
 getClassMethodType :: QualIdent -> Ident -> DTM PredType
@@ -486,7 +308,7 @@ renameDecl _ _ = internalError "Dictionary.renameDecl"
 -- class dictionary from the provided class dictionary.
 
 createStubs :: Decl PredType -> DTM [Decl Type]
-createStubs (ClassDecl _ _ cls _ _) = do
+createStubs (ClassDecl _ _ _ cls _ _) = do
   m <- getModuleIdent
   vEnv <- getValueEnv
   clsEnv <- getClassEnv
@@ -502,52 +324,42 @@ createStubs (ClassDecl _ _ cls _ _) = do
         splitAt (length sclss) $ map (TypeArrow dictTy) superDictAndMethodTys
   superDictVs <- mapM (freshVar "_#super" . instType) superDictTys
   methodVs <- mapM (freshVar "_#meth" . instType) methodTys
-  methodDictTyss <- zipWithM (computeMethodDictTypes ocls) fs methodTys
-  methodDictVss <- mapM (mapM $ freshVar "_#dict" . instType) methodDictTyss
   let patternVs   = superDictVs ++ methodVs
       pattern     = createDictPattern (instType dictTy) ocls patternVs
       superStubs  = zipWith3 (createSuperDictStubDecl pattern ocls)
                       superStubTys sclss superDictVs
-      methodStubs = zipWith4 (createMethodStubDecl pattern)
-                      methodStubTys fs methodVs methodDictVss
+      methodStubs = zipWith3 (createMethodStubDecl pattern)
+                      methodStubTys fs methodVs
   return $ superStubs ++ methodStubs
 createStubs _ = return []
 
--- Computes the additional dictionary arguments of a transformed method type
--- which correspond to the constraints of the original class method's type.
-
-computeMethodDictTypes :: QualIdent -> Ident -> Type -> DTM [Type]
-computeMethodDictTypes cls f ty = do
-  PredType _ ty' <- getClassMethodType cls f
-  return $ take (length tys - arrowArity ty') tys
-  where tys = arrowArgs ty
-
-createDictPattern :: a -> QualIdent -> [(a, Ident)] -> Pattern a
+createDictPattern :: Type -> QualIdent -> [(Type, Ident)] -> Pattern Type
 createDictPattern a cls = constrPattern a (qDictConstrId cls)
 
-createSuperDictStubDecl :: Pattern a -> QualIdent -> a -> QualIdent
-                        -> (a, Ident) -> Decl a
+createSuperDictStubDecl :: Pattern Type -> QualIdent -> Type -> QualIdent
+                        -> (Type, Ident) -> Decl Type
 createSuperDictStubDecl t cls a super v =
-  createStubDecl t a (superDictStubId cls super) v []
+  createStubDecl t a (superDictStubId cls super) v
 
-createMethodStubDecl :: Pattern a -> a -> Ident -> (a, Ident) -> [(a, Ident)]
-                     -> Decl a
+createMethodStubDecl :: Pattern Type -> Type -> Ident -> (Type, Ident) -> Decl Type
 createMethodStubDecl = createStubDecl
 
-createStubDecl :: Pattern a -> a -> Ident -> (a, Ident) -> [(a, Ident)]
-               -> Decl a
-createStubDecl t a f v us =
-  FunctionDecl NoSpanInfo a f [createStubEquation t f v us]
+createStubDecl :: Pattern Type -> Type -> Ident -> (Type, Ident) -> Decl Type
+createStubDecl t a f v =
+  FunctionDecl NoSpanInfo a f [createStubEquation t f v]
 
-createStubEquation :: Pattern a -> Ident -> (a, Ident) -> [(a, Ident)]
-                   -> Equation a
-createStubEquation t f v us =
-  mkEquation NoSpanInfo f (t : map (uncurry (VariablePattern NoSpanInfo)) us) $
-    apply (uncurry mkVar v) (map (uncurry mkVar) us)
+createStubEquation :: Pattern Type -> Ident -> (Type, Ident) -> Equation Type
+createStubEquation t f v = 
+  mkEquation NoSpanInfo f [VariablePattern NoSpanInfo (TypeArrow unitType (typeOf t)) (mkIdent "_#temp")] $
+    mkLet [FunctionDecl NoSpanInfo (TypeArrow (typeOf t) (fst v)) (mkIdent "_#lambda")
+      [mkEquation NoSpanInfo (mkIdent "_#lambda") [t] $ uncurry mkVar v]]
+      (apply (Variable NoSpanInfo (TypeArrow (typeOf t) (fst v)) (qualify $ mkIdent "_#lambda"))
+        [apply (Variable NoSpanInfo (TypeArrow unitType (typeOf t)) (qualify $ mkIdent "_#temp"))
+          [Constructor NoSpanInfo unitType qUnitId]])
 
 superDictStubType :: QualIdent -> QualIdent -> Type -> Type
 superDictStubType cls super ty =
-  TypeArrow (dictType $ Pred cls ty) (dictType $ Pred super ty)
+  TypeArrow (rtDictType $ Pred cls ty) (rtDictType $ Pred super ty)
 
 -- -----------------------------------------------------------------------------
 -- Entering new bindings into the environments
@@ -561,7 +373,7 @@ bindDictType :: ModuleIdent -> ClassEnv -> TypeInfo -> TCEnv -> TCEnv
 bindDictType m clsEnv (TypeClass cls k ms) = bindEntity m tc ti
   where ti    = DataType tc (KindArrow k KindStar) [c]
         tc    = qDictTypeId cls
-        c     = DataConstr (dictConstrId cls) 0 ps tys
+        c     = DataConstr (dictConstrId cls) (map rtDictType (Set.toAscList ps) ++ tys)
         sclss = superClasses cls clsEnv
         ps    = Set.fromList [Pred scls (TypeVariable 0) | scls <- sclss]
         tys   = map (generalizeMethodType . transformMethodPredType . methodType) ms
@@ -591,7 +403,7 @@ bindClassDict m clsEnv cls vEnv = bindEntity m c dc vEnv
         dc = DataConstructor c a (replicate a anonId) tySc
         a  = Set.size ps + arrowArity ty
         pty@(PredType ps ty) = classDictConstrPredType vEnv clsEnv cls
-        tySc = ForAllExist 1 0 pty
+        tySc = ForAll 1 pty
 
 bindDefaultMethods :: ModuleIdent -> QualIdent -> [(Ident, Int)] -> ValueEnv
                    -> ValueEnv
@@ -607,7 +419,7 @@ bindSuperStubs :: ModuleIdent -> QualIdent -> [QualIdent] -> ValueEnv
 bindSuperStubs m = flip . foldr . bindSuperStub m
 
 bindSuperStub :: ModuleIdent -> QualIdent -> QualIdent -> ValueEnv -> ValueEnv
-bindSuperStub m cls scls = bindEntity m f $ Value f False 1 $ polyType ty
+bindSuperStub m cls scls = bindEntity m f $ Value f Nothing 1 $ polyType ty
   where f  = qSuperDictStubId cls scls
         ty = superDictStubType cls scls (TypeVariable 0)
 
@@ -626,7 +438,7 @@ bindInstFuns m tcEnv clsEnv ((cls, tc), (m', ps, is)) =
 bindInstDict :: ModuleIdent -> QualIdent -> Type -> ModuleIdent -> PredSet
              -> ValueEnv -> ValueEnv
 bindInstDict m cls ty m' ps =
-  bindMethod m (qInstFunId m' cls ty) 0 $ PredType ps $ dictType $ Pred cls ty
+  bindMethod m (qInstFunId m' cls ty) 1 $ PredType ps $ rtDictType $ Pred cls ty
 
 bindInstMethods :: ModuleIdent -> ClassEnv -> QualIdent -> Type -> ModuleIdent
                 -> PredSet -> [(Ident, Int)] -> ValueEnv -> ValueEnv
@@ -642,7 +454,7 @@ bindInstMethod m cls ty m' ps is f vEnv = bindMethod m f' a pty vEnv
 
 bindMethod :: ModuleIdent -> QualIdent -> Int -> PredType -> ValueEnv
            -> ValueEnv
-bindMethod m f n pty = bindEntity m f $ Value f False n $ typeScheme pty
+bindMethod m f n pty = bindEntity m f $ Value f Nothing n $ typeScheme pty
 
 -- The function 'bindEntity' introduces a binding for an entity into a top-level
 -- environment. Depending on whether the entity is defined in the current module
@@ -675,10 +487,9 @@ dictTransTypeInfo (TypeVar _) =
   internalError "Dictionary.dictTransTypeInfo: type variable"
 
 dictTransDataConstr :: DataConstr -> DataConstr
-dictTransDataConstr (DataConstr c n ps tys) =
-  DataConstr c n emptyPredSet $ map dictType (Set.toAscList ps) ++ tys
-dictTransDataConstr (RecordConstr c n ps _ tys) =
-  dictTransDataConstr $ DataConstr c n ps tys
+dictTransDataConstr (DataConstr c tys) = DataConstr c tys
+dictTransDataConstr (RecordConstr c _ tys) =
+  dictTransDataConstr $ DataConstr c tys
 
 -- For the same reason as in 'bindClassEntities' it is safe to use 'fromMaybe 0'
 -- in 'dictTransClassMethod'. Note that type classes are removed anyway in the
@@ -693,16 +504,16 @@ dictTransValues :: ValueEnv -> ValueEnv
 dictTransValues = fmap dictTransValueInfo
 
 dictTransValueInfo :: ValueInfo -> ValueInfo
-dictTransValueInfo (DataConstructor c a ls (ForAllExist n n' pty)) =
-  DataConstructor c a' ls' $ ForAllExist n n' $ predType ty
+dictTransValueInfo (DataConstructor c a ls (ForAll n pty)) =
+  DataConstructor c a' ls' $ ForAll n $ predType ty
   where a'  = arrowArity ty
         ls' = replicate (a' - a) anonId ++ ls
         ty  = transformPredType pty
-dictTransValueInfo (NewtypeConstructor c l (ForAllExist n n' pty)) =
-  NewtypeConstructor c l (ForAllExist n n' (predType (unpredType pty)))
+dictTransValueInfo (NewtypeConstructor c l (ForAll n pty)) =
+  NewtypeConstructor c l (ForAll n (predType (unpredType pty)))
 dictTransValueInfo (Value f cm a (ForAll n pty)) =
-  Value f False a' $ ForAll n $ predType ty
-  where a' = a + if cm then 1 else arrowArity ty - arrowArity (unpredType pty)
+  Value f Nothing a' $ ForAll n $ predType ty
+  where a' = a + if isJust cm then 1 else arrowArity ty - arrowArity (unpredType pty)
         ty = transformPredType pty
 dictTransValueInfo (Label l cs (ForAll n pty)) =
   Label l cs $ ForAll n $ predType $ unpredType pty
@@ -716,11 +527,11 @@ addExports (Just (Exporting p es)) es' = Just $ Exporting p $ es ++ es'
 addExports Nothing                 _   = internalError "Dictionary.addExports"
 
 dictExports :: Decl a -> DTM [Export]
-dictExports (ClassDecl _ _ cls _ _) = do
+dictExports (ClassDecl _ _ _ cls _ _) = do
   m <- getModuleIdent
   clsEnv <- getClassEnv
   return $ classExports m clsEnv cls
-dictExports (InstanceDecl _ _ cls ty _) = do
+dictExports (InstanceDecl _ _ _ cls ty _) = do
   m <- getModuleIdent
   clsEnv <- getClassEnv
   return $ instExports m clsEnv cls (toType [] ty)
@@ -751,7 +562,7 @@ class DictTrans a where
   dictTrans :: a PredType -> DTM (a Type)
 
 instance DictTrans Module where
-  dictTrans (Module spi ps m es is ds) = do
+  dictTrans (Module spi li ps m es is ds) = do
     liftedDs <- concatMapM liftDecls ds
     stubDs <- concatMapM createStubs ds
     tcEnv <- getTyConsEnv
@@ -764,7 +575,7 @@ instance DictTrans Module where
     modifyValueEnv $ dictTransValues
     modifyTyConsEnv $ dictTransTypes
     dictEs <- addExports es <$> concatMapM dictExports ds
-    return $ Module spi ps m dictEs is $ transDs ++ stubDs
+    return $ Module spi li ps m dictEs is $ transDs ++ stubDs
 
 -- We use and transform the type from the type constructor environment for
 -- transforming a constructor declaration as it contains the reduced and
@@ -793,17 +604,16 @@ instance DictTrans Decl where
       dictTrans $ FunctionDecl p pty v [Equation p (FunLhs NoSpanInfo v []) rhs]
     _ -> withLocalDictEnv $ PatternDecl p <$> dictTrans t <*> dictTrans rhs
   dictTrans d@(FreeDecl                _ _) = return $ fmap unpredType d
-  dictTrans d@(ExternalDecl            _ _) = return $ fmap unpredType d
+  dictTrans d@(ExternalDecl            _ _) = return $ fmap transformPredType d
   dictTrans d                               =
     internalError $ "Dictionary.dictTrans: " ++ show d
 
 dictTransConstrDecl :: [Ident] -> ConstrDecl -> DataConstr -> ConstrDecl
-dictTransConstrDecl tvs (ConstrDecl p evs _ c tes) dc =
-  ConstrDecl p evs [] c $ map (fromType $ tvs ++ evs ++ bvs) tys
-  where DataConstr _ _ _ tys = dictTransDataConstr dc
-        bvs = nub $ bv tes
-dictTransConstrDecl tvs (ConOpDecl p evs cx ty1 op ty2) dc =
-  dictTransConstrDecl tvs (ConstrDecl p evs cx op [ty1, ty2]) dc
+dictTransConstrDecl tvs (ConstrDecl p c tes) dc =
+  ConstrDecl p c $ map (fromType $ tvs ++ bvs) (constrTypes dc)
+  where bvs = nub $ bv tes
+dictTransConstrDecl tvs (ConOpDecl p ty1 op ty2) dc =
+  dictTransConstrDecl tvs (ConstrDecl p op [ty1, ty2]) dc
 dictTransConstrDecl _ d _ = internalError $ "Dictionary.dictTrans: " ++ show d
 
 instance DictTrans Equation where
@@ -818,8 +628,8 @@ instance DictTrans Equation where
     internalError $ "Dictionary.dictTrans: " ++ show eq
 
 instance DictTrans Rhs where
-  dictTrans (SimpleRhs p e []) = simpleRhs p <$> dictTrans e
-  dictTrans rhs                =
+  dictTrans (SimpleRhs p _ e []) = simpleRhs p <$> dictTrans e
+  dictTrans rhs                  =
     internalError $ "Dictionary.dictTrans: " ++ show rhs
 
 instance DictTrans Pattern where
@@ -856,12 +666,12 @@ instance DictTrans Expression where
   dictTrans (Lambda       _ ts e) = withLocalValueEnv $ withLocalDictEnv $ do
     ts' <- mapM dictTrans ts
     modifyValueEnv $ bindPatterns ts'
-    Lambda NoSpanInfo ts' <$> dictTrans e
-  dictTrans (Let          _ ds e) = withLocalValueEnv $ do
+    mkLambda ts' <$> dictTrans e
+  dictTrans (Let        _ _ ds e) = withLocalValueEnv $ do
     modifyValueEnv $ bindDecls ds
-    Let NoSpanInfo <$> mapM dictTrans ds <*> dictTrans e
-  dictTrans (Case      _ ct e as) =
-    Case NoSpanInfo ct <$> dictTrans e <*> mapM dictTrans as
+    mkLet <$> mapM dictTrans ds <*> dictTrans e
+  dictTrans (Case    _ _ ct e as) =
+    mkCase ct <$> dictTrans e <*> mapM dictTrans as
   dictTrans e                   =
     internalError $ "Dictionary.dictTrans: " ++ show e
 
@@ -881,7 +691,7 @@ instance DictTrans Alt where
 
 addDictArgs :: [Pred] -> [Pattern PredType] -> DTM [Pattern Type]
 addDictArgs pls ts = do
-  dictVars <- mapM (freshVar "_#dict" . dictType) pls
+  dictVars <- mapM (freshVar "_#dict" . rtDictType) pls
   clsEnv <- getClassEnv
   modifyDictEnv $ (++) $ dicts clsEnv $ zip pls (map (uncurry mkVar) dictVars)
   (++) (map (uncurry (VariablePattern NoSpanInfo )) dictVars)
@@ -917,7 +727,7 @@ instFunApp :: ModuleIdent -> [Pred] -> Pred -> DTM (Expression Type)
 instFunApp m pls p@(Pred cls ty) = apply (Variable NoSpanInfo ty' f)
   <$> mapM dictArg pls
   where f   = qInstFunId m cls ty
-        ty' = foldr1 TypeArrow $ map dictType $ pls ++ [p]
+        ty' = foldr1 TypeArrow $ map rtDictType $ pls ++ [p]
 
 instPredList :: Pred -> DTM (ModuleIdent, [Pred])
 instPredList (Pred cls ty) = case unapplyType True ty of
@@ -960,7 +770,7 @@ predListMatch :: [Pred] -> Type -> Maybe Type
 predListMatch []     ty = Just ty
 predListMatch (p:ps) ty = case ty of
   TypeForall _ ty'                                 -> predListMatch (p : ps) ty'
-  TypeArrow ty1 ty2 | ty1 == dictType (instPred p) -> predListMatch ps ty2
+  TypeArrow ty1 ty2 | ty1 == rtDictType (instPred p) -> predListMatch ps ty2
   _                                                -> Nothing
 
 splits :: [a] -> [([a], [a])]
@@ -992,11 +802,11 @@ class Specialize a where
   specialize :: a Type -> DTM (a Type)
 
 instance Specialize Module where
-  specialize (Module spi ps m es is ds) = do
+  specialize (Module spi li ps m es is ds) = do
     clsEnv <- getClassEnv
     inEnv <- getInstEnv
     setSpEnv $ initSpEnv clsEnv inEnv
-    Module spi ps m es is <$> mapM specialize ds
+    Module spi li ps m es is <$> mapM specialize ds
 
 instance Specialize Decl where
   specialize (FunctionDecl p ty f eqs) =
@@ -1008,8 +818,8 @@ instance Specialize Equation where
   specialize (Equation p lhs rhs) = Equation p lhs <$> specialize rhs
 
 instance Specialize Rhs where
-  specialize (SimpleRhs p e []) = simpleRhs p <$> specialize e
-  specialize rhs                =
+  specialize (SimpleRhs p _ e []) = simpleRhs p <$> specialize e
+  specialize rhs                  =
     internalError $ "Dictionary.specialize: " ++ show rhs
 
 instance Specialize Expression where
@@ -1035,14 +845,14 @@ specialize' (Apply       _ e1 e2) es = do
 specialize' (Lambda       _ ts e) es = do
   e' <- specialize e
   return $ apply (Lambda NoSpanInfo ts e') es
-specialize' (Let          _ ds e) es = do
+specialize' (Let        _ _ ds e) es = do
   ds' <- mapM specialize ds
   e' <- specialize e
-  return $ apply (Let NoSpanInfo ds' e') es
-specialize' (Case      _ ct e as) es = do
+  return $ apply (mkLet ds' e') es
+specialize' (Case    _ _ ct e as) es = do
   e' <- specialize e
   as' <- mapM specialize as
-  return $ apply (Case NoSpanInfo ct e' as') es
+  return $ apply (mkCase ct e' as') es
 specialize' e                   _  =
   internalError $ "Dictionary.specialize': " ++ show e
 
@@ -1061,12 +871,12 @@ instance Specialize Alt where
 -- transformation.
 
 cleanup :: Module a -> DTM (Module a)
-cleanup (Module spi ps m es is ds) = do
+cleanup (Module spi li ps m es is ds) = do
   cleanedEs <- traverse cleanupExportSpec es
   cleanedDs <- concatMapM cleanupInfixDecl ds
   cleanupTyConsEnv
   cleanupPrecEnv
-  return $ Module spi ps m cleanedEs is cleanedDs
+  return $ Module spi li ps m cleanedEs is cleanedDs
 
 cleanupExportSpec :: ExportSpec -> DTM ExportSpec
 cleanupExportSpec (Exporting p es) = Exporting p <$> concatMapM cleanupExport es
@@ -1160,18 +970,8 @@ dictTransIDecl m vEnv clsEnv (IInstanceDecl _ _ cls ty _ mm) =
         ms   = classMethods qcls clsEnv
 
 dictTransIConstrDecl :: ModuleIdent -> [Ident] -> ConstrDecl -> ConstrDecl
-dictTransIConstrDecl m tvs (ConstrDecl     p evs cx c tys) =
-  ConstrDecl p evs [] c $ transformIContext m (tvs ++ evs) cx ++ tys
-dictTransIConstrDecl m tvs (ConOpDecl p evs cx ty1 op ty2) =
-  dictTransIConstrDecl m tvs (ConstrDecl p evs cx op [ty1, ty2])
-dictTransIConstrDecl m tvs (RecordDecl      p evs cx c fs) =
-  RecordDecl p evs [] c $
-    map toFieldDecl (transformIContext m (tvs ++ evs) cx) ++ fs
-  where toFieldDecl = FieldDecl NoSpanInfo [anonId]
-
-transformIContext :: ModuleIdent -> [Ident] -> Context -> [TypeExpr]
-transformIContext m tvs cx =
-   map (fromQualType m tvs . dictType) (Set.toAscList $ toQualPredSet m tvs cx)
+dictTransIConstrDecl _ _ (ConOpDecl p ty1 op ty2) = ConstrDecl p op [ty1, ty2]
+dictTransIConstrDecl _ _ cd                       = cd
 
 iFunctionDeclFromValue :: ModuleIdent -> ValueEnv -> QualIdent -> IDecl
 iFunctionDeclFromValue m vEnv f = case qualLookupValue f vEnv of
@@ -1183,10 +983,9 @@ iFunctionDeclFromValue m vEnv f = case qualLookupValue f vEnv of
 iConstrDeclFromDataConstructor :: ModuleIdent -> ValueEnv -> QualIdent
                                -> ConstrDecl
 iConstrDeclFromDataConstructor m vEnv c = case qualLookupValue c vEnv of
-  [DataConstructor _ _ _ (ForAllExist n n' pty)] ->
-    ConstrDecl NoSpanInfo evs [] (unqualify c) tys
-    where evs = take n' $ drop n identSupply
-          tys = map (fromQualType m identSupply) $ arrowArgs $ unpredType pty
+  [DataConstructor _ _ _ (ForAll _ pty)] ->
+    ConstrDecl NoSpanInfo (unqualify c) tys
+    where tys = map (fromQualType m identSupply) $ arrowArgs $ unpredType pty
   _ -> internalError $ "Dictionary.iConstrDeclFromDataConstructor: " ++ show c
 
 -- -----------------------------------------------------------------------------
@@ -1246,6 +1045,9 @@ freshVar name ty = ((,) ty) . mkIdent . (name ++) .  show <$> getNextId
 -- The function 'dictType' returns the type of the dictionary corresponding to
 -- a particular C-T instance.
 
+rtDictType :: Pred -> Type
+rtDictType = TypeArrow unitType . dictType
+
 dictType :: Pred -> Type
 dictType (Pred cls ty) = TypeApply (TypeConstructor $ qDictTypeId cls) ty
 
@@ -1254,7 +1056,7 @@ dictType (Pred cls ty) = TypeApply (TypeConstructor $ qDictTypeId cls) ty
 
 transformPredType :: PredType -> Type
 transformPredType (PredType ps ty) =
-  foldr (TypeArrow . dictType) ty $ Set.toList ps
+  foldr (TypeArrow . rtDictType) ty $ Set.toList ps
 
 -- The function 'transformMethodPredType' first deletes the implicit class
 -- constraint and then transforms the resulting predicated type as above.
@@ -1322,8 +1124,8 @@ varType m v vEnv = case qualLookupValue (qualify v) vEnv of
 
 conType :: QualIdent -> ValueEnv -> TypeScheme
 conType c vEnv = case qualLookupValue c vEnv of
-  [DataConstructor  _ _ _ (ForAllExist n _ pty)] -> ForAll n pty
-  [NewtypeConstructor _ _ (ForAllExist n _ pty)] -> ForAll n pty
+  [DataConstructor  _ _ _ (ForAll n pty)] -> ForAll n pty
+  [NewtypeConstructor _ _ (ForAll n pty)] -> ForAll n pty
   _ -> internalError $ "Dictionary.conType: " ++ show c
 
 funType :: QualIdent -> ValueEnv -> TypeScheme
@@ -1334,8 +1136,8 @@ funType f vEnv = case qualLookupValue f vEnv of
 
 opType :: QualIdent -> ValueEnv -> TypeScheme
 opType op vEnv = case qualLookupValue op vEnv of
-  [DataConstructor  _ _ _ (ForAllExist n _ pty)] -> ForAll n pty
-  [NewtypeConstructor _ _ (ForAllExist n _ pty)] -> ForAll n pty
+  [DataConstructor  _ _ _ (ForAll n pty)] -> ForAll n pty
+  [NewtypeConstructor _ _ (ForAll n pty)] -> ForAll n pty
   [Value _ _ _                             tySc] -> tySc
   [Label _ _                               tySc] -> tySc
   _ -> internalError $ "Dictionary.opType " ++ show op
